/* ===========================================================================
Copyright (C) 2002-2005  3Dlabs Inc. Ltd.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    Neither the name of 3Dlabs Inc. Ltd. nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
=========================================================================== */

//
// The source of the shaders itself are in plain text files. Files ending in
// .vert are vertex shaders, and files ending in .frag are fragment shaders.
//

#include <antiglut.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#ifdef __WIN32__ /*[*/
#include <io.h>
#else
#include <sys/types.h>
#include <unistd.h>
#endif /*]*/

////////////////
// Structures //
////////////////

//
// Shader types
//
typedef enum {
    EVertexShader,
    EFragmentShader,
} EShaderType;


////////////////
// Prototypes //
////////////////

static int printOglError(char *file, int line);

#define printOpenGLError() printOglError(__FILE__, __LINE__)

//
// Handle of the window we're rendering to
//
static GLint window;

//
// Movement variables
//
static float fXDiff = 206;
static float fYDiff = 16;
static float fZDiff = 10;
static int xLastIncr = 0;
static int yLastIncr = 0;
static float fXInertia = -0.5;
static float fYInertia = 0;
static float fXInertiaOld;
static float fYInertiaOld;
static float fScale = 1.0;
static float ftime = 0;
static int xLast = -1;
static int yLast = -1;
static char bmModifiers;
static int Rotate = 1;

//
// Rotation defines
//
#define INERTIA_THRESHOLD       1.0f
#define INERTIA_FACTOR          0.5f
#define SCALE_FACTOR            0.01f
#define SCALE_INCREMENT         0.1f
#define TIMER_FREQUENCY_MILLIS  20

static GLfloat RotL = 1 * 3.14f / 180;
static int LastTime = 0;

//
// Handles for the currently active program object, with its two shader objects
//
static GLuint ProgramObject = 0;
static GLuint VertexShaderObject = 0;
static GLuint FragmentShaderObject = 0;

static void drawCube(void)
{
    #define	size 1.0f
    #define scale 0.2f
    #define	delta 0.1f

    float A[3] = { size,  size,  size * scale + delta };
    float B[3] = { size,  size, -size * scale + delta };
    float C[3] = { size, -size, -size * scale };
    float D[3] = { size, -size,  size * scale };
    float E[3] = {-size,  size,  size * scale + delta };
    float F[3] = {-size,  size, -size * scale + delta };
    float G[3] = {-size, -size, -size * scale };
    float H[3] = {-size, -size,  size * scale };

    float I[3] = { 1.0f,  0.0f,  0.0f};
    float K[3] = {-1.0f,  0.0f,  0.0f};
    float L[3] = { 0.0f,  0.0f, -1.0f};
    float M[3] = { 0.0f,  0.0f,  1.0f};
    float N[3] = { 0.0f,  1.0f,  0.0f};
    float O[3] = { 0.0f, -1.0f,  0.0f};


    glBegin(GL_QUADS);

    glNormal3fv(I);

    glTexCoord2f(1,1);
    glVertex3fv(D);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(A);

    glNormal3fv(K);

    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(H);
    glTexCoord2f(0,0);
    glVertex3fv(E);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(L);

    glTexCoord2f(1,1);
    glVertex3fv(C);
    glTexCoord2f(0,1);
    glVertex3fv(G);
    glTexCoord2f(0,0);
    glVertex3fv(F);
    glTexCoord2f(1,0);
    glVertex3fv(B);

    glNormal3fv(M);

    glTexCoord2f(1,1);
    glVertex3fv(H);
    glTexCoord2f(0,1);
    glVertex3fv(D);
    glTexCoord2f(0,0);
    glVertex3fv(A);
    glTexCoord2f(1,0);
    glVertex3fv(E);

    glNormal3fv(N);

    glTexCoord2f(1,1);
    glVertex3fv(E);
    glTexCoord2f(0,1);
    glVertex3fv(A);
    glTexCoord2f(0,0);
    glVertex3fv(B);
    glTexCoord2f(1,0);
    glVertex3fv(F);

    glNormal3fv(O);
    
    glTexCoord2f(1,1);
    glVertex3fv(G);
    glTexCoord2f(0,1);
    glVertex3fv(C);
    glTexCoord2f(0,0);
    glVertex3fv(D);
    glTexCoord2f(1,0);
    glVertex3fv(H);
    
    glEnd();
#undef size
#undef scale
#undef delta

}

//
// Get the location of a uniform variable
//
static
GLint getUniLoc(GLuint program, const GLchar *name)
{
    GLint loc;

    loc = glGetUniformLocation(program, name);

    if (loc == -1)
        printf("No such uniform named \"%s\"\n", name);

    printOpenGLError();  // Check for OpenGL errors
    return loc;
}

//
// Print out the information log for a shader object
//
static
void printShaderInfoLog(GLuint shader)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  // Check for OpenGL errors

    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  // Check for OpenGL errors

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
        printf("Shader InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  // Check for OpenGL errors
}


//
// Print out the information log for a program object
//
static
void printProgramInfoLog(GLuint program)
{
    int infologLength = 0;
    int charsWritten  = 0;
    GLchar *infoLog;

    printOpenGLError();  // Check for OpenGL errors

    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);

    printOpenGLError();  // Check for OpenGL errors

    if (infologLength > 0)
    {
        infoLog = (GLchar *)malloc(infologLength);
        if (infoLog == NULL)
        {
            printf("ERROR: Could not allocate InfoLog buffer\n");
            exit(1);
        }
        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
        printf("Program InfoLog:\n%s\n\n", infoLog);
        free(infoLog);
    }
    printOpenGLError();  // Check for OpenGL errors
}


static
int shaderSize(char *fileName, EShaderType shaderType)
{
    //
    // Returns the size in bytes of the shader fileName.
    // If an error occurred, it returns -1.
    //
    // File name convention:
    //
    // <fileName>.vert
    // <fileName>.frag
    //
    int fd;
    char name[256] = "glprogs/";
    int count = -1;

    strcat(name, fileName);

    switch (shaderType)
    {
        case EVertexShader:
            strcat(name, ".vert");
            break;
        case EFragmentShader:
            strcat(name, ".frag");
            break;
        default:
            printf("ERROR: unknown shader file type\n");
            exit(1);
            break;
    }

    //
    // Open the file, seek to the end to find its length
    //
#ifdef __WIN32__ /*[*/
    fd = _open(name, _O_RDONLY);
    if (fd != -1)
    {
        count = _lseek(fd, 0, SEEK_END) + 1;
        _close(fd);
    }
#else /*][*/
    fd = open(name, O_RDONLY);
    if (fd != -1)
    {
        count = lseek(fd, 0, SEEK_END) + 1;
        close(fd);
    }
#endif /*]*/

    return count;
}


static int
readShader(char *fileName, EShaderType shaderType, char *shaderText, int shader_size)
{
    //
    // Reads a shader from the supplied file and returns the shader in the
    // arrays passed in. Returns 1 if successful, 0 if an error occurred.
    // The parameter size is an upper limit of the amount of bytes to read.
    // It is ok for it to be too big.
    //
    FILE *fh;
    char name[256] = "glprogs/";
    int count;

    strcat(name, fileName);

    switch (shaderType) 
    {
        case EVertexShader:
            strcat(name, ".vert");
            break;
        case EFragmentShader:
            strcat(name, ".frag");
            break;
        default:
            printf("ERROR: unknown shader file type\n");
            exit(1);
            break;
    }

    //
    // Open the file
    //
    fh = fopen(name, "r");
    if (!fh)
        return -1;

    //
    // Get the shader from a file.
    //
    fseek(fh, 0, SEEK_SET);
    count = (int) fread(shaderText, 1, shader_size, fh);
    shaderText[count] = '\0';

    if (ferror(fh))
        count = 0;

    fclose(fh);
    return count;
}


static
int readShaderSource(char *fileName, GLchar **vertexShader, GLchar **fragmentShader)
{
    int vSize, fSize;

    //
    // Allocate memory to hold the source of our shaders.
    //
    vSize = shaderSize(fileName, EVertexShader);
    fSize = shaderSize(fileName, EFragmentShader);

    if ((vSize == -1) || (fSize == -1))
    {
        printf("Cannot determine size of the shader %s\n", fileName);
        return 0;
    }

    *vertexShader = (GLchar *) malloc(vSize);
    *fragmentShader = (GLchar *) malloc(fSize);

    //
    // Read the source code
    //
    if (!readShader(fileName, EVertexShader, *vertexShader, vSize))
    {
        printf("Cannot read the file %s.vert\n", fileName);
        return 0;
    }

    if (!readShader(fileName, EFragmentShader, *fragmentShader, fSize))
    {
        printf("Cannot read the file %s.frag\n", fileName);
        return 0;
    }

    return 1;
}


/*public*/
static int installBrickShaders(const GLchar *brickVertex,
                        const GLchar *brickFragment)
{
    GLuint brickVS, brickFS, brickProg;   // handles to objects
    GLint  vertCompiled, fragCompiled;    // status values
    GLint  linked;

    // Create a vertex shader object and a fragment shader object

    brickVS = glCreateShader(GL_VERTEX_SHADER);
    brickFS = glCreateShader(GL_FRAGMENT_SHADER);

    // Load source code strings into shaders

    glShaderSource(brickVS, 1, &brickVertex, NULL);
    glShaderSource(brickFS, 1, &brickFragment, NULL);

    // Compile the brick vertex shader, and print out
    // the compiler log file.

    glCompileShader(brickVS);
    printOpenGLError();  // Check for OpenGL errors
    glGetShaderiv(brickVS, GL_COMPILE_STATUS, &vertCompiled);
    printShaderInfoLog(brickVS);

    // Compile the brick vertex shader, and print out
    // the compiler log file.

    glCompileShader(brickFS);
    printOpenGLError();  // Check for OpenGL errors
    glGetShaderiv(brickFS, GL_COMPILE_STATUS, &fragCompiled);
    printShaderInfoLog(brickFS);

    if (!vertCompiled || !fragCompiled)
        return 0;

    // Create a program object and attach the two compiled shaders

    brickProg = glCreateProgram();
    glAttachShader(brickProg, brickVS);
    glAttachShader(brickProg, brickFS);

    // Link the program object and print out the info log

    glLinkProgram(brickProg);
    printOpenGLError();  // Check for OpenGL errors
    glGetProgramiv(brickProg, GL_LINK_STATUS, &linked);
    printProgramInfoLog(brickProg);

    if (!linked)
        return 0;

    // Install program object as part of current state

    glUseProgram(brickProg);

    // Set up initial uniform values

    glUniform3f(getUniLoc(brickProg, "BrickColor"), 1.0f, 0.3f, 0.2f);
    glUniform3f(getUniLoc(brickProg, "MortarColor"), 0.85f, 0.86f, 0.84f);
    glUniform2f(getUniLoc(brickProg, "BrickSize"), 0.30f, 0.15f);
    glUniform2f(getUniLoc(brickProg, "BrickPct"), 0.90f, 0.85f);
    glUniform3f(getUniLoc(brickProg, "LightPosition"), 0, 0, 4.0f);

    return 1;
}

int printOglError(char *file, int line)
{
    //
    // Returns 1 if an OpenGL error occurred, 0 otherwise.
    //
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d\n", file, line);
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}


/***************************************************************************/
/* Parse GL_VERSION and return the major and minor numbers in the supplied
 * integers.
 * If it fails for any reason, major and minor will be set to 0.
 * Assumes a valid OpenGL context.
*/

static void NextClearColor(void)
{
    static int color = 0;

    switch( color++ )
    {
        case 0:  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
             break;
        case 1:  glClearColor(0.2f, 0.2f, 0.3f, 1.0f);
             break;
        default: glClearColor(0.7f, 0.7f, 0.7f, 1.0f);
             color = 0;
             break;
    }
}



/******************************************************************************/
/*
/* GLUT glue
/*
/******************************************************************************/

static
void display(void)
{
    glLoadIdentity();
	glTranslatef(0,0,-4);
	glScalef(0.5f,0.5f,0.5f);

    glRotatef(fYDiff, 1,0,0);
    glRotatef(fXDiff, 0,1,0);
    glRotatef(fZDiff, 0,0,1);

    glScalef(fScale, fScale, fScale);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    drawCube();

    glFlush();
    agSwap();
}

static
void key(unsigned scan, GLboolean press)
{
	if(!press)
		return;
    switch(scan)
    {
        case sc_b:
            NextClearColor();
            break;
        case sc_q:
        case sc_escape:
            agShutdown();
            exit(0);
            break;
        case sc_kp_add:
        case sc_equal:
        case sc_wheelup:
            fScale += SCALE_INCREMENT;
            break;
        case sc_kp_subtract:
        case sc_minus:
        case sc_wheeldown:
            fScale -= SCALE_INCREMENT;
            break;
        case sc_slash:
            fprintf(stderr, "\nKeyboard commands:\n\n"
            "b - Toggle among background clear colors\n"
            "q, <esc> - Quit\n"
            "? - Help\n"
            "<home>     - reset zoom and rotation\n"
            "<+>, <-> or <ctrl + drag> - zoom model\n"
            "<arrow keys> or <drag>    - rotate model\n\n");
            break;
    }
}

static
void mouse(unsigned press)
{
      if (press)
      {
         xLast = -1;
         yLast = -1;

         if (xLastIncr > INERTIA_THRESHOLD)
            fXInertia = (xLastIncr - INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (-xLastIncr > INERTIA_THRESHOLD) 
            fXInertia = (xLastIncr + INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (yLastIncr > INERTIA_THRESHOLD) 
            fYInertia = (yLastIncr - INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (-yLastIncr > INERTIA_THRESHOLD) 
            fYInertia = (yLastIncr + INERTIA_THRESHOLD)*INERTIA_FACTOR;
      }
      else
      {
         fXInertia = 0;
         fYInertia = 0;
      }
      xLastIncr = 0;
      yLastIncr = 0;
}


static
void motion(int x, int y,unsigned kbmods)
{
   if ((xLast != -1) || (yLast != -1))
   {
      xLastIncr = x - xLast;
      yLastIncr = y - yLast;
      if (kbmods & UI_CONTROL_MASK)
      {
         if (xLast != -1)
         {
            fZDiff += xLastIncr;
            fScale += (yLastIncr)*SCALE_FACTOR;
         }
      }
      else
      {
         if (xLast != -1)
         {
            fXDiff += xLastIncr;
            fYDiff += yLastIncr;
         }
      }
   }

   xLast = x;
   yLast = y;
}



static
void special(unsigned scan, GLboolean press)
{
	if(!press)
		return;
    switch (scan)
    {
        case sc_home:
            fXDiff    = 0;
            fYDiff    = 35;
            fZDiff    = 0;
            xLastIncr = 0;
            yLastIncr = 0;
            fXInertia = -0.5;
            fYInertia = 0;
            fScale    = 1.0;
        break;
        case sc_left:
           fXDiff--;
        break;
        case sc_right:
           fXDiff++;
        break;
        case sc_up:
           fYDiff--;
        break;
        case sc_down:
           fYDiff++;
        break;
    }
}


static void agMainLoop()
{
	ui_event_t ev;
	for(;;)
	{
		display();
		uiNextEvent(&ev);
		switch(ev.type)
		{
			case ui_key_event:
				key(ev.key.scan, ev.key.press);
				if(ev.key.scan == sc_lbutton)
					mouse(ev.key.press);
				special(ev.key.scan, ev.key.press);
				break;
			case ui_pointer_motion_event:
				if(ev.button & UI_LBUTTON_MASK)
					motion(ev.client_x, ev.client_y, ev.state);
				break;
			case ui_wm_event:
				if(ev.configure.wm_event == ui_wm_close)
					return;
			default:
				break;
		}
	}
}

/******************************************************************************/
/*
/* Main
/*
/******************************************************************************/

int main(void)
{
    int success = 0;
    GLchar *VertexShaderSource, *FragmentShaderSource;

    agWindowConfig("3Dlabs Brick Shader: AntiGlut port", 1024, 1024, GL_FALSE);
    if(!agInit(AG_REQUIRED_2_0,GL_TRUE,GL_FALSE,GL_FALSE))
    	return -1;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-1,1,-1,1,2,6);
    glMatrixMode(GL_MODELVIEW);

    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    NextClearColor();

    key(sc_slash, GL_TRUE);     // display help

    readShaderSource("brick", &VertexShaderSource, &FragmentShaderSource);
    success = installBrickShaders(VertexShaderSource, FragmentShaderSource);

    if (!success)
    {
    	agShutdown();
    	return -1;
    }
    agMainLoop();
    
    return 0;
}
