#ifndef	ANTIGLUT_H
#define	ANTIGLUT_H

/*
 * AntiGlut OpenGL helper library.
 * Version: 0.0.1
 *
 * Copyright (C) 2006 Vladimir A. Smirnov.
 *
 * #include "GPL_preamble"
 *
 * This is a limited backport from LibV framework library (as of 15 nov 2006).
 *
 * Lacks:
 *  1. Only one window can be created at the time (global data were used).
 *  2. Only GL versions 1.1 and 2.0 are supported (no extensions at all).
 *	3. No support for Delphi.
 */

/*
 * LibV abstraction layer.
 * Version 4-alpha.
 *
 * Copyright (C) 1998 - 2006 Vladimir A. Smirnov (kkatarn@rambler.ru)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so.
 *
 * <ru_koi8-r>    ,   </ru_koi8-r>
 * <ru_win1251>    ,   </ru_win1251>
 *
 * Last update: 15/nov/2006
 *
 * Important note.
 * The only question with OpenGL is an extension management.
 * Some useful stuff former available as ARB extensions now promoted to
 * core functionality; it is clear that using core functions is the
 * _right_ way to do things.
 *
 * But some stuff still aren't in the core (under Win32), and at the
 * moment of writing not all vendors (actually, almost _none_ of them)
 * provides _solid_ GL 2.0 - capable drivers (NVDA is indeed _really_
 * close to the point, though; don't know much about others).
 *
 * So, we have to implement _portable_ ways to:
 *  - determine (_after_ first call to vglInit()) the GL version
 *    and availability of _all_ functions required by that version;
 *  - determine if the specific extension is available;
 *  - instantly call any available core or extension function without
 *    that Win32-specific madness with getting pointers;
 *  - be able to fall back to different path if the specific extension
 *    isn't supported (i.e. we don't want to get 'unresolved external'
 *    errors when loading the executable).
 *
 * Moreover, under Win32 we sometimes have to call requred functions from
 * the program linked with LibV dynamically (even if the program
 * written with other language - ObjectPascal, for example).
 *
 * The best way to call GL function is via pointer stored in DLL; for C
 * program it can be done as follows:
 *
 * - in LibV's vglext.c:
 *	__declspec(dllexport) (__stdcall *proc_ptr)(...);
 *	...
 *	proc_ptr = load_gl_proc(...);
 *
 * - in program dynamically linked with LibV.dll:
 *	extern __declspec(dllimport) (__stdcall *proc_ptr)(...);
 *	... // init_extensions
 *	proc_ptr(...);
 *
 * But at least one of us - either me or dcc32.exe - is stupid. The
 * above method just don't works with Delphi. Even such a hack:
 *
 *	procedure dead_ptr(...);stdcall;external 'LibV.dll' name 'proc_ptr';
 *	type proc_t = procedure(...);stdcall;
 *	var
 *		proc_ptr : proc_t;
 *	begin
 *		asm
 *			mov eax, offset dead_ptr
 *			mov proc_ptr, [eax]
 *		end;
 *		proc_ptr(...);
 *	end.
 *
 * ... is useless. It looks (after endless debugging sessions with wdw.exe
 * and td32.exe) like Delphi uses some kind of jump table; trying to pass
 * through it _is_ insane.
 *
 * So, the (ugly...) way as it seems to me:
 *
 *	#ifdef __WIN32__
 *		#ifdef	BUILD_LIBV_DLL
 *			// we are compiling the DLL
 *			#define VGLEXTAPI __declspec(dllexport)
 *			#ifdef	DLL_TARGET_DCC
 *				// for Delphi
 *				in GL.h declare VGLEXTAPI extern gate functions;
 *				in vglext.c declare local pointers to real functions,
 *				initialize pointers and call them from gate functions.
 *			#else // !DLL_TARGET_DCC
 *				in GL.h declare VGLEXTAPI extern _pointers_;
 *				in vglext.c use wglGetProcAddress() and initialize pointers.
 *			#endif // !DLL_TARGET_DCC
 *		#else // !BUILD_LIBV_DLL
 *			// Client C program or static library
 *			#ifdef DYNAMIC_LIBV_LINK
 *				// Client have to be linked with libv.dll later
 *				#define VGLEXTAPI GLAPI - __declspec(dllimport)
 *			#else // !DYNAMIC_LIBV_LINK
 *				// Statically linked client or static library
 *				#define VGLEXTAPI
 *			#endif // !DYNAMIC_LIBV_LINK
 *			in GL.h declare extern _pointers_;
 *			in vglext.c use wglGetProcAddress() and initialize pointers.
 *		#endif // !BUILD_LIBV_DLL
 *	#else // !__WIN32__ -- thanks God !
 *		in GL.h declare extern _pointers_;
 *		in vglext.c use glXGetProcAddress() (_right_ way) or
 *		dlopen("libGL.so") (hack) and initialize pointers.
 *	#else // !__WIN32__
 *
 * Another Win32 note: according to Silicon Graphics, "there is an
 * unavoidable inconsistency with wglGetProcAddress, which returns
 * context-dependent pointers."
 *
 * Actually, it is not that bad: for the same GL implementation
 * wglGetProcAddress() always returns same pointers. So, it has to be said
 * the pointers are implementation-specific, not context-specific (at
 * least, this is true for my machine with NV 7X.xxx GL driver). Because
 * of this, it is a bad idea to protect pointers in the VGL context (call
 * via protected pointer requires the hold of global lock). We just either
 * do _not_ allow to switch to another implementation (for example, to
 * Microsoft software GL after HW GL) after first call to vglCreateWindow,
 * or do _not_ allow any extended stuff after such a switch.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * Availability of extension functions must be detected at run-time
 * _after_ successful call to vglInit() in such a way:
 *
 *	if(!vglHaveExt(GL_ARB_multitexture))
 *		die("brain-dead renderer does not support GL_ARB_multitexture.");
 *	else
 *		trace("GL_ARB_multitexture usable.");
 *	where
 *
 */

/*
 * Mesa 3-D graphics library
 * Version:  6.2
 *
 * Copyright (C) 1999-2004  Brian Paul   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Config macros
 */

/*
 * Trace each call to the malloc/free.
 */
/*
#define	MALLOC_VERBOSE
*/

/*
 * Output extra debug and info messages.
 */
#define	AG_VERBOSE

/*
 * Determine host platform
 */
#if !defined(__UNIX__) && !defined(__WIN32__)

/* these compilers are not cross-platform */
#if defined(_MSC_VER) || defined (__WATCOMC__) || defined(__BORLANDC__)
#define __WIN32__
/* under Linux, we can rely on __linux */
#elif defined(__linux)
#define __UNIX__
#endif /* defined(__linux) */

#if !defined(__UNIX__) && !defined(__WIN32__)
#error "You must explicitly specify host platform"
#endif	/* !defined(__UNIX__) && !defined(__WIN32__) */

#endif	/* !defined(__UNIX__) && !defined(__WIN32__) */

/*
 * Check for erroneous #include "GL/gl.h"
 */
#ifdef	__gl_h_

#error	"System-specific OpenGL stuff can't be used with AnliGlut."
#error	"All you need are probably already in this header."
#error	"Please, locate and remove all \"#include \"GL/gl.h\" or"
#error	"replace them with \"#include \"antiglut.h\"."
#else	/* !__gl_h_ */

/*
 * Prevent later inclusion of system-specific stuff
 */
#define	__gl_h_
#endif	/* !__gl_h_ */

#ifdef	__WIN32__

#ifdef	_MSC_VER
#define GLAPI extern __declspec(dllimport)
#else	/* !_MSC_VER */
#define GLAPI extern
#endif	/* !_MSC_VER */
#define VGLEXTAPI extern

#define GLAPIENTRY __stdcall
#define	VGLEXTAPIENTRY	__stdcall

#else	/* !__WIN32__ */

#define GLAPI extern
#define VGLEXTAPI extern
#define GLAPIENTRY
#define	VGLEXTAPIENTRY

#endif /* !__WIN32__ */

#include <stddef.h> /* for ptrdiff_t */

/*
 * ===================== Part I: OpenGL 1.1 / 2.0 stuff ==================
 */

typedef unsigned int GLenum;
typedef unsigned char GLboolean;
typedef unsigned int GLbitfield;
typedef signed char GLbyte;
typedef short GLshort;
typedef int GLint;
typedef int GLsizei;
typedef unsigned char GLubyte;
typedef unsigned short GLushort;
typedef unsigned int GLuint;
typedef float GLfloat;
typedef float GLclampf;
typedef double GLdouble;
typedef double GLclampd;
typedef void GLvoid;

#define GL_VERSION_1_1                      1

/* AttribMask */
#define GL_CURRENT_BIT                      0x00000001
#define GL_POINT_BIT                        0x00000002
#define GL_LINE_BIT                         0x00000004
#define GL_POLYGON_BIT                      0x00000008
#define GL_POLYGON_STIPPLE_BIT              0x00000010
#define GL_PIXEL_MODE_BIT                   0x00000020
#define GL_LIGHTING_BIT                     0x00000040
#define GL_FOG_BIT                          0x00000080
#define GL_DEPTH_BUFFER_BIT                 0x00000100
#define GL_ACCUM_BUFFER_BIT                 0x00000200
#define GL_STENCIL_BUFFER_BIT               0x00000400
#define GL_VIEWPORT_BIT                     0x00000800
#define GL_TRANSFORM_BIT                    0x00001000
#define GL_ENABLE_BIT                       0x00002000
#define GL_COLOR_BUFFER_BIT                 0x00004000
#define GL_HINT_BIT                         0x00008000
#define GL_EVAL_BIT                         0x00010000
#define GL_LIST_BIT                         0x00020000
#define GL_TEXTURE_BIT                      0x00040000
#define GL_SCISSOR_BIT                      0x00080000
#define GL_ALL_ATTRIB_BITS                  0x000fffff

/* ClientAttribMask */
#define GL_CLIENT_PIXEL_STORE_BIT           0x00000001
#define GL_CLIENT_VERTEX_ARRAY_BIT          0x00000002
#define GL_CLIENT_ALL_ATTRIB_BITS           0xFFFFFFFF

/* Boolean */
#define GL_FALSE                            0
#define GL_TRUE                             1

/* BeginMode */
#define GL_POINTS                           0x0000
#define GL_LINES                            0x0001
#define GL_LINE_LOOP                        0x0002
#define GL_LINE_STRIP                       0x0003
#define GL_TRIANGLES                        0x0004
#define GL_TRIANGLE_STRIP                   0x0005
#define GL_TRIANGLE_FAN                     0x0006
#define GL_QUADS                            0x0007
#define GL_QUAD_STRIP                       0x0008
#define GL_POLYGON                          0x0009

/* AccumOp */
#define GL_ACCUM                            0x0100
#define GL_LOAD                             0x0101
#define GL_RETURN                           0x0102
#define GL_MULT                             0x0103
#define GL_ADD                              0x0104

/* AlphaFunction */
#define GL_NEVER                            0x0200
#define GL_LESS                             0x0201
#define GL_EQUAL                            0x0202
#define GL_LEQUAL                           0x0203
#define GL_GREATER                          0x0204
#define GL_NOTEQUAL                         0x0205
#define GL_GEQUAL                           0x0206
#define GL_ALWAYS                           0x0207

/* BlendingFactor */
#define GL_ZERO                             0
#define GL_ONE                              1
#define GL_SRC_COLOR                        0x0300
#define GL_ONE_MINUS_SRC_COLOR              0x0301
#define GL_SRC_ALPHA                        0x0302
#define GL_ONE_MINUS_SRC_ALPHA              0x0303
#define GL_DST_ALPHA                        0x0304
#define GL_ONE_MINUS_DST_ALPHA              0x0305
#define GL_DST_COLOR                        0x0306
#define GL_ONE_MINUS_DST_COLOR              0x0307
#define GL_SRC_ALPHA_SATURATE               0x0308

/* DrawBufferMode */
#define GL_NONE                             0
#define GL_FRONT_LEFT                       0x0400
#define GL_FRONT_RIGHT                      0x0401
#define GL_BACK_LEFT                        0x0402
#define GL_BACK_RIGHT                       0x0403
#define GL_FRONT                            0x0404
#define GL_BACK                             0x0405
#define GL_LEFT                             0x0406
#define GL_RIGHT                            0x0407
#define GL_FRONT_AND_BACK                   0x0408
#define GL_AUX0                             0x0409
#define GL_AUX1                             0x040A
#define GL_AUX2                             0x040B
#define GL_AUX3                             0x040C

/* ErrorCode */
#define GL_NO_ERROR                         0
#define GL_INVALID_ENUM                     0x0500
#define GL_INVALID_VALUE                    0x0501
#define GL_INVALID_OPERATION                0x0502
#define GL_STACK_OVERFLOW                   0x0503
#define GL_STACK_UNDERFLOW                  0x0504
#define GL_OUT_OF_MEMORY                    0x0505

/* FeedbackType */
#define GL_2D                               0x0600
#define GL_3D                               0x0601
#define GL_3D_COLOR                         0x0602
#define GL_3D_COLOR_TEXTURE                 0x0603
#define GL_4D_COLOR_TEXTURE                 0x0604

/* FeedBackToken */
#define GL_PASS_THROUGH_TOKEN               0x0700
#define GL_POINT_TOKEN                      0x0701
#define GL_LINE_TOKEN                       0x0702
#define GL_POLYGON_TOKEN                    0x0703
#define GL_BITMAP_TOKEN                     0x0704
#define GL_DRAW_PIXEL_TOKEN                 0x0705
#define GL_COPY_PIXEL_TOKEN                 0x0706
#define GL_LINE_RESET_TOKEN                 0x0707

/* FogMode */
#define GL_EXP                              0x0800
#define GL_EXP2                             0x0801

/* FrontFaceDirection */
#define GL_CW                               0x0900
#define GL_CCW                              0x0901

/* GetMapQuery */
#define GL_COEFF                            0x0A00
#define GL_ORDER                            0x0A01
#define GL_DOMAIN                           0x0A02

/* GetPixelMap */
#define GL_PIXEL_MAP_I_TO_I                 0x0C70
#define GL_PIXEL_MAP_S_TO_S                 0x0C71
#define GL_PIXEL_MAP_I_TO_R                 0x0C72
#define GL_PIXEL_MAP_I_TO_G                 0x0C73
#define GL_PIXEL_MAP_I_TO_B                 0x0C74
#define GL_PIXEL_MAP_I_TO_A                 0x0C75
#define GL_PIXEL_MAP_R_TO_R                 0x0C76
#define GL_PIXEL_MAP_G_TO_G                 0x0C77
#define GL_PIXEL_MAP_B_TO_B                 0x0C78
#define GL_PIXEL_MAP_A_TO_A                 0x0C79

/* GetPointervPName */
#define GL_VERTEX_ARRAY_POINTER             0x808E
#define GL_NORMAL_ARRAY_POINTER             0x808F
#define GL_COLOR_ARRAY_POINTER              0x8090
#define GL_INDEX_ARRAY_POINTER              0x8091
#define GL_TEXTURE_COORD_ARRAY_POINTER      0x8092
#define GL_EDGE_FLAG_ARRAY_POINTER          0x8093

/* GetPName */
#define GL_CURRENT_COLOR                    0x0B00
#define GL_CURRENT_INDEX                    0x0B01
#define GL_CURRENT_NORMAL                   0x0B02
#define GL_CURRENT_TEXTURE_COORDS           0x0B03
#define GL_CURRENT_RASTER_COLOR             0x0B04
#define GL_CURRENT_RASTER_INDEX             0x0B05
#define GL_CURRENT_RASTER_TEXTURE_COORDS    0x0B06
#define GL_CURRENT_RASTER_POSITION          0x0B07
#define GL_CURRENT_RASTER_POSITION_VALID    0x0B08
#define GL_CURRENT_RASTER_DISTANCE          0x0B09
#define GL_POINT_SMOOTH                     0x0B10
#define GL_POINT_SIZE                       0x0B11
#define GL_POINT_SIZE_RANGE                 0x0B12
#define GL_POINT_SIZE_GRANULARITY           0x0B13
#define GL_LINE_SMOOTH                      0x0B20
#define GL_LINE_WIDTH                       0x0B21
#define GL_LINE_WIDTH_RANGE                 0x0B22
#define GL_LINE_WIDTH_GRANULARITY           0x0B23
#define GL_LINE_STIPPLE                     0x0B24
#define GL_LINE_STIPPLE_PATTERN             0x0B25
#define GL_LINE_STIPPLE_REPEAT              0x0B26
#define GL_LIST_MODE                        0x0B30
#define GL_MAX_LIST_NESTING                 0x0B31
#define GL_LIST_BASE                        0x0B32
#define GL_LIST_INDEX                       0x0B33
#define GL_POLYGON_MODE                     0x0B40
#define GL_POLYGON_SMOOTH                   0x0B41
#define GL_POLYGON_STIPPLE                  0x0B42
#define GL_EDGE_FLAG                        0x0B43
#define GL_CULL_FACE                        0x0B44
#define GL_CULL_FACE_MODE                   0x0B45
#define GL_FRONT_FACE                       0x0B46
#define GL_LIGHTING                         0x0B50
#define GL_LIGHT_MODEL_LOCAL_VIEWER         0x0B51
#define GL_LIGHT_MODEL_TWO_SIDE             0x0B52
#define GL_LIGHT_MODEL_AMBIENT              0x0B53
#define GL_SHADE_MODEL                      0x0B54
#define GL_COLOR_MATERIAL_FACE              0x0B55
#define GL_COLOR_MATERIAL_PARAMETER         0x0B56
#define GL_COLOR_MATERIAL                   0x0B57
#define GL_FOG                              0x0B60
#define GL_FOG_INDEX                        0x0B61
#define GL_FOG_DENSITY                      0x0B62
#define GL_FOG_START                        0x0B63
#define GL_FOG_END                          0x0B64
#define GL_FOG_MODE                         0x0B65
#define GL_FOG_COLOR                        0x0B66
#define GL_DEPTH_RANGE                      0x0B70
#define GL_DEPTH_TEST                       0x0B71
#define GL_DEPTH_WRITEMASK                  0x0B72
#define GL_DEPTH_CLEAR_VALUE                0x0B73
#define GL_DEPTH_FUNC                       0x0B74
#define GL_ACCUM_CLEAR_VALUE                0x0B80
#define GL_STENCIL_TEST                     0x0B90
#define GL_STENCIL_CLEAR_VALUE              0x0B91
#define GL_STENCIL_FUNC                     0x0B92
#define GL_STENCIL_VALUE_MASK               0x0B93
#define GL_STENCIL_FAIL                     0x0B94
#define GL_STENCIL_PASS_DEPTH_FAIL          0x0B95
#define GL_STENCIL_PASS_DEPTH_PASS          0x0B96
#define GL_STENCIL_REF                      0x0B97
#define GL_STENCIL_WRITEMASK                0x0B98
#define GL_MATRIX_MODE                      0x0BA0
#define GL_NORMALIZE                        0x0BA1
#define GL_VIEWPORT                         0x0BA2
#define GL_MODELVIEW_STACK_DEPTH            0x0BA3
#define GL_PROJECTION_STACK_DEPTH           0x0BA4
#define GL_TEXTURE_STACK_DEPTH              0x0BA5
#define GL_MODELVIEW_MATRIX                 0x0BA6
#define GL_PROJECTION_MATRIX                0x0BA7
#define GL_TEXTURE_MATRIX                   0x0BA8
#define GL_ATTRIB_STACK_DEPTH               0x0BB0
#define GL_CLIENT_ATTRIB_STACK_DEPTH        0x0BB1
#define GL_ALPHA_TEST                       0x0BC0
#define GL_ALPHA_TEST_FUNC                  0x0BC1
#define GL_ALPHA_TEST_REF                   0x0BC2
#define GL_DITHER                           0x0BD0
#define GL_BLEND_DST                        0x0BE0
#define GL_BLEND_SRC                        0x0BE1
#define GL_BLEND                            0x0BE2
#define GL_LOGIC_OP_MODE                    0x0BF0
#define GL_INDEX_LOGIC_OP                   0x0BF1
#define GL_LOGIC_OP                         GL_INDEX_LOGIC_OP
#define GL_COLOR_LOGIC_OP                   0x0BF2
#define GL_AUX_BUFFERS                      0x0C00
#define GL_DRAW_BUFFER                      0x0C01
#define GL_READ_BUFFER                      0x0C02
#define GL_SCISSOR_BOX                      0x0C10
#define GL_SCISSOR_TEST                     0x0C11
#define GL_INDEX_CLEAR_VALUE                0x0C20
#define GL_INDEX_WRITEMASK                  0x0C21
#define GL_COLOR_CLEAR_VALUE                0x0C22
#define GL_COLOR_WRITEMASK                  0x0C23
#define GL_INDEX_MODE                       0x0C30
#define GL_RGBA_MODE                        0x0C31
#define GL_DOUBLEBUFFER                     0x0C32
#define GL_STEREO                           0x0C33
#define GL_RENDER_MODE                      0x0C40
#define GL_PERSPECTIVE_CORRECTION_HINT      0x0C50
#define GL_POINT_SMOOTH_HINT                0x0C51
#define GL_LINE_SMOOTH_HINT                 0x0C52
#define GL_POLYGON_SMOOTH_HINT              0x0C53
#define GL_FOG_HINT                         0x0C54
#define GL_TEXTURE_GEN_S                    0x0C60
#define GL_TEXTURE_GEN_T                    0x0C61
#define GL_TEXTURE_GEN_R                    0x0C62
#define GL_TEXTURE_GEN_Q                    0x0C63
#define GL_PIXEL_MAP_I_TO_I_SIZE            0x0CB0
#define GL_PIXEL_MAP_S_TO_S_SIZE            0x0CB1
#define GL_PIXEL_MAP_I_TO_R_SIZE            0x0CB2
#define GL_PIXEL_MAP_I_TO_G_SIZE            0x0CB3
#define GL_PIXEL_MAP_I_TO_B_SIZE            0x0CB4
#define GL_PIXEL_MAP_I_TO_A_SIZE            0x0CB5
#define GL_PIXEL_MAP_R_TO_R_SIZE            0x0CB6
#define GL_PIXEL_MAP_G_TO_G_SIZE            0x0CB7
#define GL_PIXEL_MAP_B_TO_B_SIZE            0x0CB8
#define GL_PIXEL_MAP_A_TO_A_SIZE            0x0CB9
#define GL_UNPACK_SWAP_BYTES                0x0CF0
#define GL_UNPACK_LSB_FIRST                 0x0CF1
#define GL_UNPACK_ROW_LENGTH                0x0CF2
#define GL_UNPACK_SKIP_ROWS                 0x0CF3
#define GL_UNPACK_SKIP_PIXELS               0x0CF4
#define GL_UNPACK_ALIGNMENT                 0x0CF5
#define GL_PACK_SWAP_BYTES                  0x0D00
#define GL_PACK_LSB_FIRST                   0x0D01
#define GL_PACK_ROW_LENGTH                  0x0D02
#define GL_PACK_SKIP_ROWS                   0x0D03
#define GL_PACK_SKIP_PIXELS                 0x0D04
#define GL_PACK_ALIGNMENT                   0x0D05
#define GL_MAP_COLOR                        0x0D10
#define GL_MAP_STENCIL                      0x0D11
#define GL_INDEX_SHIFT                      0x0D12
#define GL_INDEX_OFFSET                     0x0D13
#define GL_RED_SCALE                        0x0D14
#define GL_RED_BIAS                         0x0D15
#define GL_ZOOM_X                           0x0D16
#define GL_ZOOM_Y                           0x0D17
#define GL_GREEN_SCALE                      0x0D18
#define GL_GREEN_BIAS                       0x0D19
#define GL_BLUE_SCALE                       0x0D1A
#define GL_BLUE_BIAS                        0x0D1B
#define GL_ALPHA_SCALE                      0x0D1C
#define GL_ALPHA_BIAS                       0x0D1D
#define GL_DEPTH_SCALE                      0x0D1E
#define GL_DEPTH_BIAS                       0x0D1F
#define GL_MAX_EVAL_ORDER                   0x0D30
#define GL_MAX_LIGHTS                       0x0D31
#define GL_MAX_CLIP_PLANES                  0x0D32
#define GL_MAX_TEXTURE_SIZE                 0x0D33
#define GL_MAX_PIXEL_MAP_TABLE              0x0D34
#define GL_MAX_ATTRIB_STACK_DEPTH           0x0D35
#define GL_MAX_MODELVIEW_STACK_DEPTH        0x0D36
#define GL_MAX_NAME_STACK_DEPTH             0x0D37
#define GL_MAX_PROJECTION_STACK_DEPTH       0x0D38
#define GL_MAX_TEXTURE_STACK_DEPTH          0x0D39
#define GL_MAX_VIEWPORT_DIMS                0x0D3A
#define GL_MAX_CLIENT_ATTRIB_STACK_DEPTH    0x0D3B
#define GL_SUBPIXEL_BITS                    0x0D50
#define GL_INDEX_BITS                       0x0D51
#define GL_RED_BITS                         0x0D52
#define GL_GREEN_BITS                       0x0D53
#define GL_BLUE_BITS                        0x0D54
#define GL_ALPHA_BITS                       0x0D55
#define GL_DEPTH_BITS                       0x0D56
#define GL_STENCIL_BITS                     0x0D57
#define GL_ACCUM_RED_BITS                   0x0D58
#define GL_ACCUM_GREEN_BITS                 0x0D59
#define GL_ACCUM_BLUE_BITS                  0x0D5A
#define GL_ACCUM_ALPHA_BITS                 0x0D5B
#define GL_NAME_STACK_DEPTH                 0x0D70
#define GL_AUTO_NORMAL                      0x0D80
#define GL_MAP1_COLOR_4                     0x0D90
#define GL_MAP1_INDEX                       0x0D91
#define GL_MAP1_NORMAL                      0x0D92
#define GL_MAP1_TEXTURE_COORD_1             0x0D93
#define GL_MAP1_TEXTURE_COORD_2             0x0D94
#define GL_MAP1_TEXTURE_COORD_3             0x0D95
#define GL_MAP1_TEXTURE_COORD_4             0x0D96
#define GL_MAP1_VERTEX_3                    0x0D97
#define GL_MAP1_VERTEX_4                    0x0D98
#define GL_MAP2_COLOR_4                     0x0DB0
#define GL_MAP2_INDEX                       0x0DB1
#define GL_MAP2_NORMAL                      0x0DB2
#define GL_MAP2_TEXTURE_COORD_1             0x0DB3
#define GL_MAP2_TEXTURE_COORD_2             0x0DB4
#define GL_MAP2_TEXTURE_COORD_3             0x0DB5
#define GL_MAP2_TEXTURE_COORD_4             0x0DB6
#define GL_MAP2_VERTEX_3                    0x0DB7
#define GL_MAP2_VERTEX_4                    0x0DB8
#define GL_MAP1_GRID_DOMAIN                 0x0DD0
#define GL_MAP1_GRID_SEGMENTS               0x0DD1
#define GL_MAP2_GRID_DOMAIN                 0x0DD2
#define GL_MAP2_GRID_SEGMENTS               0x0DD3
#define GL_TEXTURE_1D                       0x0DE0
#define GL_TEXTURE_2D                       0x0DE1
#define GL_FEEDBACK_BUFFER_POINTER          0x0DF0
#define GL_FEEDBACK_BUFFER_SIZE             0x0DF1
#define GL_FEEDBACK_BUFFER_TYPE             0x0DF2
#define GL_SELECTION_BUFFER_POINTER         0x0DF3
#define GL_SELECTION_BUFFER_SIZE            0x0DF4
#define GL_POLYGON_OFFSET_UNITS             0x2A00
#define GL_POLYGON_OFFSET_POINT             0x2A01
#define GL_POLYGON_OFFSET_LINE              0x2A02
#define GL_POLYGON_OFFSET_FILL              0x8037
#define GL_POLYGON_OFFSET_FACTOR            0x8038
#define GL_TEXTURE_BINDING_1D               0x8068
#define GL_TEXTURE_BINDING_2D               0x8069
#define GL_VERTEX_ARRAY                     0x8074
#define GL_NORMAL_ARRAY                     0x8075
#define GL_COLOR_ARRAY                      0x8076
#define GL_INDEX_ARRAY                      0x8077
#define GL_TEXTURE_COORD_ARRAY              0x8078
#define GL_EDGE_FLAG_ARRAY                  0x8079
#define GL_VERTEX_ARRAY_SIZE                0x807A
#define GL_VERTEX_ARRAY_TYPE                0x807B
#define GL_VERTEX_ARRAY_STRIDE              0x807C
#define GL_NORMAL_ARRAY_TYPE                0x807E
#define GL_NORMAL_ARRAY_STRIDE              0x807F
#define GL_COLOR_ARRAY_SIZE                 0x8081
#define GL_COLOR_ARRAY_TYPE                 0x8082
#define GL_COLOR_ARRAY_STRIDE               0x8083
#define GL_INDEX_ARRAY_TYPE                 0x8085
#define GL_INDEX_ARRAY_STRIDE               0x8086
#define GL_TEXTURE_COORD_ARRAY_SIZE         0x8088
#define GL_TEXTURE_COORD_ARRAY_TYPE         0x8089
#define GL_TEXTURE_COORD_ARRAY_STRIDE       0x808A
#define GL_EDGE_FLAG_ARRAY_STRIDE           0x808C

/* GetTextureParameter */
#define GL_TEXTURE_WIDTH                    0x1000
#define GL_TEXTURE_HEIGHT                   0x1001
#define GL_TEXTURE_INTERNAL_FORMAT          0x1003
#define GL_TEXTURE_COMPONENTS               GL_TEXTURE_INTERNAL_FORMAT
#define GL_TEXTURE_BORDER_COLOR             0x1004
#define GL_TEXTURE_BORDER                   0x1005
#define GL_TEXTURE_RED_SIZE                 0x805C
#define GL_TEXTURE_GREEN_SIZE               0x805D
#define GL_TEXTURE_BLUE_SIZE                0x805E
#define GL_TEXTURE_ALPHA_SIZE               0x805F
#define GL_TEXTURE_LUMINANCE_SIZE           0x8060
#define GL_TEXTURE_INTENSITY_SIZE           0x8061
#define GL_TEXTURE_PRIORITY                 0x8066
#define GL_TEXTURE_RESIDENT                 0x8067

/* HintMode */
#define GL_DONT_CARE                        0x1100
#define GL_FASTEST                          0x1101
#define GL_NICEST                           0x1102

/* LightParameter */
#define GL_AMBIENT                          0x1200
#define GL_DIFFUSE                          0x1201
#define GL_SPECULAR                         0x1202
#define GL_POSITION                         0x1203
#define GL_SPOT_DIRECTION                   0x1204
#define GL_SPOT_EXPONENT                    0x1205
#define GL_SPOT_CUTOFF                      0x1206
#define GL_CONSTANT_ATTENUATION             0x1207
#define GL_LINEAR_ATTENUATION               0x1208
#define GL_QUADRATIC_ATTENUATION            0x1209

/* ListMode */
#define GL_COMPILE                          0x1300
#define GL_COMPILE_AND_EXECUTE              0x1301

/* DataType */
#define GL_BYTE                             0x1400
#define GL_UNSIGNED_BYTE                    0x1401
#define GL_SHORT                            0x1402
#define GL_UNSIGNED_SHORT                   0x1403
#define GL_INT                              0x1404
#define GL_UNSIGNED_INT                     0x1405
#define GL_FLOAT                            0x1406
#define GL_2_BYTES                          0x1407
#define GL_3_BYTES                          0x1408
#define GL_4_BYTES                          0x1409
#define GL_DOUBLE                           0x140A
#define GL_DOUBLE_EXT                       0x140A

/* LogicOp */
#define GL_CLEAR                            0x1500
#define GL_AND                              0x1501
#define GL_AND_REVERSE                      0x1502
#define GL_COPY                             0x1503
#define GL_AND_INVERTED                     0x1504
#define GL_NOOP                             0x1505
#define GL_XOR                              0x1506
#define GL_OR                               0x1507
#define GL_NOR                              0x1508
#define GL_EQUIV                            0x1509
#define GL_INVERT                           0x150A
#define GL_OR_REVERSE                       0x150B
#define GL_COPY_INVERTED                    0x150C
#define GL_OR_INVERTED                      0x150D
#define GL_NAND                             0x150E
#define GL_SET                              0x150F

/* MaterialParameter */
#define GL_EMISSION                         0x1600
#define GL_SHININESS                        0x1601
#define GL_AMBIENT_AND_DIFFUSE              0x1602
#define GL_COLOR_INDEXES                    0x1603

/* MatrixMode */
#define GL_MODELVIEW                        0x1700
#define GL_PROJECTION                       0x1701
#define GL_TEXTURE                          0x1702

/* PixelCopyType */
#define GL_COLOR                            0x1800
#define GL_DEPTH                            0x1801
#define GL_STENCIL                          0x1802

/* PixelFormat */
#define GL_COLOR_INDEX                      0x1900
#define GL_STENCIL_INDEX                    0x1901
#define GL_DEPTH_COMPONENT                  0x1902
#define GL_RED                              0x1903
#define GL_GREEN                            0x1904
#define GL_BLUE                             0x1905
#define GL_ALPHA                            0x1906
#define GL_RGB                              0x1907
#define GL_RGBA                             0x1908
#define GL_LUMINANCE                        0x1909
#define GL_LUMINANCE_ALPHA                  0x190A

/* PolygonMode */
#define GL_POINT                            0x1B00
#define GL_LINE                             0x1B01
#define GL_FILL                             0x1B02

/* RenderingMode */
#define GL_RENDER                           0x1C00
#define GL_FEEDBACK                         0x1C01
#define GL_SELECT                           0x1C02

/* ShadingModel */
#define GL_FLAT                             0x1D00
#define GL_SMOOTH                           0x1D01

/* StencilOp */
#define GL_KEEP                             0x1E00
#define GL_REPLACE                          0x1E01
#define GL_INCR                             0x1E02
#define GL_DECR                             0x1E03

/* StringName */
#define GL_VENDOR                           0x1F00
#define GL_RENDERER                         0x1F01
#define GL_VERSION                          0x1F02
#define GL_EXTENSIONS                       0x1F03

/* TextureCoordName */
#define GL_S                                0x2000
#define GL_T                                0x2001
#define GL_R                                0x2002
#define GL_Q                                0x2003

/* TextureEnvMode */
#define GL_MODULATE                         0x2100
#define GL_DECAL                            0x2101

/* TextureEnvParameter */
#define GL_TEXTURE_ENV_MODE                 0x2200
#define GL_TEXTURE_ENV_COLOR                0x2201

/* TextureEnvTarget */
#define GL_TEXTURE_ENV                      0x2300

/* TextureGenMode */
#define GL_EYE_LINEAR                       0x2400
#define GL_OBJECT_LINEAR                    0x2401
#define GL_SPHERE_MAP                       0x2402

/* TextureGenParameter */
#define GL_TEXTURE_GEN_MODE                 0x2500
#define GL_OBJECT_PLANE                     0x2501
#define GL_EYE_PLANE                        0x2502

/* TextureMagFilter */
#define GL_NEAREST                          0x2600
#define GL_LINEAR                           0x2601

/* TextureMinFilter */
#define GL_NEAREST_MIPMAP_NEAREST           0x2700
#define GL_LINEAR_MIPMAP_NEAREST            0x2701
#define GL_NEAREST_MIPMAP_LINEAR            0x2702
#define GL_LINEAR_MIPMAP_LINEAR             0x2703

/* TextureParameterName */
#define GL_TEXTURE_MAG_FILTER               0x2800
#define GL_TEXTURE_MIN_FILTER               0x2801
#define GL_TEXTURE_WRAP_S                   0x2802
#define GL_TEXTURE_WRAP_T                   0x2803

/* TextureTarget */
#define GL_PROXY_TEXTURE_1D                 0x8063
#define GL_PROXY_TEXTURE_2D                 0x8064

/* TextureWrapMode */
#define GL_CLAMP                            0x2900
#define GL_REPEAT                           0x2901

/* PixelInternalFormat */
#define GL_R3_G3_B2                         0x2A10
#define GL_ALPHA4                           0x803B
#define GL_ALPHA8                           0x803C
#define GL_ALPHA12                          0x803D
#define GL_ALPHA16                          0x803E
#define GL_LUMINANCE4                       0x803F
#define GL_LUMINANCE8                       0x8040
#define GL_LUMINANCE12                      0x8041
#define GL_LUMINANCE16                      0x8042
#define GL_LUMINANCE4_ALPHA4                0x8043
#define GL_LUMINANCE6_ALPHA2                0x8044
#define GL_LUMINANCE8_ALPHA8                0x8045
#define GL_LUMINANCE12_ALPHA4               0x8046
#define GL_LUMINANCE12_ALPHA12              0x8047
#define GL_LUMINANCE16_ALPHA16              0x8048
#define GL_INTENSITY                        0x8049
#define GL_INTENSITY4                       0x804A
#define GL_INTENSITY8                       0x804B
#define GL_INTENSITY12                      0x804C
#define GL_INTENSITY16                      0x804D
#define GL_RGB4                             0x804F
#define GL_RGB5                             0x8050
#define GL_RGB8                             0x8051
#define GL_RGB10                            0x8052
#define GL_RGB12                            0x8053
#define GL_RGB16                            0x8054
#define GL_RGBA2                            0x8055
#define GL_RGBA4                            0x8056
#define GL_RGB5_A1                          0x8057
#define GL_RGBA8                            0x8058
#define GL_RGB10_A2                         0x8059
#define GL_RGBA12                           0x805A
#define GL_RGBA16                           0x805B

/* InterleavedArrayFormat */
#define GL_V2F                              0x2A20
#define GL_V3F                              0x2A21
#define GL_C4UB_V2F                         0x2A22
#define GL_C4UB_V3F                         0x2A23
#define GL_C3F_V3F                          0x2A24
#define GL_N3F_V3F                          0x2A25
#define GL_C4F_N3F_V3F                      0x2A26
#define GL_T2F_V3F                          0x2A27
#define GL_T4F_V4F                          0x2A28
#define GL_T2F_C4UB_V3F                     0x2A29
#define GL_T2F_C3F_V3F                      0x2A2A
#define GL_T2F_N3F_V3F                      0x2A2B
#define GL_T2F_C4F_N3F_V3F                  0x2A2C
#define GL_T4F_C4F_N3F_V4F                  0x2A2D

/* ClipPlaneName */
#define GL_CLIP_PLANE0                      0x3000
#define GL_CLIP_PLANE1                      0x3001
#define GL_CLIP_PLANE2                      0x3002
#define GL_CLIP_PLANE3                      0x3003
#define GL_CLIP_PLANE4                      0x3004
#define GL_CLIP_PLANE5                      0x3005

/* LightName */
#define GL_LIGHT0                           0x4000
#define GL_LIGHT1                           0x4001
#define GL_LIGHT2                           0x4002
#define GL_LIGHT3                           0x4003
#define GL_LIGHT4                           0x4004
#define GL_LIGHT5                           0x4005
#define GL_LIGHT6                           0x4006
#define GL_LIGHT7                           0x4007

/*
 * Static link: OpenGL 1.1 functions.
 */
#ifdef __cplusplus
extern "C" {
#endif

GLAPI void GLAPIENTRY glAccum (GLenum op, GLfloat value);
GLAPI void GLAPIENTRY glAlphaFunc (GLenum func, GLclampf ref);
GLAPI GLboolean GLAPIENTRY glAreTexturesResident (GLsizei n, const GLuint *textures, GLboolean *residences);
GLAPI void GLAPIENTRY glArrayElement (GLint i);
GLAPI void GLAPIENTRY glBegin (GLenum mode);
GLAPI void GLAPIENTRY glBindTexture (GLenum target, GLuint texture);
GLAPI void GLAPIENTRY glBitmap (GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap);
GLAPI void GLAPIENTRY glBlendFunc (GLenum sfactor, GLenum dfactor);
GLAPI void GLAPIENTRY glCallList (GLuint list);
GLAPI void GLAPIENTRY glCallLists (GLsizei n, GLenum type, const GLvoid *lists);
GLAPI void GLAPIENTRY glClear (GLbitfield mask);
GLAPI void GLAPIENTRY glClearAccum (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
GLAPI void GLAPIENTRY glClearColor (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
GLAPI void GLAPIENTRY glClearDepth (GLclampd depth);
GLAPI void GLAPIENTRY glClearIndex (GLfloat c);
GLAPI void GLAPIENTRY glClearStencil (GLint s);
GLAPI void GLAPIENTRY glClipPlane (GLenum plane, const GLdouble *equation);
GLAPI void GLAPIENTRY glColor3b (GLbyte red, GLbyte green, GLbyte blue);
GLAPI void GLAPIENTRY glColor3bv (const GLbyte *v);
GLAPI void GLAPIENTRY glColor3d (GLdouble red, GLdouble green, GLdouble blue);
GLAPI void GLAPIENTRY glColor3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glColor3f (GLfloat red, GLfloat green, GLfloat blue);
GLAPI void GLAPIENTRY glColor3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glColor3i (GLint red, GLint green, GLint blue);
GLAPI void GLAPIENTRY glColor3iv (const GLint *v);
GLAPI void GLAPIENTRY glColor3s (GLshort red, GLshort green, GLshort blue);
GLAPI void GLAPIENTRY glColor3sv (const GLshort *v);
GLAPI void GLAPIENTRY glColor3ub (GLubyte red, GLubyte green, GLubyte blue);
GLAPI void GLAPIENTRY glColor3ubv (const GLubyte *v);
GLAPI void GLAPIENTRY glColor3ui (GLuint red, GLuint green, GLuint blue);
GLAPI void GLAPIENTRY glColor3uiv (const GLuint *v);
GLAPI void GLAPIENTRY glColor3us (GLushort red, GLushort green, GLushort blue);
GLAPI void GLAPIENTRY glColor3usv (const GLushort *v);
GLAPI void GLAPIENTRY glColor4b (GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha);
GLAPI void GLAPIENTRY glColor4bv (const GLbyte *v);
GLAPI void GLAPIENTRY glColor4d (GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha);
GLAPI void GLAPIENTRY glColor4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glColor4f (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
GLAPI void GLAPIENTRY glColor4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glColor4i (GLint red, GLint green, GLint blue, GLint alpha);
GLAPI void GLAPIENTRY glColor4iv (const GLint *v);
GLAPI void GLAPIENTRY glColor4s (GLshort red, GLshort green, GLshort blue, GLshort alpha);
GLAPI void GLAPIENTRY glColor4sv (const GLshort *v);
GLAPI void GLAPIENTRY glColor4ub (GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha);
GLAPI void GLAPIENTRY glColor4ubv (const GLubyte *v);
GLAPI void GLAPIENTRY glColor4ui (GLuint red, GLuint green, GLuint blue, GLuint alpha);
GLAPI void GLAPIENTRY glColor4uiv (const GLuint *v);
GLAPI void GLAPIENTRY glColor4us (GLushort red, GLushort green, GLushort blue, GLushort alpha);
GLAPI void GLAPIENTRY glColor4usv (const GLushort *v);
GLAPI void GLAPIENTRY glColorMask (GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha);
GLAPI void GLAPIENTRY glColorMaterial (GLenum face, GLenum mode);
GLAPI void GLAPIENTRY glColorPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glCopyPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum type);
GLAPI void GLAPIENTRY glCopyTexImage1D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border);
GLAPI void GLAPIENTRY glCopyTexImage2D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border);
GLAPI void GLAPIENTRY glCopyTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width);
GLAPI void GLAPIENTRY glCopyTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glCullFace (GLenum mode);
GLAPI void GLAPIENTRY glDeleteLists (GLuint list, GLsizei range);
GLAPI void GLAPIENTRY glDeleteTextures (GLsizei n, const GLuint *textures);
GLAPI void GLAPIENTRY glDepthFunc (GLenum func);
GLAPI void GLAPIENTRY glDepthMask (GLboolean flag);
GLAPI void GLAPIENTRY glDepthRange (GLclampd zNear, GLclampd zFar);
GLAPI void GLAPIENTRY glDisable (GLenum cap);
GLAPI void GLAPIENTRY glDisableClientState (GLenum array);
GLAPI void GLAPIENTRY glDrawArrays (GLenum mode, GLint first, GLsizei count);
GLAPI void GLAPIENTRY glDrawBuffer (GLenum mode);
GLAPI void GLAPIENTRY glDrawElements (GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
GLAPI void GLAPIENTRY glDrawPixels (GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glEdgeFlag (GLboolean flag);
GLAPI void GLAPIENTRY glEdgeFlagPointer (GLsizei stride, const GLboolean *pointer);
GLAPI void GLAPIENTRY glEdgeFlagv (const GLboolean *flag);
GLAPI void GLAPIENTRY glEnable (GLenum cap);
GLAPI void GLAPIENTRY glEnableClientState (GLenum array);
GLAPI void GLAPIENTRY glEnd (void);
GLAPI void GLAPIENTRY glEndList (void);
GLAPI void GLAPIENTRY glEvalCoord1d (GLdouble u);
GLAPI void GLAPIENTRY glEvalCoord1dv (const GLdouble *u);
GLAPI void GLAPIENTRY glEvalCoord1f (GLfloat u);
GLAPI void GLAPIENTRY glEvalCoord1fv (const GLfloat *u);
GLAPI void GLAPIENTRY glEvalCoord2d (GLdouble u, GLdouble v);
GLAPI void GLAPIENTRY glEvalCoord2dv (const GLdouble *u);
GLAPI void GLAPIENTRY glEvalCoord2f (GLfloat u, GLfloat v);
GLAPI void GLAPIENTRY glEvalCoord2fv (const GLfloat *u);
GLAPI void GLAPIENTRY glEvalMesh1 (GLenum mode, GLint i1, GLint i2);
GLAPI void GLAPIENTRY glEvalMesh2 (GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2);
GLAPI void GLAPIENTRY glEvalPoint1 (GLint i);
GLAPI void GLAPIENTRY glEvalPoint2 (GLint i, GLint j);
GLAPI void GLAPIENTRY glFeedbackBuffer (GLsizei size, GLenum type, GLfloat *buffer);
GLAPI void GLAPIENTRY glFinish (void);
GLAPI void GLAPIENTRY glFlush (void);
GLAPI void GLAPIENTRY glFogf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glFogfv (GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glFogi (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glFogiv (GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glFrontFace (GLenum mode);
GLAPI void GLAPIENTRY glFrustum (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
GLAPI GLuint GLAPIENTRY glGenLists (GLsizei range);
GLAPI void GLAPIENTRY glGenTextures (GLsizei n, GLuint *textures);
GLAPI void GLAPIENTRY glGetBooleanv (GLenum pname, GLboolean *params);
GLAPI void GLAPIENTRY glGetClipPlane (GLenum plane, GLdouble *equation);
GLAPI void GLAPIENTRY glGetDoublev (GLenum pname, GLdouble *params);
GLAPI GLenum GLAPIENTRY glGetError (void);
GLAPI void GLAPIENTRY glGetFloatv (GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetIntegerv (GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetLightfv (GLenum light, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetLightiv (GLenum light, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetMapdv (GLenum target, GLenum query, GLdouble *v);
GLAPI void GLAPIENTRY glGetMapfv (GLenum target, GLenum query, GLfloat *v);
GLAPI void GLAPIENTRY glGetMapiv (GLenum target, GLenum query, GLint *v);
GLAPI void GLAPIENTRY glGetMaterialfv (GLenum face, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetMaterialiv (GLenum face, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetPixelMapfv (GLenum map, GLfloat *values);
GLAPI void GLAPIENTRY glGetPixelMapuiv (GLenum map, GLuint *values);
GLAPI void GLAPIENTRY glGetPixelMapusv (GLenum map, GLushort *values);
GLAPI void GLAPIENTRY glGetPointerv (GLenum pname, GLvoid* *params);
GLAPI void GLAPIENTRY glGetPolygonStipple (GLubyte *mask);
GLAPI const GLubyte * GLAPIENTRY glGetString (GLenum name);
GLAPI void GLAPIENTRY glGetTexEnvfv (GLenum target, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexEnviv (GLenum target, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexGendv (GLenum coord, GLenum pname, GLdouble *params);
GLAPI void GLAPIENTRY glGetTexGenfv (GLenum coord, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexGeniv (GLenum coord, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexImage (GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels);
GLAPI void GLAPIENTRY glGetTexLevelParameterfv (GLenum target, GLint level, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexLevelParameteriv (GLenum target, GLint level, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexParameterfv (GLenum target, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexParameteriv (GLenum target, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glHint (GLenum target, GLenum mode);
GLAPI void GLAPIENTRY glIndexMask (GLuint mask);
GLAPI void GLAPIENTRY glIndexPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glIndexd (GLdouble c);
GLAPI void GLAPIENTRY glIndexdv (const GLdouble *c);
GLAPI void GLAPIENTRY glIndexf (GLfloat c);
GLAPI void GLAPIENTRY glIndexfv (const GLfloat *c);
GLAPI void GLAPIENTRY glIndexi (GLint c);
GLAPI void GLAPIENTRY glIndexiv (const GLint *c);
GLAPI void GLAPIENTRY glIndexs (GLshort c);
GLAPI void GLAPIENTRY glIndexsv (const GLshort *c);
GLAPI void GLAPIENTRY glIndexub (GLubyte c);
GLAPI void GLAPIENTRY glIndexubv (const GLubyte *c);
GLAPI void GLAPIENTRY glInitNames (void);
GLAPI void GLAPIENTRY glInterleavedArrays (GLenum format, GLsizei stride, const GLvoid *pointer);
GLAPI GLboolean GLAPIENTRY glIsEnabled (GLenum cap);
GLAPI GLboolean GLAPIENTRY glIsList (GLuint list);
GLAPI GLboolean GLAPIENTRY glIsTexture (GLuint texture);
GLAPI void GLAPIENTRY glLightModelf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glLightModelfv (GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glLightModeli (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glLightModeliv (GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glLightf (GLenum light, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glLightfv (GLenum light, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glLighti (GLenum light, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glLightiv (GLenum light, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glLineStipple (GLint factor, GLushort pattern);
GLAPI void GLAPIENTRY glLineWidth (GLfloat width);
GLAPI void GLAPIENTRY glListBase (GLuint base);
GLAPI void GLAPIENTRY glLoadIdentity (void);
GLAPI void GLAPIENTRY glLoadMatrixd (const GLdouble *m);
GLAPI void GLAPIENTRY glLoadMatrixf (const GLfloat *m);
GLAPI void GLAPIENTRY glLoadName (GLuint name);
GLAPI void GLAPIENTRY glLogicOp (GLenum opcode);
GLAPI void GLAPIENTRY glMap1d (GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points);
GLAPI void GLAPIENTRY glMap1f (GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points);
GLAPI void GLAPIENTRY glMap2d (GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points);
GLAPI void GLAPIENTRY glMap2f (GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points);
GLAPI void GLAPIENTRY glMapGrid1d (GLint un, GLdouble u1, GLdouble u2);
GLAPI void GLAPIENTRY glMapGrid1f (GLint un, GLfloat u1, GLfloat u2);
GLAPI void GLAPIENTRY glMapGrid2d (GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2);
GLAPI void GLAPIENTRY glMapGrid2f (GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2);
GLAPI void GLAPIENTRY glMaterialf (GLenum face, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glMaterialfv (GLenum face, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glMateriali (GLenum face, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glMaterialiv (GLenum face, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glMatrixMode (GLenum mode);
GLAPI void GLAPIENTRY glMultMatrixd (const GLdouble *m);
GLAPI void GLAPIENTRY glMultMatrixf (const GLfloat *m);
GLAPI void GLAPIENTRY glNewList (GLuint list, GLenum mode);
GLAPI void GLAPIENTRY glNormal3b (GLbyte nx, GLbyte ny, GLbyte nz);
GLAPI void GLAPIENTRY glNormal3bv (const GLbyte *v);
GLAPI void GLAPIENTRY glNormal3d (GLdouble nx, GLdouble ny, GLdouble nz);
GLAPI void GLAPIENTRY glNormal3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glNormal3f (GLfloat nx, GLfloat ny, GLfloat nz);
GLAPI void GLAPIENTRY glNormal3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glNormal3i (GLint nx, GLint ny, GLint nz);
GLAPI void GLAPIENTRY glNormal3iv (const GLint *v);
GLAPI void GLAPIENTRY glNormal3s (GLshort nx, GLshort ny, GLshort nz);
GLAPI void GLAPIENTRY glNormal3sv (const GLshort *v);
GLAPI void GLAPIENTRY glNormalPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glOrtho (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
GLAPI void GLAPIENTRY glPassThrough (GLfloat token);
GLAPI void GLAPIENTRY glPixelMapfv (GLenum map, GLint mapsize, const GLfloat *values);
GLAPI void GLAPIENTRY glPixelMapuiv (GLenum map, GLint mapsize, const GLuint *values);
GLAPI void GLAPIENTRY glPixelMapusv (GLenum map, GLint mapsize, const GLushort *values);
GLAPI void GLAPIENTRY glPixelStoref (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glPixelStorei (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glPixelTransferf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glPixelTransferi (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glPixelZoom (GLfloat xfactor, GLfloat yfactor);
GLAPI void GLAPIENTRY glPointSize (GLfloat size);
GLAPI void GLAPIENTRY glPolygonMode (GLenum face, GLenum mode);
GLAPI void GLAPIENTRY glPolygonOffset (GLfloat factor, GLfloat units);
GLAPI void GLAPIENTRY glPolygonStipple (const GLubyte *mask);
GLAPI void GLAPIENTRY glPopAttrib (void);
GLAPI void GLAPIENTRY glPopClientAttrib (void);
GLAPI void GLAPIENTRY glPopMatrix (void);
GLAPI void GLAPIENTRY glPopName (void);
GLAPI void GLAPIENTRY glPrioritizeTextures (GLsizei n, const GLuint *textures, const GLclampf *priorities);
GLAPI void GLAPIENTRY glPushAttrib (GLbitfield mask);
GLAPI void GLAPIENTRY glPushClientAttrib (GLbitfield mask);
GLAPI void GLAPIENTRY glPushMatrix (void);
GLAPI void GLAPIENTRY glPushName (GLuint name);
GLAPI void GLAPIENTRY glRasterPos2d (GLdouble x, GLdouble y);
GLAPI void GLAPIENTRY glRasterPos2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos2f (GLfloat x, GLfloat y);
GLAPI void GLAPIENTRY glRasterPos2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos2i (GLint x, GLint y);
GLAPI void GLAPIENTRY glRasterPos2iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos2s (GLshort x, GLshort y);
GLAPI void GLAPIENTRY glRasterPos2sv (const GLshort *v);
GLAPI void GLAPIENTRY glRasterPos3d (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glRasterPos3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos3f (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glRasterPos3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos3i (GLint x, GLint y, GLint z);
GLAPI void GLAPIENTRY glRasterPos3iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos3s (GLshort x, GLshort y, GLshort z);
GLAPI void GLAPIENTRY glRasterPos3sv (const GLshort *v);
GLAPI void GLAPIENTRY glRasterPos4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
GLAPI void GLAPIENTRY glRasterPos4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
GLAPI void GLAPIENTRY glRasterPos4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos4i (GLint x, GLint y, GLint z, GLint w);
GLAPI void GLAPIENTRY glRasterPos4iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos4s (GLshort x, GLshort y, GLshort z, GLshort w);
GLAPI void GLAPIENTRY glRasterPos4sv (const GLshort *v);
GLAPI void GLAPIENTRY glReadBuffer (GLenum mode);
GLAPI void GLAPIENTRY glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);
GLAPI void GLAPIENTRY glRectd (GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2);
GLAPI void GLAPIENTRY glRectdv (const GLdouble *v1, const GLdouble *v2);
GLAPI void GLAPIENTRY glRectf (GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2);
GLAPI void GLAPIENTRY glRectfv (const GLfloat *v1, const GLfloat *v2);
GLAPI void GLAPIENTRY glRecti (GLint x1, GLint y1, GLint x2, GLint y2);
GLAPI void GLAPIENTRY glRectiv (const GLint *v1, const GLint *v2);
GLAPI void GLAPIENTRY glRects (GLshort x1, GLshort y1, GLshort x2, GLshort y2);
GLAPI void GLAPIENTRY glRectsv (const GLshort *v1, const GLshort *v2);
GLAPI GLint GLAPIENTRY glRenderMode (GLenum mode);
GLAPI void GLAPIENTRY glRotated (GLdouble angle, GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glRotatef (GLfloat angle, GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glScaled (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glScalef (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glScissor (GLint x, GLint y, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glSelectBuffer (GLsizei size, GLuint *buffer);
GLAPI void GLAPIENTRY glShadeModel (GLenum mode);
GLAPI void GLAPIENTRY glStencilFunc (GLenum func, GLint ref, GLuint mask);
GLAPI void GLAPIENTRY glStencilMask (GLuint mask);
GLAPI void GLAPIENTRY glStencilOp (GLenum fail, GLenum zfail, GLenum zpass);
GLAPI void GLAPIENTRY glTexCoord1d (GLdouble s);
GLAPI void GLAPIENTRY glTexCoord1dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord1f (GLfloat s);
GLAPI void GLAPIENTRY glTexCoord1fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord1i (GLint s);
GLAPI void GLAPIENTRY glTexCoord1iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord1s (GLshort s);
GLAPI void GLAPIENTRY glTexCoord1sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord2d (GLdouble s, GLdouble t);
GLAPI void GLAPIENTRY glTexCoord2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord2f (GLfloat s, GLfloat t);
GLAPI void GLAPIENTRY glTexCoord2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord2i (GLint s, GLint t);
GLAPI void GLAPIENTRY glTexCoord2iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord2s (GLshort s, GLshort t);
GLAPI void GLAPIENTRY glTexCoord2sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord3d (GLdouble s, GLdouble t, GLdouble r);
GLAPI void GLAPIENTRY glTexCoord3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord3f (GLfloat s, GLfloat t, GLfloat r);
GLAPI void GLAPIENTRY glTexCoord3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord3i (GLint s, GLint t, GLint r);
GLAPI void GLAPIENTRY glTexCoord3iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord3s (GLshort s, GLshort t, GLshort r);
GLAPI void GLAPIENTRY glTexCoord3sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord4d (GLdouble s, GLdouble t, GLdouble r, GLdouble q);
GLAPI void GLAPIENTRY glTexCoord4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord4f (GLfloat s, GLfloat t, GLfloat r, GLfloat q);
GLAPI void GLAPIENTRY glTexCoord4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord4i (GLint s, GLint t, GLint r, GLint q);
GLAPI void GLAPIENTRY glTexCoord4iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord4s (GLshort s, GLshort t, GLshort r, GLshort q);
GLAPI void GLAPIENTRY glTexCoord4sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoordPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glTexEnvf (GLenum target, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexEnvfv (GLenum target, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexEnvi (GLenum target, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexEnviv (GLenum target, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexGend (GLenum coord, GLenum pname, GLdouble param);
GLAPI void GLAPIENTRY glTexGendv (GLenum coord, GLenum pname, const GLdouble *params);
GLAPI void GLAPIENTRY glTexGenf (GLenum coord, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexGenfv (GLenum coord, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexGeni (GLenum coord, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexGeniv (GLenum coord, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexImage1D (GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexImage2D (GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexParameterf (GLenum target, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexParameterfv (GLenum target, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexParameteri (GLenum target, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexParameteriv (GLenum target, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTranslated (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glTranslatef (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glVertex2d (GLdouble x, GLdouble y);
GLAPI void GLAPIENTRY glVertex2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex2f (GLfloat x, GLfloat y);
GLAPI void GLAPIENTRY glVertex2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex2i (GLint x, GLint y);
GLAPI void GLAPIENTRY glVertex2iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex2s (GLshort x, GLshort y);
GLAPI void GLAPIENTRY glVertex2sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertex3d (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glVertex3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex3f (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glVertex3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex3i (GLint x, GLint y, GLint z);
GLAPI void GLAPIENTRY glVertex3iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex3s (GLshort x, GLshort y, GLshort z);
GLAPI void GLAPIENTRY glVertex3sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertex4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
GLAPI void GLAPIENTRY glVertex4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
GLAPI void GLAPIENTRY glVertex4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex4i (GLint x, GLint y, GLint z, GLint w);
GLAPI void GLAPIENTRY glVertex4iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex4s (GLshort x, GLshort y, GLshort z, GLshort w);
GLAPI void GLAPIENTRY glVertex4sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertexPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glViewport (GLint x, GLint y, GLsizei width, GLsizei height);

#ifdef __cplusplus
}
#endif

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.2
 * New stuff:
 *  - Three-Dimensional Texturing (replaces EXT_texture3D);
 *  - BGRA Pixel Formats (replaces EXT_bgra);
 *  - Packed Pixel Formats (replaces EXT_packed_pixels);
 *  - Normal Rescaling (replaces EXT_rescale_normal);
 *  - Separate Specular Color (EXT_separate_specular_color);
 *  - Texture Coordinate Edge Clamping (replaces SGIS_texture_edge_clamp);
 *  - Texture Level of Detail Control (replaces SGIS_texture_lod);
 *  - Vertex Array Draw Element Range (replaces EXT_draw_range_elements);
 *  - Imaging Subset (former ARB_imaging):
 *    - Color Tables (replaces EXT_color_table);
 *    - Convolution (replaces EXT_convolution and HP_convolution_border_modes);
 *    - Color Matrix (replaces SGI_color_matrix)
 *    - Pixel Pipeline Statistics (replaces EXT_histogram);
 *    - Constant Blend Color (replaces EXT_blend_color)
 *    - New Blending Equations (replaces EXT_blend_minmax and EXT_blend_subtract).
 */

#define GL_UNSIGNED_BYTE_3_3_2            0x8032
#define GL_UNSIGNED_SHORT_4_4_4_4         0x8033
#define GL_UNSIGNED_SHORT_5_5_5_1         0x8034
#define GL_UNSIGNED_INT_8_8_8_8           0x8035
#define GL_UNSIGNED_INT_10_10_10_2        0x8036
#define GL_RESCALE_NORMAL                 0x803A
#define GL_TEXTURE_BINDING_3D             0x806A
#define GL_PACK_SKIP_IMAGES               0x806B
#define GL_PACK_IMAGE_HEIGHT              0x806C
#define GL_UNPACK_SKIP_IMAGES             0x806D
#define GL_UNPACK_IMAGE_HEIGHT            0x806E
#define GL_TEXTURE_3D                     0x806F
#define GL_PROXY_TEXTURE_3D               0x8070
#define GL_TEXTURE_DEPTH                  0x8071
#define GL_TEXTURE_WRAP_R                 0x8072
#define GL_MAX_3D_TEXTURE_SIZE            0x8073
#define GL_UNSIGNED_BYTE_2_3_3_REV        0x8362
#define GL_UNSIGNED_SHORT_5_6_5           0x8363
#define GL_UNSIGNED_SHORT_5_6_5_REV       0x8364
#define GL_UNSIGNED_SHORT_4_4_4_4_REV     0x8365
#define GL_UNSIGNED_SHORT_1_5_5_5_REV     0x8366
#define GL_UNSIGNED_INT_8_8_8_8_REV       0x8367
#define GL_UNSIGNED_INT_2_10_10_10_REV    0x8368
#define GL_BGR                            0x80E0
#define GL_BGRA                           0x80E1
#define GL_MAX_ELEMENTS_VERTICES          0x80E8
#define GL_MAX_ELEMENTS_INDICES           0x80E9
#define GL_CLAMP_TO_EDGE                  0x812F
#define GL_TEXTURE_MIN_LOD                0x813A
#define GL_TEXTURE_MAX_LOD                0x813B
#define GL_TEXTURE_BASE_LEVEL             0x813C
#define GL_TEXTURE_MAX_LEVEL              0x813D
#define GL_LIGHT_MODEL_COLOR_CONTROL      0x81F8
#define GL_SINGLE_COLOR                   0x81F9
#define GL_SEPARATE_SPECULAR_COLOR        0x81FA
#define GL_SMOOTH_POINT_SIZE_RANGE        0x0B12
#define GL_SMOOTH_POINT_SIZE_GRANULARITY  0x0B13
#define GL_SMOOTH_LINE_WIDTH_RANGE        0x0B22
#define GL_SMOOTH_LINE_WIDTH_GRANULARITY  0x0B23
#define GL_ALIASED_POINT_SIZE_RANGE       0x846D
#define GL_ALIASED_LINE_WIDTH_RANGE       0x846E

VGLEXTAPI void (VGLEXTAPIENTRY *glBlendColor) (GLclampf, GLclampf, GLclampf, GLclampf);
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendEquation) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawRangeElements) (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTable) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTableParameterfv) (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTableParameteriv) (GLenum, GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyColorTable) (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTable) (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTableParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTableParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorSubTable) (GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyColorSubTable) (GLenum, GLsizei, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionFilter1D) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameterf) (GLenum, GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameterfv) (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameteri) (GLenum, GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameteriv) (GLenum, GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyConvolutionFilter1D) (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyConvolutionFilter2D) (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionFilter) (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetSeparableFilter) (GLenum, GLenum, GLenum, GLvoid *, GLvoid *, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSeparableFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogram) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogramParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogramParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmax) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmaxParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmaxParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glHistogram) (GLenum, GLsizei, GLenum, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glMinmax) (GLenum, GLenum, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glResetHistogram) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glResetMinmax) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexImage3D) (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei);

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.3
 * New stuff:
 *  - Compressed Textures (replaces ARB_texture_compression)
 *  - Cube Map Textures (replaces ARB_texture_cube_map)
 *  - Multisample (replaces ARB_multisample)
 *  - Multitexture (replaces GL 1.2.1 ARB_multitexture)
 *  - Texture Add Environment Mode (replaces ARB_texture_env_add)
 *  - Texture Combine Environment Mode (replaces ARB_texture_env_combine)
 *  - Texture Dot3 Environment Mode (replaces ARB_texture_env_dot3)
 *  - Texture Border Clamp (replaces ARB_texture_border_clamp)
 *  - Transpose Matrix (replaces ARB_transpose_matrix)
 */

#define GL_TEXTURE0                       0x84C0
#define GL_TEXTURE1                       0x84C1
#define GL_TEXTURE2                       0x84C2
#define GL_TEXTURE3                       0x84C3
#define GL_TEXTURE4                       0x84C4
#define GL_TEXTURE5                       0x84C5
#define GL_TEXTURE6                       0x84C6
#define GL_TEXTURE7                       0x84C7
#define GL_TEXTURE8                       0x84C8
#define GL_TEXTURE9                       0x84C9
#define GL_TEXTURE10                      0x84CA
#define GL_TEXTURE11                      0x84CB
#define GL_TEXTURE12                      0x84CC
#define GL_TEXTURE13                      0x84CD
#define GL_TEXTURE14                      0x84CE
#define GL_TEXTURE15                      0x84CF
#define GL_TEXTURE16                      0x84D0
#define GL_TEXTURE17                      0x84D1
#define GL_TEXTURE18                      0x84D2
#define GL_TEXTURE19                      0x84D3
#define GL_TEXTURE20                      0x84D4
#define GL_TEXTURE21                      0x84D5
#define GL_TEXTURE22                      0x84D6
#define GL_TEXTURE23                      0x84D7
#define GL_TEXTURE24                      0x84D8
#define GL_TEXTURE25                      0x84D9
#define GL_TEXTURE26                      0x84DA
#define GL_TEXTURE27                      0x84DB
#define GL_TEXTURE28                      0x84DC
#define GL_TEXTURE29                      0x84DD
#define GL_TEXTURE30                      0x84DE
#define GL_TEXTURE31                      0x84DF
#define GL_ACTIVE_TEXTURE                 0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE          0x84E1
#define GL_MAX_TEXTURE_UNITS              0x84E2
#define GL_TRANSPOSE_MODELVIEW_MATRIX     0x84E3
#define GL_TRANSPOSE_PROJECTION_MATRIX    0x84E4
#define GL_TRANSPOSE_TEXTURE_MATRIX       0x84E5
#define GL_TRANSPOSE_COLOR_MATRIX         0x84E6
#define GL_MULTISAMPLE                    0x809D
#define GL_SAMPLE_ALPHA_TO_COVERAGE       0x809E
#define GL_SAMPLE_ALPHA_TO_ONE            0x809F
#define GL_SAMPLE_COVERAGE                0x80A0
#define GL_SAMPLE_BUFFERS                 0x80A8
#define GL_SAMPLES                        0x80A9
#define GL_SAMPLE_COVERAGE_VALUE          0x80AA
#define GL_SAMPLE_COVERAGE_INVERT         0x80AB
#define GL_MULTISAMPLE_BIT                0x20000000
#define GL_NORMAL_MAP                     0x8511
#define GL_REFLECTION_MAP                 0x8512
#define GL_TEXTURE_CUBE_MAP               0x8513
#define GL_TEXTURE_BINDING_CUBE_MAP       0x8514
#define GL_TEXTURE_CUBE_MAP_POSITIVE_X    0x8515
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_X    0x8516
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Y    0x8517
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y    0x8518
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Z    0x8519
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z    0x851A
#define GL_PROXY_TEXTURE_CUBE_MAP         0x851B
#define GL_MAX_CUBE_MAP_TEXTURE_SIZE      0x851C
#define GL_COMPRESSED_ALPHA               0x84E9
#define GL_COMPRESSED_LUMINANCE           0x84EA
#define GL_COMPRESSED_LUMINANCE_ALPHA     0x84EB
#define GL_COMPRESSED_INTENSITY           0x84EC
#define GL_COMPRESSED_RGB                 0x84ED
#define GL_COMPRESSED_RGBA                0x84EE
#define GL_TEXTURE_COMPRESSION_HINT       0x84EF
#define GL_TEXTURE_COMPRESSED_IMAGE_SIZE  0x86A0
#define GL_TEXTURE_COMPRESSED             0x86A1
#define GL_NUM_COMPRESSED_TEXTURE_FORMATS 0x86A2
#define GL_COMPRESSED_TEXTURE_FORMATS     0x86A3
#define GL_CLAMP_TO_BORDER                0x812D
#define GL_COMBINE                        0x8570
#define GL_COMBINE_RGB                    0x8571
#define GL_COMBINE_ALPHA                  0x8572
#define GL_SOURCE0_RGB                    0x8580
#define GL_SOURCE1_RGB                    0x8581
#define GL_SOURCE2_RGB                    0x8582
#define GL_SOURCE0_ALPHA                  0x8588
#define GL_SOURCE1_ALPHA                  0x8589
#define GL_SOURCE2_ALPHA                  0x858A
#define GL_OPERAND0_RGB                   0x8590
#define GL_OPERAND1_RGB                   0x8591
#define GL_OPERAND2_RGB                   0x8592
#define GL_OPERAND0_ALPHA                 0x8598
#define GL_OPERAND1_ALPHA                 0x8599
#define GL_OPERAND2_ALPHA                 0x859A
#define GL_RGB_SCALE                      0x8573
#define GL_ADD_SIGNED                     0x8574
#define GL_INTERPOLATE                    0x8575
#define GL_SUBTRACT                       0x84E7
#define GL_CONSTANT                       0x8576
#define GL_PRIMARY_COLOR                  0x8577
#define GL_PREVIOUS                       0x8578
#define GL_DOT3_RGB                       0x86AE
#define GL_DOT3_RGBA                      0x86AF

VGLEXTAPI void (VGLEXTAPIENTRY *glActiveTexture) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glClientActiveTexture) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1d) (GLenum, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1f) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1i) (GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1s) (GLenum, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2d) (GLenum, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2f) (GLenum, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2i) (GLenum, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2s) (GLenum, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3d) (GLenum, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3f) (GLenum, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3i) (GLenum, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3s) (GLenum, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4d) (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4f) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4i) (GLenum, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4s) (GLenum, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixf) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixd) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixf) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixd) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSampleCoverage) (GLclampf, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage3D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage2D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage1D) (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage2D) (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage1D) (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetCompressedTexImage) (GLenum, GLint, GLvoid *);

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.4
 * New stuff:
 *  - Vertex programs (replaces ARB_vertex_program)
 *  - Fragment programs (replaces ARB_fragment_program and ARB_fragment_program_shadow)
 *  - Automatic Mipmap Generation (replaces SGIS_generate_mipmap)
 *  - Blend Squaring (replaces NV_blend_square)
 *  - Changes to the Imaging Subset (replaces EXT_blend_color)
 *  - Depth Textures and Shadows (replaces ARB_depth_texture and ARB_shadow)
 *  - Fog Coordinate (replaces EXT_fog_coord)
 *  - Multiple Draw Arrays (replaces EXT_multi_draw_arrays)
 *  - Point Parameters (replaces ARB_point_parameters)
 *  - Secondary Color (replaces EXT_secondary_color)
 *  - Separate Blend Functions (replaces EXT_blend_func_separate)
 *  - Stencil Wrap (replaces EXT_stencil_wrap)
 *  - Texture Crossbar Environment Mode (replaces ARB_texture_env_crossbar)
 *  - Texture LOD Bias (replaces EXT_texture_lod_bias)
 *  - Texture Mirrored Repeat (replaces ARB_texture_mirrored_repeat)
 *  - Window Raster Position (replaces ARB_window_pos)
 */

#define GL_BLEND_DST_RGB                  0x80C8
#define GL_BLEND_SRC_RGB                  0x80C9
#define GL_BLEND_DST_ALPHA                0x80CA
#define GL_BLEND_SRC_ALPHA                0x80CB
#define GL_POINT_SIZE_MIN                 0x8126
#define GL_POINT_SIZE_MAX                 0x8127
#define GL_POINT_FADE_THRESHOLD_SIZE      0x8128
#define GL_POINT_DISTANCE_ATTENUATION     0x8129
#define GL_GENERATE_MIPMAP                0x8191
#define GL_GENERATE_MIPMAP_HINT           0x8192
#define GL_DEPTH_COMPONENT16              0x81A5
#define GL_DEPTH_COMPONENT24              0x81A6
#define GL_DEPTH_COMPONENT32              0x81A7
#define GL_MIRRORED_REPEAT                0x8370
#define GL_FOG_COORDINATE_SOURCE          0x8450
#define GL_FOG_COORDINATE                 0x8451
#define GL_FRAGMENT_DEPTH                 0x8452
#define GL_CURRENT_FOG_COORDINATE         0x8453
#define GL_FOG_COORDINATE_ARRAY_TYPE      0x8454
#define GL_FOG_COORDINATE_ARRAY_STRIDE    0x8455
#define GL_FOG_COORDINATE_ARRAY_POINTER   0x8456
#define GL_FOG_COORDINATE_ARRAY           0x8457
#define GL_COLOR_SUM                      0x8458
#define GL_CURRENT_SECONDARY_COLOR        0x8459
#define GL_SECONDARY_COLOR_ARRAY_SIZE     0x845A
#define GL_SECONDARY_COLOR_ARRAY_TYPE     0x845B
#define GL_SECONDARY_COLOR_ARRAY_STRIDE   0x845C
#define GL_SECONDARY_COLOR_ARRAY_POINTER  0x845D
#define GL_SECONDARY_COLOR_ARRAY          0x845E
#define GL_MAX_TEXTURE_LOD_BIAS           0x84FD
#define GL_TEXTURE_FILTER_CONTROL         0x8500
#define GL_TEXTURE_LOD_BIAS               0x8501
#define GL_INCR_WRAP                      0x8507
#define GL_DECR_WRAP                      0x8508
#define GL_TEXTURE_DEPTH_SIZE             0x884A
#define GL_DEPTH_TEXTURE_MODE             0x884B
#define GL_TEXTURE_COMPARE_MODE           0x884C
#define GL_TEXTURE_COMPARE_FUNC           0x884D
#define GL_COMPARE_R_TO_TEXTURE           0x884E

VGLEXTAPI void (VGLEXTAPIENTRY *glBlendFuncSeparate) (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordf) (GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordfv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordd) (GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoorddv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordPointer) (GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawArrays) (GLenum, GLint *, GLsizei *, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawElements) (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterf) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterfv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameteri) (GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameteriv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3b) (GLbyte, GLbyte, GLbyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3bv) (const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3d) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3f) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3s) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3sv) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ub) (GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ubv) (const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ui) (GLuint, GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3uiv) (const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3us) (GLushort, GLushort, GLushort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3usv) (const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColorPointer) (GLint, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2d) (GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2f) (GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2i) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2s) (GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2sv) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3d) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3f) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3s) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3sv) (const GLshort *);

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.5
 * New stuff:
 *  - Vertex Buffer Objects (replaces ARB_vertex_buffer_object)
 *  - Occlusion Queries (replaces ARB_occlusion_query)
 *  - Shadow Functions (replaces EXT_shadow_funcs)
 */

#define GL_BUFFER_SIZE                    0x8764
#define GL_BUFFER_USAGE                   0x8765
#define GL_QUERY_COUNTER_BITS             0x8864
#define GL_CURRENT_QUERY                  0x8865
#define GL_QUERY_RESULT                   0x8866
#define GL_QUERY_RESULT_AVAILABLE         0x8867
#define GL_ARRAY_BUFFER                   0x8892
#define GL_ELEMENT_ARRAY_BUFFER           0x8893
#define GL_ARRAY_BUFFER_BINDING           0x8894
#define GL_ELEMENT_ARRAY_BUFFER_BINDING   0x8895
#define GL_VERTEX_ARRAY_BUFFER_BINDING    0x8896
#define GL_NORMAL_ARRAY_BUFFER_BINDING    0x8897
#define GL_COLOR_ARRAY_BUFFER_BINDING     0x8898
#define GL_INDEX_ARRAY_BUFFER_BINDING     0x8899
#define GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING 0x889A
#define GL_EDGE_FLAG_ARRAY_BUFFER_BINDING 0x889B
#define GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING 0x889C
#define GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING 0x889D
#define GL_WEIGHT_ARRAY_BUFFER_BINDING    0x889E
#define GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING 0x889F
#define GL_READ_ONLY                      0x88B8
#define GL_WRITE_ONLY                     0x88B9
#define GL_READ_WRITE                     0x88BA
#define GL_BUFFER_ACCESS                  0x88BB
#define GL_BUFFER_MAPPED                  0x88BC
#define GL_BUFFER_MAP_POINTER             0x88BD
#define GL_STREAM_DRAW                    0x88E0
#define GL_STREAM_READ                    0x88E1
#define GL_STREAM_COPY                    0x88E2
#define GL_STATIC_DRAW                    0x88E4
#define GL_STATIC_READ                    0x88E5
#define GL_STATIC_COPY                    0x88E6
#define GL_DYNAMIC_DRAW                   0x88E8
#define GL_DYNAMIC_READ                   0x88E9
#define GL_DYNAMIC_COPY                   0x88EA
#define GL_SAMPLES_PASSED                 0x8914
/*
 * GL 1.5 changes of token names.
 */
#define	FOG_COORD_SRC		FOG_COORDINATE_SOURCE
#define	FOG_COORD			FOG_COORDINATE
#define	CURRENT_FOG_COORD	CURRENT_FOG_COORDINATE
#define	FOG_COORD_ARRAY_TYPE	FOG_COORDINATE_ARRAY TYPE
#define	FOG_COORD_ARRAY_STRIDE	FOG_COORDINATE_ARRAY_STRIDE
#define	FOG_COORD_ARRAY_POINTER	FOG_COORDINATE_ARRAY_POINTER
#define	FOG_COORD_ARRAY		FOG_COORDINATE_ARRAY
#define	FOG_COORD_ARRAY_BUFFER_BINDING	FOG_COORDINATE_ARRAY_BUFFER_BINDING
#define	SRC0_RGB			SOURCE0_RGB
#define	SRC1_RGB			SOURCE1_RGB
#define	SRC2_RGB			SOURCE2_RGB
#define	SRC0_ALPHA			SOURCE0_ALPHA
#define	SRC1_ALPHA			SOURCE1_ALPHA
#define	SRC2_ALPHA			SOURCE2_ALPHA

typedef ptrdiff_t GLintptr;
typedef ptrdiff_t GLsizeiptr;

VGLEXTAPI void (VGLEXTAPIENTRY *glGenQueries) (GLsizei, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteQueries) (GLsizei, const GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsQuery) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBeginQuery) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glEndQuery) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryiv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectuiv) (GLuint, GLenum, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindBuffer) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteBuffers) (GLsizei, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGenBuffers) (GLsizei, GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsBuffer) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferData) (GLenum, GLsizeiptr, const GLvoid *, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferSubData) (GLenum, GLintptr, GLsizeiptr, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferSubData) (GLenum, GLintptr, GLsizeiptr, GLvoid *);
VGLEXTAPI GLvoid* (VGLEXTAPIENTRY *glMapBuffer) (GLenum, GLenum);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glUnmapBuffer) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferPointerv) (GLenum, GLenum, GLvoid* *);

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 2.0
 * New stuff:
 *  - Programmable Shading (replaces GL 1.5 ARB_shader_objects,
 *    ARB_vertex_shader, ARB_fragment_shader, ARB_shading_language_100);
 *  - Multiple Render Targets (replaces ARB_draw_buffers);
 *  - Non-Power-Of-Two Textures (replaces ARB_texture_non_power_of_two);
 *  - Point Sprites (replaces ARB_point_sprite);
 *  - Separate Stencil (replaces EXT_stencil_two_side).
 */

#define GL_BLEND_EQUATION_RGB             GL_BLEND_EQUATION
#define GL_VERTEX_ATTRIB_ARRAY_ENABLED    0x8622
#define GL_VERTEX_ATTRIB_ARRAY_SIZE       0x8623
#define GL_VERTEX_ATTRIB_ARRAY_STRIDE     0x8624
#define GL_VERTEX_ATTRIB_ARRAY_TYPE       0x8625
#define GL_CURRENT_VERTEX_ATTRIB          0x8626
#define GL_VERTEX_PROGRAM_POINT_SIZE      0x8642
#define GL_VERTEX_PROGRAM_TWO_SIDE        0x8643
#define GL_VERTEX_ATTRIB_ARRAY_POINTER    0x8645
#define GL_STENCIL_BACK_FUNC              0x8800
#define GL_STENCIL_BACK_FAIL              0x8801
#define GL_STENCIL_BACK_PASS_DEPTH_FAIL   0x8802
#define GL_STENCIL_BACK_PASS_DEPTH_PASS   0x8803
#define GL_MAX_DRAW_BUFFERS               0x8824
#define GL_DRAW_BUFFER0                   0x8825
#define GL_DRAW_BUFFER1                   0x8826
#define GL_DRAW_BUFFER2                   0x8827
#define GL_DRAW_BUFFER3                   0x8828
#define GL_DRAW_BUFFER4                   0x8829
#define GL_DRAW_BUFFER5                   0x882A
#define GL_DRAW_BUFFER6                   0x882B
#define GL_DRAW_BUFFER7                   0x882C
#define GL_DRAW_BUFFER8                   0x882D
#define GL_DRAW_BUFFER9                   0x882E
#define GL_DRAW_BUFFER10                  0x882F
#define GL_DRAW_BUFFER11                  0x8830
#define GL_DRAW_BUFFER12                  0x8831
#define GL_DRAW_BUFFER13                  0x8832
#define GL_DRAW_BUFFER14                  0x8833
#define GL_DRAW_BUFFER15                  0x8834
#define GL_BLEND_EQUATION_ALPHA           0x883D
#define GL_POINT_SPRITE                   0x8861
#define GL_COORD_REPLACE                  0x8862
#define GL_MAX_VERTEX_ATTRIBS             0x8869
#define GL_VERTEX_ATTRIB_ARRAY_NORMALIZED 0x886A
#define GL_MAX_TEXTURE_COORDS             0x8871
#define GL_MAX_TEXTURE_IMAGE_UNITS        0x8872
#define GL_FRAGMENT_SHADER                0x8B30
#define GL_VERTEX_SHADER                  0x8B31
#define GL_MAX_FRAGMENT_UNIFORM_COMPONENTS 0x8B49
#define GL_MAX_VERTEX_UNIFORM_COMPONENTS  0x8B4A
#define GL_MAX_VARYING_FLOATS             0x8B4B
#define GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS 0x8B4C
#define GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS 0x8B4D
#define GL_SHADER_TYPE                    0x8B4F
#define GL_FLOAT_VEC2                     0x8B50
#define GL_FLOAT_VEC3                     0x8B51
#define GL_FLOAT_VEC4                     0x8B52
#define GL_INT_VEC2                       0x8B53
#define GL_INT_VEC3                       0x8B54
#define GL_INT_VEC4                       0x8B55
#define GL_BOOL                           0x8B56
#define GL_BOOL_VEC2                      0x8B57
#define GL_BOOL_VEC3                      0x8B58
#define GL_BOOL_VEC4                      0x8B59
#define GL_FLOAT_MAT2                     0x8B5A
#define GL_FLOAT_MAT3                     0x8B5B
#define GL_FLOAT_MAT4                     0x8B5C
#define GL_SAMPLER_1D                     0x8B5D
#define GL_SAMPLER_2D                     0x8B5E
#define GL_SAMPLER_3D                     0x8B5F
#define GL_SAMPLER_CUBE                   0x8B60
#define GL_SAMPLER_1D_SHADOW              0x8B61
#define GL_SAMPLER_2D_SHADOW              0x8B62
#define GL_DELETE_STATUS                  0x8B80
#define GL_COMPILE_STATUS                 0x8B81
#define GL_LINK_STATUS                    0x8B82
#define GL_VALIDATE_STATUS                0x8B83
#define GL_INFO_LOG_LENGTH                0x8B84
#define GL_ATTACHED_SHADERS               0x8B85
#define GL_ACTIVE_UNIFORMS                0x8B86
#define GL_ACTIVE_UNIFORM_MAX_LENGTH      0x8B87
#define GL_SHADER_SOURCE_LENGTH           0x8B88
#define GL_ACTIVE_ATTRIBUTES              0x8B89
#define GL_ACTIVE_ATTRIBUTE_MAX_LENGTH    0x8B8A
#define GL_FRAGMENT_SHADER_DERIVATIVE_HINT 0x8B8B
#define GL_SHADING_LANGUAGE_VERSION       0x8B8C
#define GL_CURRENT_PROGRAM                0x8B8D
#define GL_POINT_SPRITE_COORD_ORIGIN      0x8CA0
#define GL_LOWER_LEFT                     0x8CA1
#define GL_UPPER_LEFT                     0x8CA2
#define GL_STENCIL_BACK_REF               0x8CA3
#define GL_STENCIL_BACK_VALUE_MASK        0x8CA4
#define GL_STENCIL_BACK_WRITEMASK         0x8CA5

typedef char GLchar;

VGLEXTAPI void (VGLEXTAPIENTRY *glBlendEquationSeparate) (GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawBuffers) (GLsizei, const GLenum *);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilOpSeparate) (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilFuncSeparate) (GLenum, GLenum, GLint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilMaskSeparate) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glAttachShader) (GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindAttribLocation) (GLuint, GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompileShader) (GLuint);
VGLEXTAPI GLuint (VGLEXTAPIENTRY *glCreateProgram) (void);
VGLEXTAPI GLuint (VGLEXTAPIENTRY *glCreateShader) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteShader) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDetachShader) (GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDisableVertexAttribArray) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnableVertexAttribArray) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveAttrib) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveUniform) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetAttachedShaders) (GLuint, GLsizei, GLsizei *, GLuint *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetAttribLocation) (GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderSource) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetUniformLocation) (GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformfv) (GLuint, GLint, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformiv) (GLuint, GLint, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribdv) (GLuint, GLenum, GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribfv) (GLuint, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribPointerv) (GLuint, GLenum, GLvoid* *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsProgram) (GLuint);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsShader) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glLinkProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glShaderSource) (GLuint, GLsizei, const GLchar* *, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUseProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1f) (GLint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2f) (GLint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3f) (GLint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4f) (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1i) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3i) (GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4i) (GLint, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix2fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix3fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix4fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glValidateProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1d) (GLuint, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1f) (GLuint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1s) (GLuint, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2d) (GLuint, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2f) (GLuint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2s) (GLuint, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3d) (GLuint, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3f) (GLuint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3s) (GLuint, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nbv) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Niv) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nsv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nub) (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nubv) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nuiv) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nusv) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4bv) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4d) (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4f) (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4iv) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4s) (GLuint, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4ubv) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4uiv) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4usv) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttribPointer) (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);


/*
 * =================== Part II: AntiGlut framework library ===============
 */

#define	AG_DEFAULT_WIDTH	0
#define	AG_DEFAULT_HEIGHT	0
#define	AG_WINDOWED			GL_FALSE
#define	AG_FULLSCREEN		GL_TRUE
#define	AG_REQUIRED_1_1		0
#define	AG_REQUIRED_2_0		1

/*
 * UI event types.
 * Stored in ui_event_t -> type.
 */
typedef enum
{
	ui_no_event,

	/*
	 * Key & mouse buttons press or release events.
	 * Message-related information stored in ui_event_t -> key.
	 */
	ui_key_event,
	/*
	 * Pointer motion event.
	 * Message-related information stored in ui_event_t itself,
	 * both ui_event_t -> key fields and most of ui_event_t -> configure
	 * fields are irrelevant.
	 */
	ui_pointer_motion_event,
	/*
	 * Key & mouse buttons press or release events.
	 * Message-related information stored in ui_event_t -> configure.
	 */
	ui_window_focus_in_event,	/* can look at configure.focused */
	ui_window_focus_out_event,	/* can look at configure.focused */
	ui_window_expose_event,		/* look at configure.exposed_XXX */
	ui_window_map_event,		/* can look at configure.visible */
	ui_window_unmap_event,		/* can look at configure.visible */
	ui_window_move_event,		/* look at configure.[x|y] */
	ui_window_size_event,		/* look at configure.[width|height] */
	/*
	 * Message from window manager. Look at configure.wm_event.
	 */
	ui_wm_event,

	ui_max_event
}	ui_event_type_t;

/*
 * Scancodes for "Windows" PC keyboard + 3-button & wheel mouse.
 *
 * Stored in ui_key_event_t -> key -> scan
 */
typedef enum
{
	sc_unknown,

	sc_escape,
	sc_f1,
	sc_f2,
	sc_f3,
	sc_f4,
	sc_f5,
	sc_f6,
	sc_f7,
	sc_f8,
	sc_f9,
	sc_f10,
	sc_f11,
	sc_f12,
	sc_print,
	sc_scroll_lock,
	sc_pause,

	sc_grave,
	sc_1,
	sc_2,
	sc_3,
	sc_4,
	sc_5,
	sc_6,
	sc_7,
	sc_8,
	sc_9,
	sc_0,
	sc_minus,
	sc_equal,
	sc_backslash,
	sc_backspace,

	sc_tab,
	sc_q,
	sc_w,
	sc_e,
	sc_r,
	sc_t,
	sc_y,
	sc_u,
	sc_i,
	sc_o,
	sc_p,
	sc_lbracket,
	sc_rbracket,
	sc_enter,

	sc_caps_lock,
	sc_a,
	sc_s,
	sc_d,
	sc_f,
	sc_g,
	sc_h,
	sc_j,
	sc_k,
	sc_l,
	sc_semicolon,
	sc_apostrophe,

	sc_lshift,
	sc_z,
	sc_x,
	sc_c,
	sc_v,
	sc_b,
	sc_n,
	sc_m,
	sc_comma,
	sc_period,
	sc_slash,
	sc_rshift,

	sc_lcontrol,
	sc_lsuper, /* Left WIN */
	sc_lalt,
	sc_space,
	sc_ralt,
	sc_rsuper, /* Right WIN */
	sc_menu,
	sc_rcontrol,

	sc_insert,
	sc_home,
	sc_prior,
	sc_delete,
	sc_end,
	sc_next,

	sc_up,
	sc_down,
	sc_left,
	sc_right,

	sc_num_lock,
	sc_kp_slash,
	sc_kp_multiply,
	sc_kp_subtract,
	sc_kp_home,
	sc_kp_up,
	sc_kp_prior,
	sc_kp_add,
	sc_kp_left,
	sc_kp_begin,
	sc_kp_right,
	sc_kp_end,
	sc_kp_down,
	sc_kp_next,
	sc_kp_enter,
	sc_kp_insert,
	sc_kp_delete,
	
	sc_lbutton,
	sc_mbutton,
	sc_rbutton,

	sc_wheelup,
	sc_wheeldown,

	sc_maxcode
} ui_scancode_t;

/*
 * Mouse buttons state masks.
 * Stored in ui_event_t -> button
 */
#define	UI_LBUTTON_MASK	(1<<0)
#define	UI_MBUTTON_MASK (1<<1)
#define	UI_RBUTTON_MASK (1<<2)

/*
 * Modifier key state masks.
 * Stored in ui_event_t-> state
 */
#define	UI_NUM_LOCK_MASK	(1<<0)	/* X11: Mod2 */
#define	UI_CAPS_LOCK_MASK	(1<<1)
#define	UI_SCROLL_LOCK_MASK	(1<<2)	/* X11: Mod5 */
#define	UI_SHIFT_MASK		(1<<3)
#define	UI_CONTROL_MASK		(1<<4)
#define	UI_ALT_MASK			(1<<5)	/* X11: Mod1 */
#define	UI_SUPER_MASK		(1<<6)	/* X11: Mod4 */

/*
 * WM requests: only one is known at this time
 */
typedef enum
{
	ui_wm_unknown,

	ui_wm_close,	/* closing request (under Win32 - Alt+F4 pressed) */

	ui_wm_max
}	ui_wm_event_t;

typedef struct ui_key_event_s
{
	GLboolean press;	/* == true if key/button is pressed */
	unsigned scan;	/* scancode: cpecific for LibV, defined above */
	unsigned acode;	/* 8-bit ASCII for current locale, try not to use it */
	unsigned ucode;	/* machine-endian 16-bit unicode */
}	ui_key_event_t, *ui_key_event_p;

typedef struct ui_configure_event_s
{
	GLboolean focused;	/* == true if window has input focus */
	GLboolean visible;	/* == true if window is visible (mapped) */
	/* damaged rectangle for ui_window_expose event */
	unsigned exposed_left, exposed_top, exposed_right, exposed_bottom;
	int x, y;			/* position of window */
	unsigned width, height;	/* size of window */
	/* window manager request */
	unsigned wm_event;
}	ui_configure_event_t, *ui_configure_event_p;

typedef struct ui_event_s
{
	int		type;
	/*
	 * Time since first event.
	 */
	GLfloat	time;
	/*
	 * Mouse pointer position, client space.
	 */
	int		client_x, client_y;
	/*
	 * Mouse pointer position, screen space. Try _not_ to use it
	 * without necessity.
	 */
	int		screen_x, screen_y;
	/*
	 * Mouse pointer deltas since last move.
	 */
	int		client_dx, client_dy;
	unsigned button; /* state of mouse buttons */
	unsigned state; /* state of keyboard modifiers */

	ui_key_event_t key;	/* ui_key_event - specific data */

	ui_configure_event_t configure; /* window - configuration specific data */
}	ui_event_t, *ui_event_p;

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
	/*
	 * Window settings. Must be called before agInit()
	 */
	GLboolean agWindowConfig(char *title,GLuint width,GLuint height,
		GLboolean fullscreen);
	/*
	 * Initialization.
	 */
	GLboolean agInit(GLuint required_version,GLboolean need_depth,
		GLboolean need_stencil,GLboolean need_dst_alpha);
	/*
	 * Buffer swapping.
	 */
	void agSwap(void);
	/*
	 * Termination.
	 */
	void agShutdown(void);
	/*
	 * Enables or disables keyboard autorepeat.
	 */
	void	uiAutoRepeat(GLboolean enable);
	/*
	 * Warps mouse cursor to another position
	 */
	void	uiWarpPointer(unsigned client_x,unsigned client_y);
	/*
	 * If 'enable': saves pointer position, hides pointer, starts
	 * incremental tracking of pointer coords (screen position of pointer
	 * _no more_ corresponds to window position, which can be arbitrary
	 * number between [-MAXLONG;MAXLONG]). If !'enable': restores
	 * pointer position, shows pointer. It is recommended to use deltas
	 * (i.e. msg->client_dx and msg->client_dy) instead of coords.
	 */
	void	uiAutoWarp(GLboolean enable);
	/*
	 * Polls the system (moving the messages from all threads' queues),
	 * returns number of events waiting in queue of window of current
	 * thread; _never blocks_.
	 */
	GLboolean	uiPending(void);
	/*
	 * Removes the oldest event waiting in queue of active GL
	 * window. Polls the system queue and _can block_.
	 */
	GLboolean	uiNextEvent(ui_event_p ev_return);
	/*
	 * Returns pointer to n#th event from head of the queue.
	 * Neither polls system, nor blocks.
	 */
	ui_event_p	uiPeekEvent(unsigned offset_from_first_push);
	/* Same as above, but returns n#th event from tail of the queue. */
	ui_event_p	uiPeekLastEvent(unsigned offset_from_last_push);
	/*
	 * The two following routines behaves similar to previous ones,
	 * but also removes message from queue.
	 */
	GLboolean	uiRemoveEvent(unsigned offset_from_first_push,
		ui_event_p ev_return);
	GLboolean	uiRemoveLastEvent(unsigned offset_from_last_push,
		ui_event_p ev_return);
	/*
	 * Helper routines.
	 */
	void	agWindowTitle(char *title); /* Changes the title of the window */
	GLuint	agTickCount(void); /* Returns ticks in ms (arbitrary start) */
	GLvoid	agSleep(GLuint milliseconds); /* Delays executions */
#ifdef __cplusplus
}
#endif /* __cplusplus */

/*
 * Misc macros
 */
#ifdef	AG_VERBOSE

#define	TRACE(str)			puts(str)
#define	TRACE_1ARG(fmt,arg)	printf(fmt,arg);puts("");

#else	/* !AG_VERBOSE */

#define	TRACE(str)
#define	TRACE_1ARG(fmt,arg)

#endif	/* !AG_VERBOSE */

#endif	/* !ANTIGLUT_H */
