#include "antiglut.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>

/*
 * The only global data
 */
static struct ag_window_config_s
{
	char title[512];
	GLuint width, height;
	GLboolean fullscreen;
}	ag_window_config = {"AntiGlut",
		0, 0, 		/* Desktop dimensions */
		GL_TRUE};	/* Fullscreen */

static void	*global_context = NULL;

#define	alloc_context()	(gl_context_p)memAllocZ(sizeof(gl_context_t))

/*
 * Common config macro(s)
 */
#define	DISCARD_DUPLICATE_MOUSEMOVE

/*
 * Wrappers for old LibV memory management routines.
 * Macro MALLOC_VERBOSE may be set in antiglut.h.
 */

static GLuint num_mem_blocks = 0;

#define	memZero(ptr,size)	memset(ptr,0,size)

static void	*memAllocZ(GLuint size)
{
	void *ptr;
	assert(size);
	ptr = malloc(size);
#ifdef	MALLOC_VERBOSE
    printf("malloc(): %u bytes: %X\n",size,ptr);
#endif	/* MALLOC_VERBOSE */
	if(ptr)
	{
		memset(ptr,0,size);
		num_mem_blocks++;
		return ptr;
	}
	TRACE("Not enought memory.");
    assert(0);
	return NULL;
}

#define	memAlloc	memAllocZ

static void	memFree(void *ptr)
{
	assert(ptr);
#ifdef	MALLOC_VERBOSE
    printf("free(): %X; ",ptr);
#endif	/* MALLOC_VERBOSE */
	if(num_mem_blocks)
	{
		num_mem_blocks--;
		free(ptr);
#ifdef	MALLOC_VERBOSE
	    printf("(%u block(s) left)\n",num_mem_blocks);
#endif	/* MALLOC_VERBOSE */
		return;
	}
	TRACE("Extra free().");
    assert(0);
}

static char *strDup(char *str)
{
	unsigned len;
	char *ptr;
	if(!str)
    	return NULL;
	len = strlen(str)+1;
	ptr = (char*)memAlloc(len+1);
    memcpy(ptr,str,len);
    return ptr;
}

/*
 * Template for deque ([F|L]IFO), used for implementation of message queue.
 * Circular doubly-linked list is used.
 */

#define	DEQUE_STRUCT(prefix,storage_class) \
\
typedef struct prefix##_deque_s \
{ \
	storage_class	store; \
	struct prefix##_deque_s *next, *prev; \
}	prefix##_deque_t, *prefix##_deque_p;

#define	IMPLEMENT_DEQUE_API(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPush(prefix##_deque_p deq,storage_class *store) \
{ \
	prefix##_deque_p node = \
		(prefix##_deque_p)memAllocZ(sizeof(prefix##_deque_t)); \
	memcpy(&node->store,store,sizeof(storage_class)); \
	if(!deq) \
	{ \
		node->next = node; \
		node->prev = node; \
		return node; \
	} \
	/* insert _before_ que */ \
	node->next = deq; \
	node->prev = deq->prev; \
	deq->prev = node; \
	node->prev->next = node; \
	return node; \
} \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPop(prefix##_deque_p deq,storage_class *store_ret) \
{ \
	prefix##_deque_p next, prev; \
	if(!deq) \
		return NULL; \
	if(store_ret) \
		memcpy(store_ret,&deq->store,sizeof(storage_class)); \
	next = deq->next; \
	prev = deq->prev; \
	next->prev = prev; \
	prev->next = next; \
	if(next == deq) \
		next = NULL; \
	memFree(deq); \
	return next; \
} \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPopBack(prefix##_deque_p deq,storage_class *store_ret) \
{ \
	return deq ? prefix##DeqPop(deq->prev,store_ret) : NULL; \
} \
\
static GLboolean	calling_convention \
prefix##DeqPeek(prefix##_deque_p deq, \
	unsigned offset_from_last_push,storage_class **store_ptr_ret) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return GL_FALSE; \
	do \
	{ \
		if(count == offset_from_last_push) \
		{ \
			if(store_ptr_ret) \
				*store_ptr_ret = &ptr->store; \
			return GL_TRUE; \
		} \
		count++; \
		ptr = ptr->next; \
	} while(ptr != deq); \
	return GL_FALSE; \
} \
\
static GLboolean	calling_convention \
prefix##DeqPeekBack(prefix##_deque_p deq, \
	unsigned offset_from_first_push, storage_class **store_ptr_ret) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return GL_FALSE; \
	do \
	{ \
		if(count == offset_from_first_push) \
		{ \
			if(store_ptr_ret) \
				*store_ptr_ret = &ptr->prev->store; \
			return GL_TRUE; \
		} \
		count++; \
		ptr = ptr->prev; \
	} while(ptr != deq); \
	return GL_FALSE; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqRemove(prefix##_deque_p deq, \
	unsigned offset_from_last_push) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return NULL; \
	do \
	{ \
		if(count == offset_from_last_push) \
		{ \
			if(ptr == deq) \
				return prefix##DeqPop(ptr,NULL); \
			prefix##DeqPop(ptr,NULL); \
			return deq; \
		} \
		count++; \
		ptr = ptr->next; \
	} while(ptr != deq); \
	return deq; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqRemoveBack(prefix##_deque_p deq, \
	unsigned offset_from_first_push) \
{ \
	unsigned count = 0; \
	prefix##_deque_p ptr = deq; \
	if(!ptr) \
		return NULL; \
	do \
	{ \
		ptr = ptr->prev; \
		if(count == offset_from_first_push) \
		{ \
			if(ptr == deq) \
				return prefix##DeqPop(ptr,NULL); \
			prefix##DeqPop(ptr,NULL); \
			return deq; \
		} \
		count++; \
	} while(ptr != deq); \
	return deq; \
} \
\
static void	calling_convention \
prefix##DeqFree(prefix##_deque_p deq) \
{ \
	prefix##_deque_p head, next; \
	if(!deq) \
		return; \
	head = deq; \
	do \
	{ \
		next = deq->next; \
		memFree(deq); \
		deq = next; \
	} while(deq != head); \
}

#define	DEQ_PUSH(prefix,deq,store) \
	((deq) = prefix##DeqPush(deq,store))

#define	DEQ_POP(prefix,deq,store_ret) \
	((deq) = prefix##DeqPop(deq,store_ret))

#define	DEQ_POP_BACK(prefix,deq,store_ret) \
	((deq) = prefix##DeqPopBack(deq,store_ret))

#define	DEQ_PEEK(prefix,deq,offset_from_last_push,store_ret) \
	prefix##DeqPeek(deq,offset_from_last_push,store_ret)

#define	DEQ_PEEK_BACK(prefix,deq,offset_from_first_push,store_ret) \
	prefix##DeqPeekBack(deq,offset_from_first_push,store_ret)

#define	DEQ_REMOVE(prefix,deq,offset_from_last_push) \
	((deq) = prefix##DeqRemove(deq,offset_from_last_push))

#define	DEQ_REMOVE_BACK(prefix,deq,offset_from_first_push) \
	((deq) = prefix##DeqRemoveBack(deq,offset_from_first_push))

#define	DEQ_FREE(prefix,deq)	prefix##DeqFree(deq)

/*
 * Event que declaration and implementation
 */

#define	DEFAULT_CALLING_CONVENTION

DEQUE_STRUCT(event,ui_event_t)

IMPLEMENT_DEQUE_API(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#ifdef	__UNIX__

#define	IMPLEMENT_DEQUE_PB(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p	calling_convention \
prefix##DeqPushBack(prefix##_deque_p deq,storage_class *store) \
{ \
	prefix##_deque_p new_head = prefix##DeqPush(deq,store); \
	return deq ? deq : new_head; \
} \

#define	DEQ_PUSH_BACK(prefix,deq,store) \
	((deq) = prefix##DeqPushBack(deq,store))

IMPLEMENT_DEQUE_PB(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#else	/* !__UNIX__ */

#define	IMPLEMENT_DEQUE_SCANS(prefix,calling_convention,storage_class) \
\
static prefix##_deque_p calling_convention \
prefix##DeqScanForward(prefix##_deque_p deq, \
	GLboolean (calling_convention *reject)(storage_class*,void*),void *arg) \
{ \
	prefix##_deque_p head, ptr, tail; \
	if(!deq) \
		return NULL; \
	deq = deq->prev; \
	head = NULL; \
	tail = deq; \
	do \
	{ \
		ptr = deq->prev; \
		if(reject(&deq->store,arg)) \
			memFree(deq); \
		else \
		{ \
			deq->next = head; \
			head = deq; \
		} \
		deq = ptr; \
	} while(deq != tail); \
	if(!head) \
		return NULL; \
	for(tail = head, ptr = head->next;ptr;ptr = ptr->next) \
	{ \
		ptr->prev = tail; \
		tail = ptr; \
	} \
	head->prev = tail; \
	tail->next = head; \
	return head; \
} \
\
static prefix##_deque_p calling_convention \
prefix##DeqScanBackward(prefix##_deque_p deq, \
	GLboolean (calling_convention *reject)(storage_class*,void*),void *arg) \
{ \
	prefix##_deque_p head, ptr, tail; \
	if(!deq) \
		return NULL; \
	head = deq; \
	tail = NULL; \
	do \
	{ \
		ptr = deq->next; \
		if(reject(&deq->store,arg)) \
			memFree(deq); \
		else \
		{ \
			deq->prev = tail; \
			tail = deq; \
		} \
		deq = ptr; \
	} while(deq != head); \
	if(!tail) \
		return NULL; \
	for(head = tail, ptr = tail->prev;ptr;ptr = ptr->prev) \
	{ \
		ptr->next = head; \
		head = ptr; \
	} \
	head->prev = tail; \
	tail->next = head; \
	return head; \
}

#define	DEQ_SCAN_FORWARD(prefix,deq,reject,arg) \
	((deq) = prefix##DeqScanForward(deq,reject,arg))

#define	DEQ_SCAN_BACKWARD(prefix,deq,reject,arg) \
	((deq) = prefix##DeqScanBackward(deq,reject,arg))

IMPLEMENT_DEQUE_SCANS(event,DEFAULT_CALLING_CONVENTION,ui_event_t)

#endif	/* !__UNIX__ */

/*
 * Some WM/GL data is common for both platforms...
 */

typedef struct vidmode_s
{
	GLuint	width;
	GLuint	height;
	GLuint	bpp;		/* set correctly but not used under X11 */
	GLuint	refresh;	/* set correctly but not used under X11 */
#ifndef	__WIN32__
	GLuint	modeline;	/* not used under Win32 */
#endif	/* !__WIN32__ */
}	vidmode_t, *vidmode_p;

#define	MAX_WINDOW_TITLE	128

#define	ARCH_COMMON_WINDOW_FIELDS \
	char	title[MAX_WINDOW_TITLE]; /* caption */ \
	GLuint	x, y; /* [left,top] */ \
	GLuint	width, height; \
	GLuint	root_width, root_height; \
	/* X11 - time of 1st input msg; Win32 - time of wnd creation */ \
	GLuint start_time; \
	GLboolean focused; /* affected by winsystem | archWinFocus() */ \
	GLboolean visible; /* affected by winsystem | archWinShow() */ \
	GLuint pointer_hide_count; /* affected by archWinShowPointer() */ \
	GLuint ms_grab_count; /* affected by archWinGrabMouse() */ \
	GLuint kb_grab_count; /* affected by archWinGrabKeyboard() */ \
	GLint  current_vidmode; /* affected by archWinSetVidMode() */ \
	GLuint saved_pointer_hide_count; /* affected by ui_win_loose_control() */ \
	GLuint saved_ms_grab_count; /* affected by ui_win_loose_control() */ \
	GLuint saved_kb_grab_count; /* affected by ui_win_loose_control() */ \
	GLint  saved_vidmode; /* affected by ui_win_loose_control() */ \
	/* last ms coords, client space, as returned by windowing system */ \
	GLuint last_mx, last_my; \
	/* last ms coords, client space, as tracked by library */ \
	GLuint last_vmx, last_vmy; \
	GLuint last_key_state; /* keyboard modifiers of last input msg */ \
	GLuint last_button_state; /* mouse btns of last mouse msg */ \
	GLboolean auto_repeat_disabled; /* !filter autorepeated keys */ \
	/* mouse pointer hidden, 'infinite' mouse mode */ \
	GLboolean auto_warp_enabled; \
	GLuint auto_warp_saved_x, auto_warp_saved_y; \
	GLboolean zap_next_mousemove; /* eat next MouseMove msg */ \
	GLubyte keyboard_state[sc_maxcode]; /* 0 if sc_XXX key is released */ \
	event_deque_p msg_que; /* message deque */

#define	INIT_COMMON_WINDOW_FIELDS(winobj) \
	(winobj)->title[0] = 0; \
	(winobj)->x = (winobj)->y = 0; \
	(winobj)->width = (winobj)->height = 256; \
	(winobj)->root_width = (winobj)->root_height = 0; \
	(winobj)->start_time = 0; \
	(winobj)->visible = GL_FALSE; \
	(winobj)->focused = GL_FALSE; \
	(winobj)->pointer_hide_count = 0; \
	(winobj)->ms_grab_count = 0; \
	(winobj)->kb_grab_count = 0; \
	(winobj)->current_vidmode = -1; \
	(winobj)->saved_pointer_hide_count = 0; \
	(winobj)->saved_ms_grab_count = 0; \
	(winobj)->saved_kb_grab_count = 0; \
	(winobj)->saved_vidmode = -1; \
	(winobj)->auto_warp_enabled = GL_FALSE; \
	(winobj)->last_mx = (winobj)->last_my = 0; \
	(winobj)->last_key_state = (winobj)->last_button_state = 0; \
	(winobj)->auto_repeat_disabled = GL_FALSE; \
	(winobj)->zap_next_mousemove = GL_FALSE; \
	memZero((winobj)->keyboard_state,sc_maxcode); \
	(winobj)->msg_que = NULL;

#define	CLEAR_COMMON_WINDOW_FIELDS(winobj) \
	eventDeqFree((winobj)->msg_que); \
	(winobj)->msg_que = NULL;

#define ARCH_COMMON_GL_CONTEXT_FIELDS \
	arch_window_p wnd; \
	char *vendor, *renderer, *version, *extensions;

#define ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx) \
	if((ctx)->vendor) { \
		memFree((ctx)->vendor); \
		(ctx)->vendor = NULL; \
	} \
	if((ctx)->renderer) { \
		memFree((ctx)->renderer); \
		(ctx)->renderer = NULL; \
	} \
	if((ctx)->version) { \
		memFree((ctx)->version); \
		(ctx)->version = NULL; \
	} \
	if((ctx)->extensions) { \
		memFree((ctx)->extensions); \
		(ctx)->extensions = NULL; \
	} \
	if((ctx)->wnd) { \
		archWinClose((ctx)->wnd); \
		(ctx)->wnd = NULL; \
	}

#define	IF_NO_GLCONTEXT(ctx) \
	(ctx) = (gl_context_p)(global_context); \
	if((ctx) == NULL)

/*
 * ... and other are not.
 */

#ifndef	__UNIX__

#include <windows.h>

#define	MAX_CLASS_NAME	32

typedef struct arch_window_s
{
	ARCH_COMMON_WINDOW_FIELDS
	char		class_name[MAX_CLASS_NAME];
	WNDCLASS	wc;
	HWND		wnd;
	HDC			dc;
	DWORD		delta_w, delta_h;
	int			wheel_delta;
	BYTE		oem_kbstate[256];
	UINT		auto_capture_button;
	GLuint		num_vidmodes; /* fixme: common */
	vidmode_p	vidmodes; /* fixme: common */
}	arch_window_t, *arch_window_p;

typedef struct gl_context_s
{
	ARCH_COMMON_GL_CONTEXT_FIELDS
	HGLRC wgl_context;
}	gl_context_t, *gl_context_p;

#else	/* __UNIX__ */

/*
 * X11 configuration macro(s).
 */
#define	ARCH_HAVE_XF86_VIDMODE

#include <locale.h>
#include <dlfcn.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <GL/glx.h>
#include <sys/time.h>

#ifdef	ARCH_HAVE_XF86_VIDMODE
#include <X11/extensions/xf86vmode.h>
#endif	/* ARCH_HAVE_XF86_VIDMODE */

typedef struct x11_display_connection_s
{
	Window root;
	GLuint scrnum;
	GLuint screen_width, screen_height;
	char *lang;
	Display *dpy;
	XIM im;
	Atom wm_protocols_atom;
	Atom wm_delete_window_atom;
	Cursor null_cursor;
	GLboolean grab_keyboard;
	GLuint connection_fd;
#ifdef	ARCH_HAVE_XF86_VIDMODE
	GLuint num_vidmodes;
	GLboolean vidmode_active;
	XF86VidModeModeInfo **xf86_vidmodes;
	vidmode_p vidmodes;
#endif	/* ARCH_HAVE_XF86_VIDMODE */
}	x11_display_connection_t, *x11_display_connection_p;

typedef struct arch_window_s
{
	ARCH_COMMON_WINDOW_FIELDS
	x11_display_connection_p cn;
	Window		parent, wnd;
	XIC			ic;
	GLuint	event_mask;
	GLuint	keycode_to_scan[256];
	Colormap	cmap;
}	arch_window_t, *arch_window_p;

typedef struct gl_context_s
{
	ARCH_COMMON_GL_CONTEXT_FIELDS
	XVisualInfo *visual_info;
	GLXContext glx_context;
}	gl_context_t, *gl_context_p;

static x11_display_connection_p x11_open_connection();
static void x11_close_connection(x11_display_connection_p cn);

#endif	/* !__WIN32__ */

/*
 * archXXX routines
 */
static GLboolean
archWinSetVidMode(arch_window_p aw,GLint new_mode);

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab);

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y);

#ifndef	__UNIX__
static void
archWinFocus(arch_window_p aw);
#endif	/* !__UNIX__ */

static void
archWinClose(arch_window_p aw);

/*
 * ui_default_wnd_proc()
 * Returns 'GL_FALSE' if message must _not_ be stored in queue.
 */
static GLboolean
ui_default_wnd_proc(arch_window_p aw,ui_event_p ev)
{
	GLuint x, y, cx, cy, dx, dy;
	switch(ev->type)
	{
		case ui_pointer_motion_event:
			if(aw->zap_next_mousemove)
			{
				aw->zap_next_mousemove = GL_FALSE;
				return GL_FALSE;
			}
			x = ev->client_x;
			y = ev->client_y;
#ifdef	DISCARD_DUPLICATE_MOUSEMOVE
			if(aw->last_mx == x && aw->last_my == y)
				return GL_FALSE;
#endif	/* DISCARD_DUPLICATE_MOUSEMOVE */
			dx = x - aw->last_mx;
			dy = y - aw->last_my;
			if(aw->auto_warp_enabled)
            {
            	cx = aw->width >> 1;
            	cy = aw->height >> 1;
				aw->last_mx = cx;
				aw->last_my = cy;
				aw->last_vmx += dx;
				aw->last_vmy += dy;
				ev->client_x = aw->last_vmx;
				ev->client_y = aw->last_vmy;
				aw->zap_next_mousemove = GL_TRUE;
				archWinWarpPointer(aw, cx, cy);
				ev->client_dx = dx;
				ev->client_dy = dy;
            }
			else
			{
				ev->client_dx = x - aw->last_mx;
				ev->client_dy = y - aw->last_my;
				aw->last_vmx = aw->last_mx = x;
				aw->last_vmy = aw->last_my = y;
			}
			return GL_TRUE;
		default:
			break;
	}
	if(aw->auto_warp_enabled)
    {
		ev->client_x = aw->last_vmx;
		ev->client_y = aw->last_vmy;
	}
	ev->client_dx = 0;
	ev->client_dy = 0;
	return GL_TRUE;
}

#ifdef	__UNIX__
static void
archWinGrabKeyboard(arch_window_p aw,GLboolean grab);
#endif	/* __UNIX__ */

static void
ui_win_loose_control(arch_window_p aw,GLboolean hiding)
{
	if(hiding)
	{
		/* check if we need to uninstall grabs / restore vidmode */
		if(aw->ms_grab_count > 0)
		{
			aw->saved_ms_grab_count = aw->ms_grab_count;
			aw->ms_grab_count = 1;
			archWinGrabMouse(aw, GL_FALSE);
		}
		if(aw->kb_grab_count > 0)
		{
			aw->saved_kb_grab_count = aw->kb_grab_count;
			aw->kb_grab_count = 1;
#ifdef	__UNIX__
			archWinGrabKeyboard(aw, GL_FALSE);
#endif	/* __UNIX__ */
		}
		if(aw->current_vidmode != -1)
		{
			aw->saved_vidmode = aw->current_vidmode;
			archWinSetVidMode(aw, -1);
		}
		/*
		 * On exit:
		 *  ms_grab_count == kb_grab_count == 0;
		 *  current_vidmode == 0;
		 */
	}
	else
	{
		/* check if we need to reinstall grabs / set vidmode */
		if(aw->saved_ms_grab_count > 0)
		{
			aw->ms_grab_count = 0;
			archWinGrabMouse(aw, GL_TRUE);
			aw->ms_grab_count = aw->saved_ms_grab_count;
			aw->saved_ms_grab_count = 0;
		}
		if(aw->saved_kb_grab_count > 0)
		{
			aw->kb_grab_count = 0;
#ifdef	__UNIX__
			archWinGrabKeyboard(aw, GL_TRUE);
#endif	/* __UNIX__ */
			aw->kb_grab_count = aw->saved_kb_grab_count;
			aw->saved_kb_grab_count = 0;
		}
		if(aw->saved_vidmode != -1)
		{
			archWinSetVidMode(aw, aw->saved_vidmode);
			aw->saved_vidmode = -1;
		}
		/*
		 * On exit:
		 *  saved_ms_grab_count == saved_kb_grab_count == 0;
		 *  saved_vidmode == -1;
		 */
	}
}

static void
ui_make_default_key_event(ui_key_event_p ev)
{
	ev->press = GL_FALSE;
	ev->scan = sc_unknown;
	ev->acode = 0;
	ev->ucode = 0;
}

static void
ui_set_configure_from_aw(arch_window_p aw,ui_configure_event_p ev)
{
	ev->focused = aw->focused;
	ev->visible = aw->visible;
	if(aw->visible)
	{
		ev->x = aw->x;
		ev->y = aw->y;
		ev->width = aw->width;
		ev->height = aw->height;
	}
	else
	{
		ev->x = ev->y = 0;
		ev->width = ev->height = 0;
	}
	ev->exposed_left = ev->exposed_top = 0;
	ev->exposed_right = ev->exposed_bottom = 0;
	ev->wm_event = ui_wm_unknown;
}

static GLfloat
ui_make_message_time(GLuint *start,GLuint current)
{
	GLuint delta1, delta2;
	if(!*start)
		*start = current;
	if(*start <= current)
	{
		delta1 = current - *start;
		delta2 = 0;
	}
	else
	{
		/* overflow ! */
		GLuint last;
#ifdef	_ARCH_64_
		last = 0xFFFFFFFFFFFFFFFF;
#else
		last = 0xFFFFFFFF;
#endif
		delta1 = last - *start;
		delta2 = current;
		/* FIXME: this is wrong */
		*start = current;
	}
	return (GLfloat)0.001 * ((GLfloat)delta1 + (GLfloat)delta2);
}

/*
 * Windows - specific stuff
 */

#ifndef __UNIX__

static char *LIBV_HANDLE_PROP = "LIBV_HANDLE";

/*
 * Translates wParam and lParam into LibV scan code.
 */
static GLuint
win32_make_scan_code(WPARAM wParam,LPARAM lParam)
{
	GLubyte scan = (GLubyte)((lParam >> 16) & 0xFF);
	GLubyte ext = (GLubyte)((lParam & (1 << 24)) != 0);
	switch(wParam)
	{
		case VK_ESCAPE:	return sc_escape;

		case VK_F1: return sc_f1;
		case VK_F2: return sc_f2;
		case VK_F3: return sc_f3;
		case VK_F4: return sc_f4;
		case VK_F5: return sc_f5;
		case VK_F6: return sc_f6;
		case VK_F7: return sc_f7;
		case VK_F8: return sc_f8;
		case VK_F9: return sc_f9;
		case VK_F10: return sc_f10;
		case VK_F11: return sc_f11;
		case VK_F12: return sc_f12;
		case VK_SCROLL: return sc_scroll_lock;

		case 192: return sc_grave;
		case '1': return sc_1;
		case '2': return sc_2;
		case '3': return sc_3;
		case '4': return sc_4;
		case '5': return sc_5;
		case '6': return sc_6;
		case '7': return sc_7;
		case '8': return sc_8;
		case '9': return sc_9;
		case '0': return sc_0;
		case 189: return sc_minus;
		case 187: return sc_equal;
		case 220: return sc_backslash;
		case VK_BACK: return sc_backspace;

		case VK_TAB: return sc_tab;
		case 'Q': return sc_q;
		case 'W': return sc_w;
		case 'E': return sc_e;
		case 'R': return sc_r;
		case 'T': return sc_t;
		case 'Y': return sc_y;
		case 'U': return sc_u;
		case 'I': return sc_i;
		case 'O': return sc_o;
		case 'P': return sc_p;
		case 219: return sc_lbracket;
		case 221: return sc_rbracket;
		case VK_RETURN:	return ext ? sc_kp_enter : sc_enter;

		case VK_CAPITAL: return sc_caps_lock;
		case 'A': return sc_a;
		case 'S': return sc_s;
		case 'D': return sc_d;
		case 'F': return sc_f;
		case 'G': return sc_g;
		case 'H': return sc_h;
		case 'J': return sc_j;
		case 'K': return sc_k;
		case 'L': return sc_l;
		case 186: return sc_semicolon;
		case 222: return sc_apostrophe;

		case 16: return (scan == 42) ? sc_lshift : sc_rshift;
		case 'Z': return sc_z;
		case 'X': return sc_x;
		case 'C': return sc_c;
		case 'V': return sc_v;
		case 'B': return sc_b;
		case 'N': return sc_n;
		case 'M': return sc_m;
		case 188: return sc_comma;
		case 190: return sc_period;
		case 191: return sc_slash;

		case 17: return ext ? sc_rcontrol : sc_lcontrol;
		case VK_LWIN: return sc_lsuper;
		case 18: return ext ? sc_ralt : sc_lalt;
		case VK_SPACE: return sc_space;
		case VK_RWIN: return sc_rsuper;
		case VK_APPS: return sc_menu;
		case VK_NUMLOCK: return sc_num_lock;
		case VK_DIVIDE: return sc_kp_slash;
		case 106: return sc_kp_multiply;
		case 109: return sc_kp_subtract;
		case 107: return sc_kp_add;
		/* case VK_CLEAR: return sc_kp_begin; */
	}
	switch(scan)
	{
		case 71: return ext ? sc_home : sc_kp_home;
		case 72: return ext ? sc_up : sc_kp_up;
		case 73: return ext ? sc_prior : sc_kp_prior;
		case 75: return ext ? sc_left : sc_kp_left;
		case 76: return sc_kp_begin;
		case 77: return ext ? sc_right : sc_kp_right;
		case 79: return ext ? sc_end : sc_kp_end;
		case 80: return ext ? sc_down : sc_kp_down;
		case 81: return ext ? sc_next : sc_kp_next;
		case 82: return ext ? sc_insert : sc_kp_insert;
		case 83: return ext ? sc_delete : sc_kp_delete;
		case 69: return sc_pause;
		case 55: return sc_print;
	}
	return sc_unknown;
}

static void
win32_get_key_modifiers(arch_window_p aw,ui_event_p ev)
{
	BYTE *keys = aw->oem_kbstate;
	GetKeyboardState(keys);
	ev->state = ev->button = 0;
	if(keys[VK_NUMLOCK] & 1)
		ev->state |= UI_NUM_LOCK_MASK;
	if(keys[VK_CAPITAL] & 1)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(keys[VK_SCROLL] & 1)
		ev->state |= UI_SCROLL_LOCK_MASK;
    if(keys[VK_LCONTROL] & 0x80 || keys[VK_RCONTROL] & 0x80)
		ev->state |= UI_CONTROL_MASK;
    if(keys[VK_LSHIFT] & 0x80 || keys[VK_RSHIFT] & 0x80)
		ev->state |= UI_SHIFT_MASK;
	if(keys[18] & 0x80)
		ev->state |= UI_ALT_MASK;
    if(keys[VK_LMENU] & 0x80 || keys[VK_RMENU] & 0x80)
		ev->state |= UI_SUPER_MASK;
    if(keys[VK_LBUTTON] & 0x80)
		ev->button |= UI_LBUTTON_MASK;
    if(keys[VK_MBUTTON] & 0x80)
		ev->button |= UI_MBUTTON_MASK;
    if(keys[VK_RBUTTON] & 0x80)
		ev->button |= UI_RBUTTON_MASK;
}

static void
win32_get_mouse_pos(HWND wnd,POINT *psc,ui_event_p ev)
{
	POINT pcc;
	pcc.x = ev->screen_x = psc->x;
	pcc.y = ev->screen_y = psc->y;
	if(ScreenToClient(wnd,&pcc))
	{
		ev->client_x = pcc.x;
		ev->client_y = pcc.y;
	}
	else
	{
		ev->client_x = psc->x;
		ev->client_y = psc->y;
	}
}

/*
 * Should be called _after_ win32_get_key_modifiers().
 */
static void
win32_get_char_codes(arch_window_p aw,WPARAM wParam,LPARAM lParam,ui_event_p ev)
{
	WORD chars[4];
	WCHAR uchars[4];
	UINT oem_scan = (lParam >> 16) & 0xFF;
	ev->key.acode = ev->key.ucode = 0;
	if(ToAscii(wParam,oem_scan,aw->oem_kbstate,chars,0) == 1)
		ev->key.acode = chars[0] & 0xFF;
	if(ToUnicode(wParam,oem_scan,aw->oem_kbstate,uchars,4,0) == 1)
		ev->key.ucode = uchars[0];
}

#define	SIGNED_LOWORD(src) \
	(((src) & 0x8000) ? LOWORD(src) | 0xFFFF0000 : LOWORD(src))

#define	SIGNED_HIWORD(src) \
	(((src) & 0x80000000) ? HIWORD(src) | 0xFFFF0000 : HIWORD(src))

static void
win32_window_proc(arch_window_p aw,
	UINT msg,WPARAM wParam,LPARAM lParam,DWORD time,POINT *psc)
{
	POINT def_psc;
	RECT rc;
	PAINTSTRUCT ps;
	ui_event_t dst;
	GLboolean zap_message = GL_FALSE;
	if(!time)
		time = GetTickCount();
	if(!psc)
	{
		psc = &def_psc;
		GetCursorPos(psc);
	}
	dst.time = ui_make_message_time(&aw->start_time,time);
	dst.type = ui_no_event;
	switch(msg)
	{
		case WM_SHOWWINDOW:
			if(wParam)
			{
				aw->visible = GL_TRUE;
				dst.type = ui_window_map_event;
			}
			else
			{
				aw->visible = GL_FALSE;
				dst.type = ui_window_unmap_event;
			}
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_SYSCOMMAND:
			switch(wParam)
			{
				case SC_MINIMIZE:
					if(!aw->visible)
						zap_message = GL_TRUE;
					else
					{
						aw->visible = GL_FALSE;
						dst.type = ui_window_unmap_event;
					}
					break;

				case SC_MAXIMIZE:
				case SC_RESTORE:
					if(aw->visible)
						zap_message = GL_TRUE;
					else
					{
						aw->visible = GL_TRUE;
						dst.type = ui_window_map_event;
					}
					break;

				default:
					zap_message = GL_TRUE;
			}
			if(!zap_message)
			{
				win32_get_mouse_pos(aw->wnd,psc,&dst);
				win32_get_key_modifiers(aw,&dst);
				ui_make_default_key_event(&dst.key);
				ui_set_configure_from_aw(aw,&dst.configure);
			}
			break;

		case WM_SETFOCUS:
		case WM_KILLFOCUS:
			if(msg == WM_SETFOCUS)
			{
				aw->focused = GL_TRUE;
				dst.type = ui_window_focus_in_event;
			}
			else
			{
				aw->focused = GL_FALSE;
				dst.type = ui_window_focus_out_event;
			}
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_QUIT:
			dst.type = ui_wm_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.configure.wm_event = ui_wm_close;
			break;

		case WM_PAINT:
			memZero(&rc, sizeof(RECT));
			if(!GetUpdateRect(aw->wnd,&rc,FALSE))
				return;
			BeginPaint(aw->wnd,&ps);
			EndPaint(aw->wnd,&ps);
			dst.type = ui_window_expose_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.configure.exposed_left = rc.left;
			dst.configure.exposed_top = rc.top;
			dst.configure.exposed_right = rc.right;
			dst.configure.exposed_bottom = rc.bottom;
			break;

		case WM_MOVE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_move_event;
			/* win32_get_window_geometry */
			aw->x = (GLuint)SIGNED_LOWORD(lParam);
			aw->y = (GLuint)SIGNED_HIWORD(lParam);
			if(aw->x == -32000)
				aw->x = 0;
			if(aw->y == -32000)
				aw->y = 0;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_SIZE:
			if(GetUpdateRect(aw->wnd,&rc,FALSE))
				PatBlt(aw->dc,rc.left,rc.top,
					rc.right - rc.left,rc.bottom-rc.top,BLACKNESS);
			dst.type = ui_window_size_event;
			aw->width = (GLuint)LOWORD(lParam);
			aw->height = (GLuint)HIWORD(lParam);
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_make_default_key_event(&dst.key);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYUP:
			if(msg == WM_KEYDOWN || msg == WM_SYSKEYDOWN)
				dst.key.press = GL_TRUE;
			else
				dst.key.press = GL_FALSE;
			if(aw->auto_repeat_disabled && dst.key.press &&
				(lParam & (1 << 30)))
			{
				zap_message = GL_TRUE;
				break;
			}
			dst.type = ui_key_event;
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			dst.key.scan = win32_make_scan_code(wParam,lParam);
			win32_get_char_codes(aw,wParam,lParam,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			if(dst.key.scan != sc_unknown)
				aw->keyboard_state[dst.key.scan] = (GLubyte)(dst.key.press ? 1 : 0);
			break;

		case WM_MOUSEMOVE:
			dst.type = ui_pointer_motion_event;
			ui_make_default_key_event(&dst.key);
			dst.client_x = SIGNED_LOWORD(lParam);
			dst.client_y = SIGNED_HIWORD(lParam);
			dst.screen_x = psc->x;
			dst.screen_y = psc->y;
			win32_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			break;

		case WM_LBUTTONDOWN:
		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_MBUTTONUP:
		case WM_RBUTTONUP:
		case 522: /* WM_MOUSEWHEEL */
			dst.type = ui_key_event;
			switch(msg)
			{
				case WM_LBUTTONUP:
				case WM_MBUTTONUP:
				case WM_RBUTTONUP:
					dst.key.press = GL_FALSE;
					if(msg == aw->auto_capture_button)
					{
						archWinGrabMouse(aw, GL_FALSE);
						aw->auto_capture_button = 0;
						aw->zap_next_mousemove = GL_TRUE;
					}
					break;
				case WM_LBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_LBUTTONUP;
					break;
				case WM_MBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_MBUTTONUP;
					break;
				case WM_RBUTTONDOWN:
					dst.key.press = GL_TRUE;
					if(!aw->auto_capture_button)
						aw->auto_capture_button = WM_RBUTTONUP;
					break;
				default:
					dst.key.press = GL_TRUE;
			}
			if(aw->auto_capture_button)
				archWinGrabMouse(aw, GL_TRUE);
			win32_get_mouse_pos(aw->wnd,psc,&dst);
			win32_get_key_modifiers(aw,&dst);
			ui_set_configure_from_aw(aw,&dst.configure);
			dst.key.acode = dst.key.ucode = 0; 
			if(msg == 522) /* WM_MOUSEWHEEL */
			{
				aw->wheel_delta += SIGNED_HIWORD(wParam);
				if(abs(aw->wheel_delta) < 120)
				{
					zap_message = GL_TRUE;
					break;
				}
				if(aw->wheel_delta > 0)
				{
					dst.key.scan = sc_wheelup;
					aw->keyboard_state[sc_wheeldown] = 0;
					aw->keyboard_state[sc_wheelup] = 1;
				}
				else
				{
					dst.key.scan = sc_wheeldown;
					aw->keyboard_state[sc_wheeldown] = 1;
					aw->keyboard_state[sc_wheelup] = 0;
				}
				aw->wheel_delta = 0;
			}
			else
			{
				if(msg == WM_LBUTTONDOWN || msg == WM_LBUTTONUP)
					dst.key.scan = sc_lbutton;
				else if(msg == WM_RBUTTONDOWN || msg == WM_RBUTTONUP)
					dst.key.scan = sc_rbutton;
				else
					dst.key.scan = sc_mbutton;
				aw->keyboard_state[dst.key.scan] = (GLubyte)(dst.key.press ? 1 : 0);
			}
			break;

		default:
			zap_message = GL_TRUE;
			break;
	}
	if(!zap_message && ui_default_wnd_proc(aw,&dst))
		DEQ_PUSH(event,aw->msg_que,&dst);
	return;
}

/*
 * Default callback. All system-generated messages and messages
 * unknown to win32_window_proc() have to be processed here.
 */
static LRESULT CALLBACK
win32_sys_window_proc(HWND wnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	arch_window_p aw = (arch_window_p)GetProp(wnd,LIBV_HANDLE_PROP);
	if(!aw || aw->wnd != wnd)
		aw = NULL;

	switch(msg)
	{
		case WM_CREATE:
			return 0;

		case WM_CLOSE:
			PostQuitMessage(0);
			return 0;

		case WM_ERASEBKGND:
			return 1;

		case WM_SHOWWINDOW:
		case WM_PAINT:
		case WM_MOVE:
		case WM_SIZE:
		case WM_QUIT:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
		case WM_SYSCOMMAND:
			if(aw)
				win32_window_proc(aw,msg,wParam,lParam,0,NULL);
			if(msg != WM_SYSCOMMAND)
				return 0;
			break;

		default:
			break;
	}

	return DefWindowProc(wnd,msg,wParam,lParam);
}

static void
win32_query_vid_modes(arch_window_p aw)
{
#define	MAXVIDMODES		4096
	DWORD i, count = 0;
	DEVMODE dm;
	DEVMODE modes[MAXVIDMODES];
	memZero(&dm, sizeof(DEVMODE));
	dm.dmSize = sizeof(dm);
	for(i = 0;EnumDisplaySettings(NULL, i, &dm);i++)
	{
		if(dm.dmBitsPerPel >= 15) /* Hi/TrueColor */
		{
			memcpy(&modes[count++],&dm,sizeof(DEVMODE));
			if(count == MAXVIDMODES)
				break;
		}
	}
	if(aw->vidmodes)
	{
		memFree(aw->vidmodes);
		aw->vidmodes = NULL;
	}
	aw->num_vidmodes = count;
	if(count)
	{
		vidmode_p ptr;
		aw->vidmodes = (vidmode_p)memAllocZ(count * sizeof(vidmode_t));
		ptr = aw->vidmodes;
		for(i = 0;i < count;i++)
		{
			ptr->width = modes[i].dmPelsWidth;
			ptr->height = modes[i].dmPelsHeight;
			ptr->bpp = modes[i].dmBitsPerPel;
			ptr->refresh = modes[i].dmDisplayFrequency;
			if(ptr->refresh == 1)
				ptr->refresh = 0;
			ptr++;
		}
	}
#undef	MAXVIDMODES
}

static LONG
win32_change_video_mode(vidmode_p new_mode)
{
	LONG result;
	if(new_mode)
	{
		DEVMODE dm;
		HDC refdc = GetDC(NULL);
		GLuint old_bpp = GetDeviceCaps(refdc,BITSPIXEL);
		GLuint old_refresh = GetDeviceCaps(refdc,VREFRESH);
		GLuint old_width = GetSystemMetrics(SM_CXSCREEN);
		GLuint old_height = GetSystemMetrics(SM_CYSCREEN);
		ReleaseDC(NULL,refdc);
		memZero(&dm, sizeof(DEVMODE));
		dm.dmSize = sizeof(dm);
		if(!new_mode->width)
		{
			TRACE("Using current desktop width.");
			dm.dmPelsWidth = old_width;
		}
		else
			dm.dmPelsWidth = new_mode->width;
		if(!new_mode->height)
		{
			TRACE("Using current desktop height.");
			dm.dmPelsHeight = old_height;
		}
		else
			dm.dmPelsHeight = new_mode->height;
		if(!new_mode->bpp) /* FIXME: remove any ability to touch the BPP - it is uportable */
		{
			TRACE("Using current desktop display depth.");
			dm.dmBitsPerPel = old_bpp;
		}
		else
			dm.dmBitsPerPel = new_mode->bpp;
		dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
		
		if(!new_mode->refresh)
		{
			if(dm.dmPelsWidth > old_width || dm.dmPelsHeight > old_height)
			{
				TRACE("No refresh rate for hi-res mode; using 75 Hz.");
				/* suitable both for CRT and (good) LCD monitors */
				dm.dmDisplayFrequency = 75;
			}
			else
			{
				TRACE("Using current refresh rate.");
				dm.dmDisplayFrequency = old_refresh;
			}
		}
		else
			dm.dmDisplayFrequency = new_mode->refresh;
		dm.dmFields |= DM_DISPLAYFREQUENCY;
		if(new_mode->width != old_width ||
				new_mode->height != old_height ||
				new_mode->bpp != old_bpp ||
				new_mode->refresh != old_refresh)
			{
				TRACE("Calling CDS()...");
				result = ChangeDisplaySettings(&dm,0);
				if(result == DISP_CHANGE_SUCCESSFUL)
				{
					TRACE("... Ok!");
				}
			}
		else
			result = DISP_CHANGE_SUCCESSFUL;
	}
	else
		result = ChangeDisplaySettings(NULL,0);
	if(result != DISP_CHANGE_SUCCESSFUL)
	{
		TRACE("ChangeDisplaySettings() failed.");
		return GL_FALSE;
	}
	return GL_TRUE;
}

static GLboolean
archWinSetVidMode(arch_window_p aw,GLint new_mode)
{
	vidmode_p mode;
	GLboolean result;
	if(!aw->visible)
	{
		TRACE("Vidmode change with invisible window.");
		return GL_FALSE;
	}
	if(!aw->focused)
	{
		TRACE("Vidmode change with non-focused window.");
		archWinFocus(aw);
	}
	if(new_mode == -1)
	{
		/* restore default mode */
		if(aw->current_vidmode != -1)
		{
			if(aw->auto_capture_button)
				archWinGrabMouse(aw, GL_FALSE);
			win32_change_video_mode(NULL);
			if(aw->auto_capture_button)
			{
				archWinGrabMouse(aw, GL_TRUE);
				aw->zap_next_mousemove = GL_TRUE;
			}
			SetWindowPos(aw->wnd,HWND_NOTOPMOST,0,0,0,0,
				SWP_NOSIZE|SWP_NOMOVE);
			aw->current_vidmode = -1;
			aw->root_width = GetSystemMetrics(SM_CXSCREEN);
			aw->root_height = GetSystemMetrics(SM_CYSCREEN);
		}
		return GL_TRUE;
	}
	mode = aw->vidmodes + new_mode;
	if(aw->auto_capture_button)
		archWinGrabMouse(aw, GL_FALSE);
	if(win32_change_video_mode(mode))
	{
		aw->current_vidmode = new_mode;
		SetWindowPos(aw->wnd,HWND_TOP,0,0,0,0,SWP_NOSIZE|SWP_NOMOVE);
		aw->root_width = mode->width;
		aw->root_height = mode->height;
		result = GL_TRUE;
	}
	else
		result = GL_FALSE;
	if(aw->auto_capture_button)
	{
		archWinGrabMouse(aw, GL_TRUE);
		aw->zap_next_mousemove = GL_TRUE;
	}
	return result;
}

static GLboolean
win32_register_window_class(arch_window_p aw)
{
	unsigned pid = (unsigned)GetCurrentProcessId();
	unsigned tm = (unsigned)GetTickCount();
	sprintf(aw->class_name,"AntiGlut%u%u",pid,tm);
	memZero(&aw->wc,sizeof(WNDCLASS));
	aw->wc.lpszClassName = aw->class_name;
	aw->wc.hInstance = GetModuleHandle(NULL);
	aw->wc.style = CS_CLASSDC;
	aw->wc.lpfnWndProc = win32_sys_window_proc;
	aw->wc.hIcon = LoadIcon(NULL,IDI_WINLOGO);
	aw->wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	aw->wc.hbrBackground = GetStockObject(HOLLOW_BRUSH);
	aw->wc.lpszMenuName = NULL;
	if(RegisterClass(&aw->wc))
		return GL_TRUE;
	TRACE("RegisterClass() failed.");
	return GL_FALSE;
}

static arch_window_p
archWinOpenNP(char *title,GLuint left,GLuint top,GLuint width,GLuint height,
	GLboolean caption,GLboolean allow_resize)
{
	RECT rc;
	arch_window_p aw;
	DWORD style = WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
	if(allow_resize)
	{
		if(!caption)
		{
			TRACE("Can't create resizeable window without caption.");
			return NULL;
		}
		style |= WS_OVERLAPPEDWINDOW;
	}
	else
	{
		style |= WS_POPUP;
		if(caption)
			style |= WS_CAPTION;
	}
	aw = (arch_window_p)memAllocZ(sizeof(arch_window_t));
	INIT_COMMON_WINDOW_FIELDS(aw);
	if(!win32_register_window_class(aw))
	{
		archWinClose(aw);
		return NULL;
	}
	strncpy(aw->title,title,MAX_WINDOW_TITLE);
	aw->x = left;
	aw->y = top;
	aw->root_width = GetSystemMetrics(SM_CXSCREEN);
	aw->root_height = GetSystemMetrics(SM_CYSCREEN);
	if (width == 0)
		width = aw->root_width - left;
	if (height == 0)
		height = aw->root_height - top;
	aw->width = width;
	aw->height = height;
	aw->wnd = CreateWindow(aw->class_name,
		aw->title, style, left, top,
		width, height,
		NULL,
		NULL,
		GetModuleHandle(NULL),NULL);
	if(!aw->wnd)
	{
		TRACE("CreateWindow() failed.");
		archWinClose(aw);
		return NULL;
	}
	SetProp(aw->wnd,LIBV_HANDLE_PROP,(HANDLE)aw);
	GetClientRect(aw->wnd,&rc);
	aw->delta_w = aw->width - rc.right;
	aw->delta_h = aw->height - rc.bottom;
	SetWindowPos(aw->wnd,HWND_NOTOPMOST,
		0,0,
		2 * aw->width - rc.right,
		2 * aw->height - rc.bottom,
		SWP_NOMOVE | SWP_NOZORDER);
	aw->dc = GetDC(aw->wnd);
	aw->wheel_delta = 0;
	win32_query_vid_modes(aw);
	memZero(aw->oem_kbstate,256);
	aw->zap_next_mousemove = GL_FALSE;
	return aw;
}

static GLboolean
archWinShow(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		ShowWindow(aw->wnd,SW_SHOW);
		UpdateWindow(aw->wnd);
		ui_win_loose_control(aw, GL_FALSE);
		aw->visible = GL_TRUE;
	}
	else
	{
		ui_win_loose_control(aw, GL_TRUE);
		aw->visible = GL_FALSE;
		ShowWindow(aw->wnd,SW_HIDE);
	}
	return GL_TRUE;
}

static void
archWinFocus(arch_window_p aw)
{
	if(!aw->visible)
	{
		ShowWindow(aw->wnd,SW_SHOW);
		UpdateWindow(aw->wnd);
	}
	SetFocus(aw->wnd);
	SetActiveWindow(aw->wnd);
}

/*
 * Returns 'GL_TRUE' if cursor is visible at exit.
 */
static GLboolean
archWinShowPointer(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		aw->pointer_hide_count--;
		if(!aw->pointer_hide_count)
			ShowCursor(TRUE);
	}
	else
	{
		if(!aw->pointer_hide_count)
			ShowCursor(FALSE);
		aw->pointer_hide_count++;
	}
	return (GLboolean)((aw->pointer_hide_count > 0) ? GL_FALSE : GL_TRUE);
}

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y)
{
	POINT pt;
	pt.x = x;
	pt.y = y;
	ClientToScreen(aw->wnd, &pt);
	if(pt.x < 0 || pt.y < 0 ||
			(GLuint)pt.x >= aw->root_width ||
			(GLuint)pt.y >= aw->root_height)
	{
		TRACE("Unable to warp pointer outside of root window.");
		return GL_FALSE;
	}
	SetCursorPos(pt.x, pt.y);
	return GL_TRUE;
}

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab)
{
	if(!aw->visible)
	{
		TRACE("Unable to grab mouse: window invisible.");
		return;
	}
	if(grab)
	{
		if(!aw->ms_grab_count++)
			SetCapture(aw->wnd);
	}
	else
	{
		if(!--aw->ms_grab_count)
			ReleaseCapture();
	}
}

static void
archWinClose(arch_window_p aw)
{
	if(aw->pointer_hide_count > 0)
	{
		aw->pointer_hide_count = 1;
		archWinShowPointer(aw, GL_TRUE);
	}
	ui_win_loose_control(aw, GL_TRUE);
	if(aw->wnd)
	{
		if(aw->dc)
			ReleaseDC(aw->wnd,aw->dc);
		RemoveProp(aw->wnd,LIBV_HANDLE_PROP);
		DestroyWindow(aw->wnd);
	}
	if(aw->vidmodes)
		memFree(aw->vidmodes);
	UnregisterClass(aw->class_name,GetModuleHandle(NULL));
	CLEAR_COMMON_WINDOW_FIELDS(aw);
	memFree(aw);
}

struct rejection_arg_s
{
	GLboolean met_size, met_move;
	ui_event_t last_expose;
	GLuint left, top, right, bottom;
};

static GLboolean
mvsz_rejection_proc(ui_event_p ev,void *arg)
{
	struct rejection_arg_s *rj = (struct rejection_arg_s*)arg;
	/* assume scan from last (new) to first (old) */
	switch(ev->type)
	{
		case ui_window_move_event:
			if(rj->met_move)
				return GL_TRUE;
			rj->met_move = GL_TRUE;
			break;
		case ui_window_size_event:
			if(rj->met_size)
				return GL_TRUE;
			rj->met_size = GL_TRUE;
			break;
		default:
			break;
	}
	return GL_FALSE;
}

static GLboolean
expose_collapse_proc(ui_event_p ev,void *arg)
{
	struct rejection_arg_s *rj = (struct rejection_arg_s*)arg;
	/* assume scan from first (old) to last (new) */
	if(ev->type == ui_window_expose_event)
	{
		memcpy(&rj->last_expose,ev,sizeof(ui_event_t));
		if(rj->left > ev->configure.exposed_left)
			rj->left = ev->configure.exposed_left;
		if(rj->top > ev->configure.exposed_top)
			rj->top = ev->configure.exposed_top;
		if(rj->right < ev->configure.exposed_right)
			rj->right = ev->configure.exposed_right;
		if(rj->bottom < ev->configure.exposed_bottom)
			rj->bottom = ev->configure.exposed_bottom;
		return GL_TRUE;
	}
	return GL_FALSE;
}

static event_deque_p
ui_collapse_configure_events(event_deque_p msg_que)
{
	struct rejection_arg_s rj;
	if(!msg_que)
		return NULL;
	memZero(&rj, sizeof(rj));
	rj.left = rj.top = MAXLONG;
	rj.met_size = rj.met_move = GL_FALSE;
	DEQ_SCAN_BACKWARD(event,msg_que,mvsz_rejection_proc,&rj);
	DEQ_SCAN_FORWARD(event,msg_que,expose_collapse_proc,&rj);
	if(rj.left != MAXLONG)
	{
		rj.last_expose.configure.exposed_left = rj.left;
		rj.last_expose.configure.exposed_top = rj.top;
		rj.last_expose.configure.exposed_right = rj.right;
		rj.last_expose.configure.exposed_bottom = rj.bottom;
		DEQ_PUSH(event,msg_que,&rj.last_expose);
	}
	return msg_que;
}

static void
win32_poll_events(arch_window_p aw)
{
	MSG msg;
	while(PeekMessage(&msg,aw->wnd,0,0,PM_REMOVE))
	{
		win32_window_proc(aw,msg.message,msg.wParam,msg.lParam,
			msg.time, &msg.pt);
		DispatchMessage(&msg);
	}
	aw->msg_que = ui_collapse_configure_events(aw->msg_que);
}

static void
archWinPoll(arch_window_p aw,GLboolean block)
{
	if(!block)
	{
		win32_poll_events(aw);
		return;
	}
	for(;;)
	{
		win32_poll_events(aw);
		if(aw->msg_que)
			break;
		WaitMessage();
	}
}

static GLboolean
vgl_load_proc(char *proc_name,void **ptr_ret, GLboolean *terminate)
{
	PROC proc = wglGetProcAddress((LPCSTR)proc_name);
	if(*ptr_ret && *ptr_ret != proc)
	{
		/*
		 * Different implementation (M$ generic after HW or vice versa).
		 */
		*terminate = GL_TRUE;
		return GL_FALSE;
	}
	if(proc)
	{
		*ptr_ret = (void*)proc;
		return GL_TRUE;
	}
	return GL_FALSE;
}

void
agSwap(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	SwapBuffers(ctx->wnd->dc);
}

void
agShutdown(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	global_context = NULL;
	wglMakeCurrent(NULL, NULL);
	if(ctx->wgl_context)
		wglDeleteContext(ctx->wgl_context);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	memFree(ctx);
}

#ifdef __BORLANDC__
#include <float.h>
#endif

static gl_context_p
vglInit(
	char *caption,
	GLuint x, GLuint y, GLuint width, GLuint height,
	GLboolean fullscreen,
	GLboolean can_resize,
	GLboolean need_depth,
	GLboolean need_stencil,
	GLboolean need_dst_alpha)
{
	PIXELFORMATDESCRIPTOR pfd;
	gl_context_p ctx;
	GLuint bpp;
	GLuint i, pixfmt;
	GLfloat aspect;
	GLboolean with_caption;
	ctx = alloc_context();
	if(!ctx)
		return NULL;
	if(fullscreen)
		with_caption = GL_FALSE;
	else
		with_caption = GL_TRUE;
	ctx->wnd = archWinOpenNP(caption, x, y, width, height,
		with_caption,
		can_resize);
	if(!ctx->wnd)
		goto err_exit;
	width = ctx->wnd->width;
	height = ctx->wnd->height;
	archWinShow(ctx->wnd, GL_TRUE);
	bpp = GetDeviceCaps(ctx->wnd->dc, BITSPIXEL);
	if(bpp <= 8)
	{
		if(!fullscreen)
		{
			TRACE("True or high color mode required.");
			goto err_exit;
		}
		else
			bpp = 16;
	}
	if(fullscreen)
	{
		for(i = 0; i < ctx->wnd->num_vidmodes; i++)
		{
			vidmode_p mode = ctx->wnd->vidmodes + i;
			if((GLuint)mode->width == width &&
					(GLuint)mode->height == height && mode->bpp == bpp)
			{
				if(!archWinSetVidMode(ctx->wnd, i))
				{
					TRACE("archWinSetVidMode() failed.");
					goto err_exit;
				}
				break;
			}
		}
		if(i == ctx->wnd->num_vidmodes)
		{
			TRACE("Required video mode not found.");
			goto err_exit;
		}
	}
#ifdef __BORLANDC__
	_control87(MCW_EM,MCW_EM);
#endif
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
    pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.iLayerType = PFD_MAIN_PLANE;
	pfd.cColorBits = 16;
    if(need_depth)
		pfd.cDepthBits = 16;
    if(need_stencil)
		pfd.cStencilBits = 1;
    if(need_dst_alpha)
		pfd.cAlphaBits = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER
		| PFD_SUPPORT_OPENGL;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pixfmt = ChoosePixelFormat(ctx->wnd->dc, &pfd);
	if(!pixfmt)
	{
		TRACE("ChoosePixelFormat() failed.");
		goto err_exit;
	}
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	if(!DescribePixelFormat(ctx->wnd->dc, pixfmt, pfd.nSize, &pfd))
	{
		TRACE("DescribePixelFormat() failed.");
		goto err_exit;
	}
	if(need_depth && !pfd.cDepthBits)
	{
		TRACE("Renderer doesn't support Z-buffering.");
		goto err_exit;
	}
	if(need_dst_alpha && !pfd.cAlphaBits)
	{
		TRACE("Renderer doesn't support dst alpha.");
		goto err_exit;
	}
	if(need_stencil && !pfd.cStencilBits)
	{
		TRACE("Renderer doesn't support stencil.");
		goto err_exit;
	}
	if(!SetPixelFormat(ctx->wnd->dc, pixfmt, &pfd))
	{
		TRACE("SetPixelFormat() failed.");
		goto err_exit;
	}
	ctx->wgl_context = wglCreateContext(ctx->wnd->dc);
	if(!ctx->wgl_context)
	{
		TRACE("wglCreateContext() failed.");
		goto err_exit;
	}
	if(!wglMakeCurrent(ctx->wnd->dc, ctx->wgl_context))
	{
		TRACE("wglMakeCurrent() failed.");
		goto err_exit;
	}
	ctx->vendor = strDup((char*)glGetString(GL_VENDOR));
	ctx->renderer = strDup((char*)glGetString(GL_RENDERER));
	ctx->version = strDup((char*)glGetString(GL_VERSION));
	ctx->extensions = strDup((char*)glGetString(GL_EXTENSIONS));
#ifdef	AG_VERBOSE
	printf("PFD %d, %s OpenGL on %s ver. %s.\n",
		pixfmt,ctx->vendor,ctx->renderer,ctx->version);
#endif	/* AG_VERBOSE */
	archWinPoll(ctx->wnd, GL_FALSE);
	aspect = (GLfloat)ctx->wnd->width /
		(GLfloat)ctx->wnd->height;
	glViewport(0, 0, ctx->wnd->width, ctx->wnd->height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-aspect, aspect, -1, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	return ctx;

err_exit:;
	wglMakeCurrent(NULL, NULL);
	if(ctx->wgl_context)
		wglDeleteContext(ctx->wgl_context);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	memFree(ctx);
	return NULL;
}

/*
 * Helper routines.
 */

void
agWindowTitle(char *title)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	if(!title)
		return;
	strncpy(ctx->wnd->title,title,MAX_WINDOW_TITLE);
	SetWindowText(ctx->wnd->wnd,title);
}

GLuint
agTickCount(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GetTickCount();
	return GetTickCount() - ctx->wnd->start_time;
}

GLvoid
agSleep(GLuint milliseconds)
{
	Sleep(milliseconds);
}

#else	/* __UNIX__ */

static char *local_dpy_name = "localhost:0.0";
static char *default_lang = "en_US";

#ifdef	ARCH_HAVE_XF86_VIDMODE

static void
x11_set_default_mode(x11_display_connection_p cn)
{
	if(!cn->vidmode_active || !cn->xf86_vidmodes || !cn->num_vidmodes)
		return;
	XF86VidModeLockModeSwitch(cn->dpy, cn->scrnum, False);
	XF86VidModeSwitchToMode(cn->dpy, cn->scrnum, cn->xf86_vidmodes[0]);
	cn->vidmode_active = GL_FALSE;
}

static void
x11_free_mode_lines(x11_display_connection_p cn)
{
	GLuint i;
	if(cn->xf86_vidmodes)
	{
		x11_set_default_mode(cn);
		for(i = 0; i < cn->num_vidmodes; i++)
		{
			if(cn->xf86_vidmodes[i]->privsize)
#if	defined(__cplusplus) || defined(c_plusplus)
				XFree(cn->xf86_vidmodes[i]->c_private);
#else
				XFree(cn->xf86_vidmodes[i]->private);
#endif
		}
		XFree(cn->xf86_vidmodes);
		cn->xf86_vidmodes = NULL;
	}
	if(cn->vidmodes)
	{
		memFree(cn->vidmodes);
		cn->vidmodes = NULL;
	}
	cn->num_vidmodes = 0;
}

static void
x11_get_mode_lines(x11_display_connection_p cn)
{
	GLint i, mj, mn;
	vidmode_p ptr;
	GLuint depth;
	XF86VidModeModeInfo *xm;
	if(!cn->xf86_vidmodes)
	{
		if(!XF86VidModeQueryVersion(cn->dpy, &mj, &mn))
		{
			TRACE("XFree86-VidModeExtension not found.");
			return;
		}
		XF86VidModeGetAllModeLines(cn->dpy, cn->scrnum,
			(int*)&cn->num_vidmodes, &cn->xf86_vidmodes);
		if(!cn->xf86_vidmodes || !cn->num_vidmodes)
		{
			TRACE("XF86VidModeGetAllModeLines() failed.");
			return;
		}
	}
	if(cn->vidmodes)
	{
		memFree(cn->vidmodes);
		cn->vidmodes = NULL;
	}
	cn->vidmodes = (vidmode_p)memAllocZ(cn->num_vidmodes * sizeof(vidmode_t));
	if(cn->vidmodes)
	{
		depth = DefaultDepth(cn->dpy, cn->scrnum);
		for(i = 0; i < cn->num_vidmodes; i++)
		{
			xm = cn->xf86_vidmodes[i];
			ptr = cn->vidmodes + i;
			ptr->width = xm->hdisplay;
			ptr->height = xm->vdisplay;
			ptr->bpp = depth;
			ptr->refresh = xm->dotclock * 1000 / xm->htotal / xm->vtotal;
			if(xm->flags & 32)
				ptr->refresh = ptr->refresh >> 1;
		}
	}
	cn->screen_width = WidthOfScreen(DefaultScreenOfDisplay(cn->dpy));
	cn->screen_height = HeightOfScreen(DefaultScreenOfDisplay(cn->dpy));
	return;
}

static GLboolean
archWinSetVidMode(arch_window_p aw,GLint mode_num)
{
	vidmode_p ptr;
	if(!aw->visible)
	{
		TRACE("Unable to change vidmode: window invisible.");
		return GL_FALSE;
	}
	if(mode_num == -1)
	{
		x11_set_default_mode(aw->cn);
		aw->current_vidmode = -1;
		return GL_TRUE;
	}
	ptr = aw->cn->vidmodes + mode_num;
	if(aw->cn->screen_width != ptr->width || aw->cn->screen_height != ptr->height)
	{

		XWarpPointer(aw->cn->dpy, None, aw->wnd, 0, 0, 0, 0,
			ptr->width >> 1, ptr->height >> 1);
		XF86VidModeSwitchToMode(aw->cn->dpy, aw->cn->scrnum,
			aw->cn->xf86_vidmodes[ptr->modeline]);
		XF86VidModeSetViewPort(aw->cn->dpy, aw->cn->scrnum, 0, 0);
		XF86VidModeLockModeSwitch(aw->cn->dpy, aw->cn->scrnum, True);
		aw->current_vidmode = ptr->modeline;
		aw->cn->vidmode_active = GL_TRUE;
		aw->root_width = ptr->width;
		aw->root_height = ptr->height;
	}
	return GL_TRUE;
}

#else	/* !ARCH_HAVE_XF86_VIDMODE */

static GLboolean
archWinSetVidMode(arch_window_p aw,GLuint mode_num)
{
	return GL_FALSE;
}

#endif	/* !ARCH_HAVE_XF86_VIDMODE */

/*
 * Partial copyright (C) iD Software
 */
static x11_display_connection_p
x11_open_connection()
{
	char *lang;
	Pixmap cursormask;
    XGCValues xgc;
   	GC gc;
    XColor dummycolour;
	x11_display_connection_p cn = 
		(x11_display_connection_p)memAllocZ(sizeof(x11_display_connection_t));
	if(!cn)
		return NULL;
	lang = getenv("LANG");
	if(!lang)
	{
		TRACE_1ARG("LANG not set, assuming \"%s\".",default_lang);
		lang = default_lang;
	}
	if(!setlocale(LC_ALL,lang))
	{
		TRACE_1ARG("Could not set locale \"%s\".",lang);
		setlocale(LC_ALL,default_lang);
	}
	cn->dpy = XOpenDisplay(NULL);
	if(!cn->dpy)
	{
		TRACE_1ARG("XOpenDisplay(NULL) failed; trying \"%s\".",
			local_dpy_name);
		cn->dpy = XOpenDisplay(local_dpy_name);
		if(!cn->dpy)
		{
			TRACE_1ARG("XOpenDisplay(\"%s\") failed.",
				local_dpy_name);
			memFree(cn);
			return NULL;
		}
	}
	cn->scrnum = DefaultScreen(cn->dpy);
	cn->root = RootWindow(cn->dpy, cn->scrnum);
	XSetLocaleModifiers("");
	cn->im = XOpenIM(cn->dpy, NULL, NULL, NULL);
	cn->wm_protocols_atom = XInternAtom(cn->dpy, "WM_PROTOCOLS", False); 
	cn->wm_delete_window_atom = XInternAtom(cn->dpy, "WM_DELETE_WINDOW", False);
   	cursormask = XCreatePixmap(cn->dpy, cn->root, 1, 1, 1);
   	xgc.function = GXclear;
    gc = XCreateGC(cn->dpy, cursormask, GCFunction, &xgc);
   	XFillRectangle(cn->dpy, cursormask, gc, 0, 0, 1, 1);
   	dummycolour.pixel = 0;
    dummycolour.red = 0;
   	dummycolour.flags = 04;
    cn->null_cursor = XCreatePixmapCursor(cn->dpy, cursormask,
		cursormask, &dummycolour,&dummycolour, 0, 0);
    XFreePixmap(cn->dpy, cursormask);
   	XFreeGC(cn->dpy, gc);
	cn->connection_fd = ConnectionNumber(cn->dpy);
#ifdef	ARCH_HAVE_XF86_VIDMODE
	x11_get_mode_lines(cn);
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	return cn;
}

static void
x11_close_connection(x11_display_connection_p cn)
{
#ifdef	ARCH_HAVE_XF86_VIDMODE
	x11_free_mode_lines(cn);
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	if(cn->dpy)
	{
		XUngrabKeyboard(cn->dpy, CurrentTime);
		if(cn->null_cursor)
			XFreeCursor(cn->dpy, cn->null_cursor);
		if(cn->im)
			XCloseIM(cn->im);
		XCloseDisplay(cn->dpy);
	}
	memFree(cn);
}

#ifndef XK_KP_Home
#define XK_KP_Home              0xFF95
#endif
#ifndef XK_KP_Left
#define XK_KP_Left              0xFF96
#endif
#ifndef XK_KP_Up
#define XK_KP_Up                0xFF97
#endif
#ifndef XK_KP_Right
#define XK_KP_Right             0xFF98
#endif
#ifndef XK_KP_Down
#define XK_KP_Down              0xFF99
#endif
#ifndef XK_KP_Prior
#define XK_KP_Prior             0xFF9A
#endif
#ifndef XK_KP_Next
#define XK_KP_Next              0xFF9B
#endif
#ifndef XK_KP_End
#define XK_KP_End               0xFF9C
#endif
#ifndef XK_KP_Insert
#define XK_KP_Insert            0xFF9E
#endif
#ifndef XK_KP_Delete
#define XK_KP_Delete            0xFF9F
#endif

static void
X11_init_keycode_to_scan(Display *dpy,GLuint *tab)
{
	KeySym ks;
	GLuint scan;
	GLint i, mincode, maxcode;
	XDisplayKeycodes(dpy, &mincode, &maxcode);
	if(mincode < 0)
		mincode = 0;
	if(maxcode > 255)
		maxcode = 255;
	for(i = 0;i < mincode;i++)
		tab[i] = sc_unknown;
	for(i = maxcode + 1;i < 256;i++)
		tab[i] = sc_unknown;
	for(i = mincode;i <= maxcode;i++)
	{
		ks = XKeycodeToKeysym(dpy,i,0);
		switch(ks)
		{
            case XK_KP_Page_Up: 
            case XK_KP_9: scan = sc_kp_prior; break;
            case XK_Page_Up: scan = sc_prior; break;

            case XK_KP_Page_Down: 
            case XK_KP_3: scan = sc_kp_next; break;
            case XK_Page_Down: scan = sc_next; break;

            case XK_KP_Home:
            case XK_KP_7: scan = sc_kp_home; break;
            case XK_Home: scan = sc_home; break;

            case XK_KP_End:
            case XK_KP_1: scan = sc_kp_end; break;
            case XK_End: scan = sc_end; break;

            case XK_KP_Left:
            case XK_KP_4: scan = sc_kp_left; break;
            case XK_Left: scan = sc_left; break;

            case XK_KP_Right:
            case XK_KP_6: scan = sc_kp_right; break;
            case XK_Right: scan = sc_right; break;

            case XK_KP_Down:
            case XK_KP_2: scan = sc_kp_down; break;
            case XK_Down: scan = sc_down; break;

            case XK_KP_Up:   
            case XK_KP_8: scan = sc_kp_up; break;
            case XK_Up: scan = sc_up; break;

            case XK_Escape: scan = sc_escape; break;

            case XK_KP_Enter: scan = sc_kp_enter; break;
            case XK_Return: scan = sc_enter; break;

            case XK_Tab: scan = sc_tab; break;

            case XK_F1: scan = sc_f1; break;
            case XK_F2: scan = sc_f2; break;
            case XK_F3: scan = sc_f3; break;
            case XK_F4: scan = sc_f4; break;
            case XK_F5: scan = sc_f5; break;
            case XK_F6: scan = sc_f6; break;
            case XK_F7: scan = sc_f7; break;
            case XK_F8: scan = sc_f8; break;
            case XK_F9: scan = sc_f9; break;
            case XK_F10: scan = sc_f10; break;
            case XK_F11: scan = sc_f11; break;
            case XK_F12: scan = sc_f12; break;

            case XK_BackSpace: scan = sc_backspace; break;
            case XK_KP_Delete:
            case XK_KP_Decimal: scan = sc_kp_delete; break;
            case XK_Delete: scan = sc_delete; break;

            case XK_Pause: scan = sc_pause; break;

            case XK_Shift_L: scan = sc_lshift; break;
            case XK_Shift_R: scan = sc_rshift; break;

            case XK_Execute:
            case XK_Control_L: scan = sc_lcontrol; break;
            case XK_Control_R: scan = sc_rcontrol; break;

            case XK_Alt_L:  
            case XK_Meta_L: scan = sc_lalt; break;
            case XK_Alt_R:  
            case XK_Meta_R: scan = sc_ralt; break;

            case XK_KP_Begin: scan = sc_kp_begin; break;

            case XK_KP_Insert:
            case XK_KP_0: scan = sc_kp_insert; break;
            case XK_Insert: scan = sc_insert; break;

            case XK_KP_Multiply: scan = sc_kp_multiply; break;
            case XK_KP_Add: scan = sc_kp_add; break;
            case XK_KP_Subtract: scan = sc_kp_subtract; break;
            case XK_KP_Divide: scan = sc_kp_slash; break;

            case XK_exclam: scan = sc_1; break;
            case XK_at: scan = sc_2; break;
            case XK_numbersign: scan = sc_3; break;
            case XK_dollar: scan = sc_4; break;
            case XK_percent: scan = sc_5; break;
            case XK_asciicircum: scan = sc_6; break;
            case XK_ampersand: scan = sc_7; break;
            case XK_asterisk: scan = sc_8; break;
            case XK_parenleft: scan = sc_9; break;
            case XK_parenright: scan = sc_0; break;
            case XK_twosuperior: scan = sc_grave; break;
            case XK_space:
            case XK_KP_Space: scan = sc_space; break;
		
		    case XK_Sys_Req:
		    case XK_Print: scan = sc_print; break;
		    case XK_Scroll_Lock: scan = sc_scroll_lock; break;
		    case XK_Caps_Lock: scan = sc_caps_lock; break;
		    case XK_Num_Lock: scan = sc_num_lock; break;
		    case XK_Hyper_L:
		    case XK_Super_L: scan = sc_lsuper; break;
		    case XK_Hyper_R:
		    case XK_Super_R: scan = sc_rsuper; break;
		    case XK_Menu: scan = sc_menu; break;

		    case '`':
		    case '~': scan = sc_grave; break;
		    case '1': scan = sc_1; break;
		    case '2': scan = sc_2; break;
		    case '3': scan = sc_3; break;
		    case '4': scan = sc_4; break;
		    case '5': scan = sc_5; break;
		    case '6': scan = sc_6; break;
		    case '7': scan = sc_7; break;
		    case '8': scan = sc_8; break;
		    case '9': scan = sc_9; break;
		    case '0': scan = sc_0; break;
		    case '-':
		    case '_': scan = sc_minus; break;
		    case '=':
		    case '+': scan = sc_equal; break;
		    case '\\':
		    case '|': scan = sc_backslash; break;
		    case 'q':
		    case 'Q': scan = sc_q; break;
		    case 'w':
		    case 'W': scan = sc_w; break;
		    case 'e':
		    case 'E': scan = sc_e; break;
		    case 'r':
		    case 'R': scan = sc_r; break;
		    case 't':
		    case 'T': scan = sc_t; break;
		    case 'y':
		    case 'Y': scan = sc_y; break;
		    case 'u':
		    case 'U': scan = sc_u; break;
		    case 'i':
		    case 'I': scan = sc_i; break;
		    case 'o':
		    case 'O': scan = sc_o; break;
		    case 'p':
		    case 'P': scan = sc_p; break;
		    case '[':
		    case '{': scan = sc_lbracket; break;
		    case ']':
		    case '}': scan = sc_rbracket; break;
		    case 'a':
		    case 'A': scan = sc_a; break;
		    case 's':
		    case 'S': scan = sc_s; break;
		    case 'd':
		    case 'D': scan = sc_d; break;
		    case 'f':
		    case 'F': scan = sc_f; break;
		    case 'g':
		    case 'G': scan = sc_g; break;
		    case 'h':
		    case 'H': scan = sc_h; break;
		    case 'j':
		    case 'J': scan = sc_j; break;
		    case 'k':
		    case 'K': scan = sc_k; break;
		    case 'l':
		    case 'L': scan = sc_l; break;
		    case ';':
		    case ':': scan = sc_semicolon; break;
		    case '\'':
		    case '"': scan = sc_apostrophe; break;
		    case 'z':
		    case 'Z': scan = sc_z; break;
		    case 'x':
		    case 'X': scan = sc_x; break;
		    case 'c':
		    case 'C': scan = sc_c; break;
		    case 'v':
		    case 'V': scan = sc_v; break;
		    case 'b':
		    case 'B': scan = sc_b; break;
		    case 'n':
		    case 'N': scan = sc_n; break;
		    case 'm':
		    case 'M': scan = sc_m; break;
		    case ',':
		    case '<': scan = sc_comma; break;
		    case '.':
		    case '>': scan = sc_period; break;
		    case '/':
		    case '?': scan = sc_slash; break;
			default: scan = sc_unknown;
		}
		tab[i] = scan;
	}
}

static arch_window_p
archWinOpenNP(x11_display_connection_p cn,XVisualInfo *vi,
	char *title,GLuint left,GLuint top,
	GLuint width,GLuint height,GLboolean caption,GLboolean allow_resize)
{
	arch_window_p aw;
	unsigned long mask;
	XSizeHints sh;
	XSetWindowAttributes attr;
	if (width == 0)
		width = cn->screen_width - left;
	if (height == 0)
		height = cn->screen_height - top;
	memZero(&attr,sizeof(XSetWindowAttributes));
	attr.event_mask =
		StructureNotifyMask | ExposureMask |
		KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask | VisibilityChangeMask | FocusChangeMask;
	attr.do_not_propagate_mask = KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ButtonMotionMask;
	attr.background_pixmap = None;
	attr.background_pixel = None;
	attr.border_pixmap = None;
	attr.border_pixel = None;
	attr.backing_pixel = 0;
	attr.backing_store = NotUseful;
	attr.save_under = False;
	mask = CWBackPixmap | CWBackPixel | 
		CWBorderPixmap | CWBorderPixel | 
		CWBackingPixel | CWBackingStore |
		CWEventMask | CWDontPropagate | CWSaveUnder;
	attr.colormap = XCreateColormap(cn->dpy, cn->root, vi->visual, AllocNone);
	mask |= CWColormap;
	memZero(&sh, sizeof(sh));
	sh.x = left;
	sh.y = top;
	sh.width = sh.min_width = sh.max_width = sh.base_width = width;
	sh.height = sh.min_height = sh.max_height = sh.base_height = height;
  	sh.flags = USPosition | PPosition | PSize;
	if(!allow_resize)
	{
		sh.flags |= PBaseSize | PMinSize | PMaxSize;
		if(!caption)
		{
			mask |= CWOverrideRedirect;
			attr.override_redirect = True;
		}
	}
	else if(!caption)
	{
		TRACE("Can't create resizeable window in fullscreen mode.");
		XFreeColormap(cn->dpy, attr.colormap);
		return NULL;
	}
	aw = (arch_window_p)memAllocZ(sizeof(arch_window_t));
	if(!aw)
	{
		XFreeColormap(cn->dpy, attr.colormap);
		return NULL;
	}
	INIT_COMMON_WINDOW_FIELDS(aw);
	aw->cn = cn;
	X11_init_keycode_to_scan(cn->dpy, aw->keycode_to_scan);
	strncpy(aw->title,title,MAX_WINDOW_TITLE);
	aw->x = left;
	aw->y = top;
	aw->width = width;
	aw->height = height;
	aw->parent = cn->root;
	aw->event_mask = attr.event_mask;
	aw->wnd = XCreateWindow(cn->dpy, aw->parent, left, top, width, height, 0,
		vi ? vi->depth : DefaultDepth(cn->dpy, cn->scrnum),
		InputOutput,
		vi ? vi->visual : DefaultVisual(cn->dpy, cn->scrnum), mask, &attr);
	if(!aw->wnd)
	{
		TRACE("XCreateWindow() failed.");
		archWinClose(aw);
		return NULL;
	}
	aw->ic = XCreateIC(cn->im, XNInputStyle,
		XIMPreeditNothing | XIMStatusNothing,
		XNClientWindow, aw->wnd, XNFocusWindow, aw->wnd, NULL);
	XStoreName(cn->dpy, aw->wnd, aw->title);
	XSetWMNormalHints(cn->dpy, aw->wnd, &sh);
	XSetWMProtocols(cn->dpy, aw->wnd, &cn->wm_delete_window_atom, 1);
	XMapRaised(cn->dpy, aw->wnd);
	if(!caption)
		XSetInputFocus(cn->dpy, aw->wnd, RevertToParent, CurrentTime);
	aw->auto_repeat_disabled = GL_FALSE;
	aw->cmap = attr.colormap;
	
	return aw;
}

static void
X11_make_key_codes(arch_window_p aw,XEvent *event,ui_key_event_p dst)
{
	unsigned nb, nub = 0;
	unsigned char buf[68];
	wchar_t wbuf[68];
	KeySym ks; /* stub; will not be used */
	GLuint saved_type;
	/*
	 * This is a hack to make XwcLookupString() work for KeyRelease
	 * event as for KeyPress (Xlib doc says this should _not_ be done).
	 */
	saved_type = event->type;
	event->type = KeyPress;
	nub = XwcLookupString(aw->ic, &event->xkey, wbuf, 64, &ks, NULL);
	nb = XLookupString(&event->xkey, (char*)buf, 64, &ks, NULL);
	event->type = saved_type;
	dst->scan = aw->keycode_to_scan[event->xkey.keycode];
	dst->acode = (nb > 0) ? buf[0] : 0;
	dst->ucode = (nub > 0) ? wbuf[0] : 0;
}

static void
X11_make_key_states(arch_window_p aw,GLuint x_state, ui_event_p ev)
{
	ev->state = ev->button = 0;

	if(x_state & Mod2Mask)
		ev->state |= UI_NUM_LOCK_MASK;
	if(x_state & LockMask)
		ev->state |= UI_CAPS_LOCK_MASK;
	if(x_state & Mod5Mask)
		ev->state |= UI_SCROLL_LOCK_MASK;
	if(x_state & ShiftMask)
		ev->state |= UI_SHIFT_MASK;
	if(x_state & ControlMask)
		ev->state |= UI_CONTROL_MASK;
	if(x_state & Mod1Mask)
		ev->state |= UI_ALT_MASK;
	if(x_state & Mod4Mask)
		ev->state |= UI_SUPER_MASK;

	if(x_state & Button1Mask)
		ev->button |= UI_LBUTTON_MASK;
	if(x_state & Button2Mask)
		ev->button |= UI_MBUTTON_MASK;
	if(x_state & Button3Mask)
		ev->button |= UI_RBUTTON_MASK;

	aw->last_key_state = ev->state;
	aw->last_button_state = ev->button;
}

static ui_event_type_t
X11_process_event(arch_window_p aw,XEvent *event,ui_event_p dst)
{
	x11_display_connection_p cn = aw->cn;

	switch(event->type)
	{
		case KeyPress:
		case KeyRelease:

			dst->key.press = (event->type == KeyPress) ? GL_TRUE : GL_FALSE;
			if(!dst->key.press && aw->auto_repeat_disabled)
			{
				/*
				 * The idea of XAutoRepeatOn/Off() is bad one, since
				 * we have to change global X server state.
				 *
				 * There is a better way to ignore repeated keys;
				 * it was demonstrated in glut_event.c.
				 */
				XEvent head;
	    		if(XEventsQueued(cn->dpy, QueuedAfterReading))
				{
	      			XPeekEvent(cn->dpy, &head);
	      			if(head.type == KeyPress && 
						head.xkey.window == event->xkey.window &&
	        			head.xkey.keycode == event->xkey.keycode &&
	        			head.xkey.time == event->xkey.time)
					{
	        			XNextEvent(cn->dpy, &head);
						return ui_no_event;
					}
				}
			}
			dst->type = ui_key_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xkey.x;
			dst->client_y = event->xkey.y;
			dst->screen_x = event->xkey.x_root;
			dst->screen_y = event->xkey.y_root;
			X11_make_key_states(aw, event->xkey.state, dst);
			X11_make_key_codes(aw, event, &dst->key);
			if(dst->key.scan == sc_unknown)
				return ui_no_event;
			ui_set_configure_from_aw(aw, &dst->configure);
			aw->keyboard_state[dst->key.scan] = dst->key.press ? 1 : 0;
			break;

		case ButtonPress:
		case ButtonRelease:
	
			dst->key.press = (event->type == ButtonPress) ? GL_TRUE : GL_FALSE;
			switch(event->xbutton.button)
			{
				case Button1: dst->key.scan = sc_lbutton; break;
				case Button2: dst->key.scan = sc_mbutton; break;
				case Button3: dst->key.scan = sc_rbutton; break;
				case Button4:
					if(!dst->key.press) /* fake X-generated message */
						return ui_no_event;
					dst->key.scan = sc_wheelup;
					aw->keyboard_state[sc_wheeldown] = 0;
					break;
				case Button5:
					if(!dst->key.press) /* fake X-generated message */
						return ui_no_event;
					dst->key.scan = sc_wheeldown;
					aw->keyboard_state[sc_wheelup] = 0;
					break;
				default:
					return ui_no_event;
			}
			dst->type = ui_key_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xbutton.x;
			dst->client_y = event->xbutton.y;
			dst->screen_x = event->xbutton.x_root;
			dst->screen_y = event->xbutton.y_root;
			X11_make_key_states(aw, event->xbutton.state, dst);
			dst->key.acode = dst->key.ucode = 0;
			ui_set_configure_from_aw(aw, &dst->configure);
			aw->keyboard_state[dst->key.scan] = dst->key.press ? 1 : 0;
			break;

		case MotionNotify:

			dst->type = ui_pointer_motion_event;
			dst->time =
				ui_make_message_time(&aw->start_time, event->xkey.time);
			dst->client_x = event->xmotion.x;
			dst->client_y = event->xmotion.y;
			dst->screen_x = event->xmotion.x_root;
			dst->screen_y = event->xmotion.y_root;
			X11_make_key_states(aw, event->xkey.state, dst);
			ui_make_default_key_event(&dst->key);
			ui_set_configure_from_aw(aw, &dst->configure);

			break;

		case FocusIn:
		case FocusOut:
		
			if(event->type == FocusIn)
			{
				aw->focused = GL_TRUE;
				dst->type = ui_window_focus_in_event;
				if(aw->saved_vidmode != -1)
					archWinSetVidMode(aw, aw->saved_vidmode);
			}
			else
			{
				aw->focused = GL_FALSE;
				dst->type = ui_window_focus_out_event;
				if(aw->current_vidmode != -1)
				{
					aw->saved_vidmode = aw->current_vidmode;
					archWinSetVidMode(aw, -1);
				}
			}
			dst->time = 0;
			dst->client_x = 0;
			dst->client_y = 0;
			dst->screen_x = 0;
			dst->screen_y = 0;
			ui_make_default_key_event(&dst->key);
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case ConfigureNotify:
		
			if(event->xconfigure.width != aw->width ||
				event->xconfigure.height != aw->height)
				dst->type = ui_window_size_event;
			else
				dst->type = ui_window_move_event;
			dst->time = 0;
			dst->client_x = 0;
			dst->client_y = 0;
			dst->screen_x = 0;
			dst->screen_y = 0;
			aw->x = event->xconfigure.x;
			aw->y = event->xconfigure.y;
			aw->width = event->xconfigure.width;
			aw->height = event->xconfigure.height;
			ui_make_default_key_event(&dst->key);
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case MapNotify:
		case UnmapNotify:
		
			if(event->type == MapNotify)
			{
				aw->visible = GL_TRUE;
				dst->type = ui_window_map_event;
			}
			else
			{
				aw->visible = GL_FALSE;
				dst->type = ui_window_unmap_event;
			}
			dst->time = 0;
			dst->client_x = 0;
			dst->client_y = 0;
			dst->screen_x = 0;
			dst->screen_y = 0;
			ui_make_default_key_event(&dst->key);
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			break;

		case MappingNotify:

    		XRefreshKeyboardMapping((XMappingEvent *)event);
        	break;

		case ClientMessage:

			if(event->xclient.message_type != cn->wm_protocols_atom ||
					event->xclient.format != 32 ||
					(Atom)event->xclient.data.l[0] != cn->wm_delete_window_atom)
				return ui_no_event;
			dst->type = ui_wm_event;
			dst->time = 0;
			dst->client_x = 0;
			dst->client_y = 0;
			dst->screen_x = 0;
			dst->screen_y = 0;
			ui_make_default_key_event(&dst->key);
			ui_set_configure_from_aw(aw, &dst->configure);
			dst->state = aw->last_key_state;
			dst->button = aw->last_button_state;
			dst->configure.wm_event = ui_wm_close;
			break;

	default:

		return ui_no_event;
	}

	return dst->type;
}

static ui_event_type_t
X11_post_process_event(arch_window_p aw,XEvent *event,ui_event_p dst)
{
	ui_event_type_t retval = X11_process_event(aw, event, dst);
	if(retval != ui_no_event && ui_default_wnd_proc(aw, dst))
		return retval;
	return ui_no_event;
}

static void
x11_poll_events(arch_window_p aw)
{
	ui_event_t ev;
	XEvent xev;
	GLint l, t, r, b, left, top, right, bottom;
	x11_display_connection_p cn = aw->cn;
	/*
	 * There seems to be no good method to lock viewport after
	 * mode switch. This is last resort.
	 */
#ifdef	ARCH_HAVE_XF86_VIDMODE
	if(cn->vidmode_active)
	{
		int x = 0, y = 0;
		XF86VidModeGetViewPort(cn->dpy, cn->scrnum, &x, &y);
		if(x || y)
		{
			XF86VidModeSetViewPort(cn->dpy, cn->scrnum, 0, 0);
			/* XWarpPointer(cn->dpy,None,None,0,0,0,0,-x,-y); */
			archWinWarpPointer(aw, aw->last_mx, aw->last_my);
		}
	}
#endif	/* ARCH_HAVE_XF86_VIDMODE */
	while(XCheckTypedWindowEvent(cn->dpy, aw->wnd, ClientMessage ,&xev))
	{
		if(X11_post_process_event(aw, &xev, &ev) != ui_no_event)
			DEQ_PUSH(event, aw->msg_que, &ev);
	}
	/*
	 * Remove all Expose events from queue and collapse them.
	 */
	left = top = 0x7FFFFFFF;
	right = bottom = -1;
	while(XCheckTypedWindowEvent(cn->dpy, aw->wnd, Expose, &xev))
	{
		l = xev.xexpose.x;
		t = xev.xexpose.y;
		r = l + xev.xexpose.width;
		b = t + xev.xexpose.height;
		if(left > l)
			left = l;
		if(top > t)
			top = t;
		if(right < r)
			right = r;
		if(bottom < b)
			bottom = b;
	}
	if(right != -1)
	{
		ev.type = ui_window_expose_event;
		ev.time = 0;
		ev.client_x = 0;
		ev.client_y = 0;
		ev.screen_x = 0;
		ev.screen_y = 0;
		ui_make_default_key_event(&ev.key);
		ui_set_configure_from_aw(aw, &ev.configure);
		ev.configure.exposed_left = left;
		ev.configure.exposed_top = top;
		ev.configure.exposed_right = right;
		ev.configure.exposed_bottom = bottom;
		ev.state = aw->last_key_state;
		ev.button = aw->last_button_state;
		DEQ_PUSH_BACK(event, aw->msg_que, &ev);
	}
	while(XCheckWindowEvent(cn->dpy, aw->wnd, aw->event_mask, &xev))
	{
		if(X11_post_process_event(aw, &xev, &ev) != ui_no_event)
			DEQ_PUSH(event, aw->msg_que, &ev);
	}
}

/*
 * archWinWaitMessages()
 *
 * Based on interruptibleXNextEvent() from glut.
 */
static GLboolean
archWinWaitMessages(arch_window_p aw, GLuint timeout)
{
	fd_set fds;
	struct timeval	delay;
	GLuint rc;
	XFlush(aw->cn->dpy);
    if(XPending(aw->cn->dpy))
		return GL_TRUE;
	FD_ZERO(&fds);
	FD_SET(aw->cn->connection_fd, &fds);
	if(timeout != -1)
	{
		delay.tv_sec = timeout / 1000;
		delay.tv_usec = 1000 * (timeout % 1000);
		rc = select(aw->cn->connection_fd + 1, &fds, NULL, NULL, &delay);
	}
	else
		rc = select(aw->cn->connection_fd + 1, &fds, NULL, NULL, NULL);
    if(XPending(aw->cn->dpy))
		return GL_TRUE;
	return GL_FALSE;
}

static void
archWinPoll(arch_window_p aw, GLboolean block)
{
	x11_poll_events(aw);
	if(!block || aw->msg_que)
		return;
	archWinWaitMessages(aw,-1);
	x11_poll_events(aw);
}

static void
archWinGrabKeyboard(arch_window_p aw,GLboolean grab)
{
	if((grab && aw->cn->grab_keyboard) || (!grab && !aw->cn->grab_keyboard))
		return;
	if(grab)
	{
		XGrabKeyboard(aw->cn->dpy, aw->wnd, False,
			GrabModeAsync, GrabModeAsync, CurrentTime);
			aw->cn->grab_keyboard = GL_TRUE;
	}
	else
	{
		XUngrabKeyboard(aw->cn->dpy, CurrentTime);
		aw->cn->grab_keyboard = GL_FALSE;
	}
}

static void
archWinGrabMouse(arch_window_p aw,GLboolean grab)
{
	if(!aw->visible)
	{
		TRACE("Window invisible.");
		return;
	}
	if(grab)
	{
		if(!aw->ms_grab_count++)
		{
			XGrabPointer(aw->cn->dpy, aw->wnd, True, 0,
				GrabModeAsync, GrabModeAsync, aw->wnd, None, CurrentTime);
		}
	}
	else
	{
		if(!--aw->ms_grab_count)
		{
			XUngrabPointer(aw->cn->dpy, CurrentTime);
		}
	}
}

static GLboolean
archWinShow(arch_window_p aw,GLboolean show)
{
	if(show)
	{
		XMapRaised(aw->cn->dpy, aw->wnd);
		aw->visible = GL_TRUE;
		ui_win_loose_control(aw, GL_FALSE);
	}
	else
	{
		aw->visible = GL_FALSE;
		ui_win_loose_control(aw, GL_TRUE);
		XUnmapWindow(aw->cn->dpy, aw->wnd);
	}
	return GL_TRUE;
}

/*
 * Returns 'GL_TRUE' if cursor is visible at exit.
 */
static GLboolean
archWinShowPointer(arch_window_p aw,GLboolean show)
{
	GLboolean result;
	if(show)
	{
		aw->pointer_hide_count--;
		if(!aw->pointer_hide_count)
			XUndefineCursor(aw->cn->dpy,aw->wnd);
	}
	else
	{
		if(!aw->pointer_hide_count)
			XDefineCursor(aw->cn->dpy,aw->wnd,
				aw->cn->null_cursor);
		aw->pointer_hide_count++;
	}
	return result = (aw->pointer_hide_count > 0) ? GL_FALSE : GL_TRUE;
}

static GLboolean
archWinWarpPointer(arch_window_p aw,GLint x,GLint y)
{
	XWarpPointer(aw->cn->dpy, None, aw->wnd, 0, 0, 0, 0, x, y);
	return GL_TRUE;
}

static void
archWinClose(arch_window_p aw)
{
	while(aw->pointer_hide_count > 0)
		archWinShowPointer(aw, GL_TRUE);
	if(aw->wnd)
	{
		XEvent ev;
		ui_win_loose_control(aw, GL_TRUE);
		XUnmapWindow(aw->cn->dpy, aw->wnd);
		XDestroyWindow(aw->cn->dpy, aw->wnd);
		archWinWaitMessages(aw,1);
		while(XCheckWindowEvent(aw->cn->dpy, aw->wnd, aw->event_mask, &ev));
	}
	if(aw->cmap)
		XFreeColormap(aw->cn->dpy, aw->cmap);
	CLEAR_COMMON_WINDOW_FIELDS(aw);
	memFree(aw);
}

void (*(*glx_get_proc_address)(const GLubyte *procname))(void) = NULL;

static void *libgl_handle = NULL;

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate);

GLboolean (*vgl_load_proc)(char *proc_name,void **ptr_ret,GLboolean *terminate) = vgl_load_proc_init;

static GLboolean
vgl_load_proc_real(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	void *proc = (void*)glx_get_proc_address((const GLubyte *)proc_name);
	if(*ptr_ret && (*ptr_ret != proc))
	{
		/*
		 * This should never happen under X11 + GLX.
		 */
		TRACE("Different retval of glXGetProcAddress(): GLX broken ?");
		*terminate = GL_TRUE;
		return GL_FALSE;
	}
	if(proc)
	{
		*ptr_ret = (void*)proc;
		return GL_TRUE;
	}
	return GL_FALSE;
}

static GLboolean
vgl_load_proc_init(char *proc_name,void **ptr_ret,GLboolean *terminate)
{
	GLboolean result = GL_FALSE;
	if(!libgl_handle)
	{
		for(;;)
		{
			libgl_handle = dlopen("libGL.so",RTLD_LAZY);
			if(!libgl_handle)
			{
				TRACE_1ARG("glopen(\"libGL.so\"): failed: \"%s\".", dlerror());
				break;
			}
			*((void**)&glx_get_proc_address) = dlsym(libgl_handle, "glXGetProcAddress");
			if(!glx_get_proc_address)
				*((void**)&glx_get_proc_address) = dlsym(libgl_handle, "glXGetProcAddressARB");
			if(!glx_get_proc_address)
			{
				TRACE("no glXGetProcAddress[ARB]() in libGL.so.");
				dlclose(libgl_handle);
				libgl_handle = NULL;
				break;
			}
			vgl_load_proc = vgl_load_proc_real;
			result = vgl_load_proc(proc_name, ptr_ret, terminate);
			break;
		}
	}
	return result;
}

void
agSwap()
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	glXSwapBuffers(ctx->wnd->cn->dpy, ctx->wnd->wnd);
}

static void

x11_destroy_context(gl_context_p ctx)
{
/* FIXME */
	x11_display_connection_p cn = ctx->wnd->cn;
	if(ctx->glx_context)
	{
		glXMakeCurrent(ctx->wnd->cn->dpy, None, NULL);
		glXDestroyContext(ctx->wnd->cn->dpy, ctx->glx_context);
	}
	if(ctx->visual_info)
		XFree(ctx->visual_info);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(ctx);
	x11_close_connection(cn);
	memFree(ctx);
}

void
agShutdown(void)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	global_context = NULL;
	x11_destroy_context(ctx);
}

static gl_context_p
vglInit(
	char *caption,
	GLuint x, GLuint y, GLuint width, GLuint height,
	GLboolean fullscreen,
	GLboolean can_resize,
	GLboolean need_depth,
	GLboolean need_stencil,
	GLboolean need_dst_alpha)
{
	GLfloat aspect;
	gl_context_p ctx = NULL;
	GLuint i;
	GLuint visinfo_class;
	GLint attrib[64], *aptr = attrib;
	x11_display_connection_p cn;
	cn = x11_open_connection();
	if(!cn)
		return NULL;
	ctx = alloc_context();
	*aptr++ = GLX_RGBA;
	*aptr++ = GLX_DOUBLEBUFFER;
	*aptr++ = GLX_RED_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_GREEN_SIZE;
	*aptr++ = 1;
	*aptr++ = GLX_BLUE_SIZE;
	*aptr++ = 1;
	*aptr = None;
	if(need_depth)
	{
		*aptr++ = GLX_DEPTH_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if(need_stencil)
	{
		*aptr++ = GLX_STENCIL_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	if(need_dst_alpha)
	{
		*aptr++ = GLX_ALPHA_SIZE;
		*aptr++ = 1;
		*aptr = None;
	}
	ctx->visual_info = glXChooseVisual(cn->dpy, cn->scrnum, attrib);
	if(!ctx->visual_info)
	{
		TRACE("glXChooseVisual() failed.");
		goto err_exit;
	}
#if defined(__cplusplus) || defined(c_plusplus)
	visinfo_class = ctx->visual_info->c_class;
#else
	visinfo_class = ctx->visual_info->class;
#endif
	if(visinfo_class != DirectColor && visinfo_class != TrueColor)
	{
		TRACE("DirectColor or TrueColor visual required.");
		goto err_exit;
	}
	ctx->glx_context =
		glXCreateContext(cn->dpy, ctx->visual_info, NULL, True);
	if(!ctx->glx_context)
	{
		TRACE("glXCreateContext() failed.");
		goto err_exit;
	}
	ctx->wnd = archWinOpenNP(cn,ctx->visual_info,
		caption, x, y, width, height, !fullscreen, can_resize);
	if(!ctx->wnd)
		goto err_exit;
	width = ctx->wnd->width;
	height = ctx->wnd->height;
	archWinShow(ctx->wnd, GL_TRUE);
	XFlush(cn->dpy);
	if(fullscreen)
	{
		for(i = 0; i < ctx->wnd->cn->num_vidmodes; i++)
		{
			vidmode_p mode = ctx->wnd->cn->vidmodes + i;
			if((GLuint)mode->width == width && (GLuint)mode->height == height)
			{
				if(!archWinSetVidMode(ctx->wnd, i))
				{
					TRACE("archWinSetVidMode() failed.");
					goto err_exit;
				}
				break;
			}
		}
		if(i == ctx->wnd->cn->num_vidmodes)
		{
			TRACE("Required video mode not found.");
			goto err_exit;
		}
	}
	if(!glXMakeCurrent(cn->dpy, ctx->wnd->wnd, ctx->glx_context))
	{
		TRACE("glXMakeCurrent() failed.");
		goto err_exit;
	}
	ctx->vendor = strDup((char*)glGetString(GL_VENDOR));
	ctx->renderer = strDup((char*)glGetString(GL_RENDERER));
	ctx->version = strDup((char*)glGetString(GL_VERSION));
	ctx->extensions = strDup((char*)glGetString(GL_EXTENSIONS));
#ifdef	AG_VERBOSE
	printf("%s OpenGL on %s ver. %s.\n",
		ctx->vendor,ctx->renderer,ctx->version);
#endif	/* AG_VERBOSE */
	archWinPoll(ctx->wnd, GL_FALSE);
	aspect = (GLfloat)ctx->wnd->width / (GLfloat)ctx->wnd->height;
	glViewport(0, 0, ctx->wnd->width, ctx->wnd->height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-aspect, aspect, -1, 1, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	return ctx;

err_exit:;
	x11_destroy_context(ctx);
	return NULL;
}

void
agWindowTitle(char *title)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	if(!title)
		return;
	strncpy(ctx->wnd->title, title, MAX_WINDOW_TITLE);
	XStoreName(ctx->wnd->cn->dpy, ctx->wnd->wnd, title);
}

GLuint
agTickCount(void)
{
	gl_context_p ctx;
	struct timezone tz = {0,0};
	struct timeval tv;
	unsigned tick;
	gettimeofday((struct timeval*)&tv,&tz);
	tick = tv.tv_sec * 1000 + tv.tv_usec / 1000;
	IF_NO_GLCONTEXT(ctx)
		return tick;
	return tick - ctx->wnd->start_time;
}

GLvoid
agSleep(GLuint milliseconds)
{
	struct timeval delay;
	delay.tv_sec = milliseconds / 1000;
	delay.tv_usec = 1000 * (milliseconds % 1000);
	select(0,NULL,NULL,NULL,&delay);
}

#endif	/* __UNIX__ */

/*
 * OGL 2.0 loading stuff.
 */

#define	LGLP(name) \
	(flag = (GLboolean)(flag && vgl_load_proc(#name,(void**)(&name),terminate)))

#define	RGLP(name) \
	if(name != NULL) \
		*terminate = GL_TRUE; \
	name = NULL

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.2.
 */

void (VGLEXTAPIENTRY *glBlendColor) (GLclampf, GLclampf, GLclampf, GLclampf);
void (VGLEXTAPIENTRY *glBlendEquation) (GLenum);
void (VGLEXTAPIENTRY *glDrawRangeElements) (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glTexImage3D) (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glCopyTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei);
void (VGLEXTAPIENTRY *glColorTable) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glColorTableParameterfv) (GLenum, GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glColorTableParameteriv) (GLenum, GLenum, const GLint *);
void (VGLEXTAPIENTRY *glCopyColorTable) (GLenum, GLenum, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glGetColorTable) (GLenum, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetColorTableParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetColorTableParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glColorSubTable) (GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glCopyColorSubTable) (GLenum, GLsizei, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glConvolutionFilter1D) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glConvolutionFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
void (VGLEXTAPIENTRY *glConvolutionParameterf) (GLenum, GLenum, GLfloat);
void (VGLEXTAPIENTRY *glConvolutionParameterfv) (GLenum, GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glConvolutionParameteri) (GLenum, GLenum, GLint);
void (VGLEXTAPIENTRY *glConvolutionParameteriv) (GLenum, GLenum, const GLint *);
void (VGLEXTAPIENTRY *glCopyConvolutionFilter1D) (GLenum, GLenum, GLint, GLint, GLsizei);
void (VGLEXTAPIENTRY *glCopyConvolutionFilter2D) (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei);
void (VGLEXTAPIENTRY *glGetConvolutionFilter) (GLenum, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetConvolutionParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetConvolutionParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetSeparableFilter) (GLenum, GLenum, GLenum, GLvoid *, GLvoid *, GLvoid *);
void (VGLEXTAPIENTRY *glSeparableFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *, const GLvoid *);
void (VGLEXTAPIENTRY *glGetHistogram) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetHistogramParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetHistogramParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetMinmax) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
void (VGLEXTAPIENTRY *glGetMinmaxParameterfv) (GLenum, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetMinmaxParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glHistogram) (GLenum, GLsizei, GLenum, GLboolean);
void (VGLEXTAPIENTRY *glMinmax) (GLenum, GLenum, GLboolean);
void (VGLEXTAPIENTRY *glResetHistogram) (GLenum);
void (VGLEXTAPIENTRY *glResetMinmax) (GLenum);

static GLboolean
vgl_load_gl_1_2(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glColorTable);
	LGLP(glColorTableParameterfv);
	LGLP(glColorTableParameteriv);
	LGLP(glCopyColorTable);
	LGLP(glGetColorTable);
	LGLP(glGetColorTableParameterfv);
	LGLP(glGetColorTableParameteriv);
	LGLP(glColorSubTable);
	LGLP(glCopyColorSubTable);
	LGLP(glConvolutionFilter1D);
	LGLP(glConvolutionFilter2D);
	LGLP(glConvolutionParameterf);
	LGLP(glConvolutionParameterfv);
	LGLP(glConvolutionParameteri);
	LGLP(glConvolutionParameteriv);
	LGLP(glCopyConvolutionFilter1D);
	LGLP(glCopyConvolutionFilter2D);
	LGLP(glGetConvolutionFilter);
	LGLP(glGetConvolutionParameterfv);
	LGLP(glGetConvolutionParameteriv);
	LGLP(glGetSeparableFilter);
	LGLP(glSeparableFilter2D);
	LGLP(glGetHistogram);
	LGLP(glGetHistogramParameterfv);
	LGLP(glGetHistogramParameteriv);
	LGLP(glGetMinmax);
	LGLP(glGetMinmaxParameterfv);
	LGLP(glGetMinmaxParameteriv);
	LGLP(glHistogram);
	LGLP(glMinmax);
	LGLP(glResetHistogram);
	LGLP(glResetMinmax);
	if(!flag)
	{
		TRACE("vgl_load_gl_1_2(): Warning: GL_ARB_imaging not supported.");
		flag = GL_TRUE;
	}
	LGLP(glBlendColor);
	LGLP(glBlendEquation);
	LGLP(glDrawRangeElements);
	LGLP(glTexImage3D);
	LGLP(glTexSubImage3D);
	LGLP(glCopyTexSubImage3D);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.2 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.3.
 */

void (VGLEXTAPIENTRY *glActiveTexture) (GLenum);
void (VGLEXTAPIENTRY *glClientActiveTexture) (GLenum);
void (VGLEXTAPIENTRY *glMultiTexCoord1d) (GLenum, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord1dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord1f) (GLenum, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord1fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord1i) (GLenum, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord1iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord1s) (GLenum, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord1sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord2d) (GLenum, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord2dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord2f) (GLenum, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord2fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord2i) (GLenum, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord2iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord2s) (GLenum, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord2sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord3d) (GLenum, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord3dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord3f) (GLenum, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord3fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord3i) (GLenum, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord3iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord3s) (GLenum, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord3sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glMultiTexCoord4d) (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glMultiTexCoord4dv) (GLenum, const GLdouble *);
void (VGLEXTAPIENTRY *glMultiTexCoord4f) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glMultiTexCoord4fv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glMultiTexCoord4i) (GLenum, GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glMultiTexCoord4iv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glMultiTexCoord4s) (GLenum, GLshort, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glMultiTexCoord4sv) (GLenum, const GLshort *);
void (VGLEXTAPIENTRY *glLoadTransposeMatrixf) (const GLfloat *);
void (VGLEXTAPIENTRY *glLoadTransposeMatrixd) (const GLdouble *);
void (VGLEXTAPIENTRY *glMultTransposeMatrixf) (const GLfloat *);
void (VGLEXTAPIENTRY *glMultTransposeMatrixd) (const GLdouble *);
void (VGLEXTAPIENTRY *glSampleCoverage) (GLclampf, GLboolean);
void (VGLEXTAPIENTRY *glCompressedTexImage3D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexImage2D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexImage1D) (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage2D) (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glCompressedTexSubImage1D) (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glGetCompressedTexImage) (GLenum, GLint, GLvoid *);

static GLboolean
vgl_load_gl_1_3(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glActiveTexture);
	LGLP(glClientActiveTexture);
	LGLP(glMultiTexCoord1d);
	LGLP(glMultiTexCoord1dv);
	LGLP(glMultiTexCoord1f);
	LGLP(glMultiTexCoord1fv);
	LGLP(glMultiTexCoord1i);
	LGLP(glMultiTexCoord1iv);
	LGLP(glMultiTexCoord1s);
	LGLP(glMultiTexCoord1sv);
	LGLP(glMultiTexCoord2d);
	LGLP(glMultiTexCoord2dv);
	LGLP(glMultiTexCoord2f);
	LGLP(glMultiTexCoord2fv);
	LGLP(glMultiTexCoord2i);
	LGLP(glMultiTexCoord2iv);
	LGLP(glMultiTexCoord2s);
	LGLP(glMultiTexCoord2sv);
	LGLP(glMultiTexCoord3d);
	LGLP(glMultiTexCoord3dv);
	LGLP(glMultiTexCoord3f);
	LGLP(glMultiTexCoord3fv);
	LGLP(glMultiTexCoord3i);
	LGLP(glMultiTexCoord3iv);
	LGLP(glMultiTexCoord3s);
	LGLP(glMultiTexCoord3sv);
	LGLP(glMultiTexCoord4d);
	LGLP(glMultiTexCoord4dv);
	LGLP(glMultiTexCoord4f);
	LGLP(glMultiTexCoord4fv);
	LGLP(glMultiTexCoord4i);
	LGLP(glMultiTexCoord4iv);
	LGLP(glMultiTexCoord4s);
	LGLP(glMultiTexCoord4sv);
	LGLP(glLoadTransposeMatrixf);
	LGLP(glLoadTransposeMatrixd);
	LGLP(glMultTransposeMatrixf);
	LGLP(glMultTransposeMatrixd);
	LGLP(glSampleCoverage);
	LGLP(glCompressedTexImage3D);
	LGLP(glCompressedTexImage2D);
	LGLP(glCompressedTexImage1D);
	LGLP(glCompressedTexSubImage3D);
	LGLP(glCompressedTexSubImage2D);
	LGLP(glCompressedTexSubImage1D);
	LGLP(glGetCompressedTexImage);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.3 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.4.
 */

void (VGLEXTAPIENTRY *glBlendFuncSeparate) (GLenum, GLenum, GLenum, GLenum);
void (VGLEXTAPIENTRY *glFogCoordf) (GLfloat);
void (VGLEXTAPIENTRY *glFogCoordfv) (const GLfloat *);
void (VGLEXTAPIENTRY *glFogCoordd) (GLdouble);
void (VGLEXTAPIENTRY *glFogCoorddv) (const GLdouble *);
void (VGLEXTAPIENTRY *glFogCoordPointer) (GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glMultiDrawArrays) (GLenum, GLint *, GLsizei *, GLsizei);
void (VGLEXTAPIENTRY *glMultiDrawElements) (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
void (VGLEXTAPIENTRY *glPointParameterf) (GLenum, GLfloat);
void (VGLEXTAPIENTRY *glPointParameterfv) (GLenum, const GLfloat *);
void (VGLEXTAPIENTRY *glPointParameteri) (GLenum, GLint);
void (VGLEXTAPIENTRY *glPointParameteriv) (GLenum, const GLint *);
void (VGLEXTAPIENTRY *glSecondaryColor3b) (GLbyte, GLbyte, GLbyte);
void (VGLEXTAPIENTRY *glSecondaryColor3bv) (const GLbyte *);
void (VGLEXTAPIENTRY *glSecondaryColor3d) (GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glSecondaryColor3dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glSecondaryColor3f) (GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glSecondaryColor3fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glSecondaryColor3i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glSecondaryColor3iv) (const GLint *);
void (VGLEXTAPIENTRY *glSecondaryColor3s) (GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glSecondaryColor3sv) (const GLshort *);
void (VGLEXTAPIENTRY *glSecondaryColor3ub) (GLubyte, GLubyte, GLubyte);
void (VGLEXTAPIENTRY *glSecondaryColor3ubv) (const GLubyte *);
void (VGLEXTAPIENTRY *glSecondaryColor3ui) (GLuint, GLuint, GLuint);
void (VGLEXTAPIENTRY *glSecondaryColor3uiv) (const GLuint *);
void (VGLEXTAPIENTRY *glSecondaryColor3us) (GLushort, GLushort, GLushort);
void (VGLEXTAPIENTRY *glSecondaryColor3usv) (const GLushort *);
void (VGLEXTAPIENTRY *glSecondaryColorPointer) (GLint, GLenum, GLsizei, const GLvoid *);
void (VGLEXTAPIENTRY *glWindowPos2d) (GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glWindowPos2dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glWindowPos2f) (GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glWindowPos2fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glWindowPos2i) (GLint, GLint);
void (VGLEXTAPIENTRY *glWindowPos2iv) (const GLint *);
void (VGLEXTAPIENTRY *glWindowPos2s) (GLshort, GLshort);
void (VGLEXTAPIENTRY *glWindowPos2sv) (const GLshort *);
void (VGLEXTAPIENTRY *glWindowPos3d) (GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glWindowPos3dv) (const GLdouble *);
void (VGLEXTAPIENTRY *glWindowPos3f) (GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glWindowPos3fv) (const GLfloat *);
void (VGLEXTAPIENTRY *glWindowPos3i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glWindowPos3iv) (const GLint *);
void (VGLEXTAPIENTRY *glWindowPos3s) (GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glWindowPos3sv) (const GLshort *);

static GLboolean
vgl_load_gl_1_4(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glBlendFuncSeparate);
	LGLP(glFogCoordf);
	LGLP(glFogCoordfv);
	LGLP(glFogCoordd);
	LGLP(glFogCoorddv);
	LGLP(glFogCoordPointer);
	LGLP(glMultiDrawArrays);
	LGLP(glMultiDrawElements);
	LGLP(glPointParameterf);
	LGLP(glPointParameterfv);
	LGLP(glPointParameteri);
	LGLP(glPointParameteriv);
	LGLP(glSecondaryColor3b);
	LGLP(glSecondaryColor3bv);
	LGLP(glSecondaryColor3d);
	LGLP(glSecondaryColor3dv);
	LGLP(glSecondaryColor3f);
	LGLP(glSecondaryColor3fv);
	LGLP(glSecondaryColor3i);
	LGLP(glSecondaryColor3iv);
	LGLP(glSecondaryColor3s);
	LGLP(glSecondaryColor3sv);
	LGLP(glSecondaryColor3ub);
	LGLP(glSecondaryColor3ubv);
	LGLP(glSecondaryColor3ui);
	LGLP(glSecondaryColor3uiv);
	LGLP(glSecondaryColor3us);
	LGLP(glSecondaryColor3usv);
	LGLP(glSecondaryColorPointer);
	LGLP(glWindowPos2d);
	LGLP(glWindowPos2dv);
	LGLP(glWindowPos2f);
	LGLP(glWindowPos2fv);
	LGLP(glWindowPos2i);
	LGLP(glWindowPos2iv);
	LGLP(glWindowPos2s);
	LGLP(glWindowPos2sv);
	LGLP(glWindowPos3d);
	LGLP(glWindowPos3dv);
	LGLP(glWindowPos3f);
	LGLP(glWindowPos3fv);
	LGLP(glWindowPos3i);
	LGLP(glWindowPos3iv);
	LGLP(glWindowPos3s);
	LGLP(glWindowPos3sv);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.4 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 1.5.
 */

void (VGLEXTAPIENTRY *glGenQueries) (GLsizei, GLuint *);
void (VGLEXTAPIENTRY *glDeleteQueries) (GLsizei, const GLuint *);
GLboolean (VGLEXTAPIENTRY *glIsQuery) (GLuint);
void (VGLEXTAPIENTRY *glBeginQuery) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glEndQuery) (GLenum);
void (VGLEXTAPIENTRY *glGetQueryiv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetQueryObjectiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetQueryObjectuiv) (GLuint, GLenum, GLuint *);
void (VGLEXTAPIENTRY *glBindBuffer) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glDeleteBuffers) (GLsizei, const GLuint *);
void (VGLEXTAPIENTRY *glGenBuffers) (GLsizei, GLuint *);
GLboolean (VGLEXTAPIENTRY *glIsBuffer) (GLuint);
void (VGLEXTAPIENTRY *glBufferData) (GLenum, GLsizeiptr, const GLvoid *, GLenum);
void (VGLEXTAPIENTRY *glBufferSubData) (GLenum, GLintptr, GLsizeiptr, const GLvoid *);
void (VGLEXTAPIENTRY *glGetBufferSubData) (GLenum, GLintptr, GLsizeiptr, GLvoid *);
GLvoid* (VGLEXTAPIENTRY *glMapBuffer) (GLenum, GLenum);
GLboolean (VGLEXTAPIENTRY *glUnmapBuffer) (GLenum);
void (VGLEXTAPIENTRY *glGetBufferParameteriv) (GLenum, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetBufferPointerv) (GLenum, GLenum, GLvoid* *);

static GLboolean
vgl_load_gl_1_5(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
	LGLP(glGenQueries);
	LGLP(glDeleteQueries);
	LGLP(glIsQuery);
	LGLP(glBeginQuery);
	LGLP(glEndQuery);
	LGLP(glGetQueryiv);
	LGLP(glGetQueryObjectiv);
	LGLP(glGetQueryObjectuiv);
	LGLP(glBindBuffer);
	LGLP(glDeleteBuffers);
	LGLP(glGenBuffers);
	LGLP(glIsBuffer);
	LGLP(glBufferData);
	LGLP(glBufferSubData);
	LGLP(glGetBufferSubData);
	LGLP(glMapBuffer);
	LGLP(glUnmapBuffer);
	LGLP(glGetBufferParameteriv);
	LGLP(glGetBufferPointerv);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 1.5 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 2.0.
 */

void (VGLEXTAPIENTRY *glBlendEquationSeparate) (GLenum, GLenum);
void (VGLEXTAPIENTRY *glDrawBuffers) (GLsizei, const GLenum *);
void (VGLEXTAPIENTRY *glStencilOpSeparate) (GLenum, GLenum, GLenum, GLenum);
void (VGLEXTAPIENTRY *glStencilFuncSeparate) (GLenum, GLenum, GLint, GLuint);
void (VGLEXTAPIENTRY *glStencilMaskSeparate) (GLenum, GLuint);
void (VGLEXTAPIENTRY *glAttachShader) (GLuint, GLuint);
void (VGLEXTAPIENTRY *glBindAttribLocation) (GLuint, GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glCompileShader) (GLuint);
GLuint (VGLEXTAPIENTRY *glCreateProgram) (void);
GLuint (VGLEXTAPIENTRY *glCreateShader) (GLenum);
void (VGLEXTAPIENTRY *glDeleteProgram) (GLuint);
void (VGLEXTAPIENTRY *glDeleteShader) (GLuint);
void (VGLEXTAPIENTRY *glDetachShader) (GLuint, GLuint);
void (VGLEXTAPIENTRY *glDisableVertexAttribArray) (GLuint);
void (VGLEXTAPIENTRY *glEnableVertexAttribArray) (GLuint);
void (VGLEXTAPIENTRY *glGetActiveAttrib) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
void (VGLEXTAPIENTRY *glGetActiveUniform) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
void (VGLEXTAPIENTRY *glGetAttachedShaders) (GLuint, GLsizei, GLsizei *, GLuint *);
GLint (VGLEXTAPIENTRY *glGetAttribLocation) (GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glGetProgramiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetProgramInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
void (VGLEXTAPIENTRY *glGetShaderiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetShaderInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
void (VGLEXTAPIENTRY *glGetShaderSource) (GLuint, GLsizei, GLsizei *, GLchar *);
GLint (VGLEXTAPIENTRY *glGetUniformLocation) (GLuint, const GLchar *);
void (VGLEXTAPIENTRY *glGetUniformfv) (GLuint, GLint, GLfloat *);
void (VGLEXTAPIENTRY *glGetUniformiv) (GLuint, GLint, GLint *);
void (VGLEXTAPIENTRY *glGetVertexAttribdv) (GLuint, GLenum, GLdouble *);
void (VGLEXTAPIENTRY *glGetVertexAttribfv) (GLuint, GLenum, GLfloat *);
void (VGLEXTAPIENTRY *glGetVertexAttribiv) (GLuint, GLenum, GLint *);
void (VGLEXTAPIENTRY *glGetVertexAttribPointerv) (GLuint, GLenum, GLvoid* *);
GLboolean (VGLEXTAPIENTRY *glIsProgram) (GLuint);
GLboolean (VGLEXTAPIENTRY *glIsShader) (GLuint);
void (VGLEXTAPIENTRY *glLinkProgram) (GLuint);
void (VGLEXTAPIENTRY *glShaderSource) (GLuint, GLsizei, const GLchar* *, const GLint *);
void (VGLEXTAPIENTRY *glUseProgram) (GLuint);
void (VGLEXTAPIENTRY *glUniform1f) (GLint, GLfloat);
void (VGLEXTAPIENTRY *glUniform2f) (GLint, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform3f) (GLint, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform4f) (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glUniform1i) (GLint, GLint);
void (VGLEXTAPIENTRY *glUniform2i) (GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform3i) (GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform4i) (GLint, GLint, GLint, GLint, GLint);
void (VGLEXTAPIENTRY *glUniform1fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform2fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform3fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform4fv) (GLint, GLsizei, const GLfloat *);
void (VGLEXTAPIENTRY *glUniform1iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform2iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform3iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniform4iv) (GLint, GLsizei, const GLint *);
void (VGLEXTAPIENTRY *glUniformMatrix2fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glUniformMatrix3fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glUniformMatrix4fv) (GLint, GLsizei, GLboolean, const GLfloat *);
void (VGLEXTAPIENTRY *glValidateProgram) (GLuint);
void (VGLEXTAPIENTRY *glVertexAttrib1d) (GLuint, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib1dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib1f) (GLuint, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib1fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib1s) (GLuint, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib1sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib2d) (GLuint, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib2dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib2f) (GLuint, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib2fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib2s) (GLuint, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib2sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib3d) (GLuint, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib3dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib3f) (GLuint, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib3fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib3s) (GLuint, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib3sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nbv) (GLuint, const GLbyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4Niv) (GLuint, const GLint *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nsv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nub) (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
void (VGLEXTAPIENTRY *glVertexAttrib4Nubv) (GLuint, const GLubyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nuiv) (GLuint, const GLuint *);
void (VGLEXTAPIENTRY *glVertexAttrib4Nusv) (GLuint, const GLushort *);
void (VGLEXTAPIENTRY *glVertexAttrib4bv) (GLuint, const GLbyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4d) (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
void (VGLEXTAPIENTRY *glVertexAttrib4dv) (GLuint, const GLdouble *);
void (VGLEXTAPIENTRY *glVertexAttrib4f) (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
void (VGLEXTAPIENTRY *glVertexAttrib4fv) (GLuint, const GLfloat *);
void (VGLEXTAPIENTRY *glVertexAttrib4iv) (GLuint, const GLint *);
void (VGLEXTAPIENTRY *glVertexAttrib4s) (GLuint, GLshort, GLshort, GLshort, GLshort);
void (VGLEXTAPIENTRY *glVertexAttrib4sv) (GLuint, const GLshort *);
void (VGLEXTAPIENTRY *glVertexAttrib4ubv) (GLuint, const GLubyte *);
void (VGLEXTAPIENTRY *glVertexAttrib4uiv) (GLuint, const GLuint *);
void (VGLEXTAPIENTRY *glVertexAttrib4usv) (GLuint, const GLushort *);
void (VGLEXTAPIENTRY *glVertexAttribPointer) (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);

static GLboolean
vgl_load_gl_2_0(GLboolean *terminate)
{
	GLboolean flag = GL_TRUE;
    LGLP(glBlendEquationSeparate);
    LGLP(glDrawBuffers);
    LGLP(glStencilOpSeparate);
    LGLP(glStencilFuncSeparate);
    LGLP(glStencilMaskSeparate);
    LGLP(glAttachShader);
    LGLP(glBindAttribLocation);
    LGLP(glCompileShader);
	LGLP(glCreateProgram);
	LGLP(glCreateShader);
    LGLP(glDeleteProgram);
    LGLP(glDeleteShader);
    LGLP(glDetachShader);
    LGLP(glDisableVertexAttribArray);
    LGLP(glEnableVertexAttribArray);
    LGLP(glGetActiveAttrib);
    LGLP(glGetActiveUniform);
    LGLP(glGetAttachedShaders);
    LGLP(glGetAttribLocation);
    LGLP(glGetProgramiv);
    LGLP(glGetProgramInfoLog);
    LGLP(glGetShaderiv);
    LGLP(glGetShaderInfoLog);
    LGLP(glGetShaderSource);
    LGLP(glGetUniformLocation);
    LGLP(glGetUniformfv);
    LGLP(glGetUniformiv);
    LGLP(glGetVertexAttribdv);
    LGLP(glGetVertexAttribfv);
    LGLP(glGetVertexAttribiv);
    LGLP(glGetVertexAttribPointerv);
    LGLP(glIsProgram);
    LGLP(glIsShader);
    LGLP(glLinkProgram);
    LGLP(glShaderSource);
    LGLP(glUseProgram);
    LGLP(glUniform1f);
    LGLP(glUniform2f);
    LGLP(glUniform3f);
    LGLP(glUniform4f);
    LGLP(glUniform1i);
    LGLP(glUniform2i);
    LGLP(glUniform3i);
    LGLP(glUniform4i);
    LGLP(glUniform1fv);
    LGLP(glUniform2fv);
    LGLP(glUniform3fv);
    LGLP(glUniform4fv);
    LGLP(glUniform1iv);
    LGLP(glUniform2iv);
    LGLP(glUniform3iv);
    LGLP(glUniform4iv);
    LGLP(glUniformMatrix2fv);
    LGLP(glUniformMatrix3fv);
    LGLP(glUniformMatrix4fv);
    LGLP(glValidateProgram);
    LGLP(glVertexAttrib1d);
    LGLP(glVertexAttrib1dv);
    LGLP(glVertexAttrib1f);
    LGLP(glVertexAttrib1fv);
    LGLP(glVertexAttrib1s);
    LGLP(glVertexAttrib1sv);
    LGLP(glVertexAttrib2d);
    LGLP(glVertexAttrib2dv);
    LGLP(glVertexAttrib2f);
    LGLP(glVertexAttrib2fv);
    LGLP(glVertexAttrib2s);
    LGLP(glVertexAttrib2sv);
    LGLP(glVertexAttrib3d);
    LGLP(glVertexAttrib3dv);
    LGLP(glVertexAttrib3f);
    LGLP(glVertexAttrib3fv);
    LGLP(glVertexAttrib3s);
    LGLP(glVertexAttrib3sv);
    LGLP(glVertexAttrib4Nbv);
    LGLP(glVertexAttrib4Niv);
    LGLP(glVertexAttrib4Nsv);
    LGLP(glVertexAttrib4Nub);
    LGLP(glVertexAttrib4Nubv);
    LGLP(glVertexAttrib4Nuiv);
    LGLP(glVertexAttrib4Nusv);
    LGLP(glVertexAttrib4bv);
    LGLP(glVertexAttrib4d);
    LGLP(glVertexAttrib4dv);
    LGLP(glVertexAttrib4f);
    LGLP(glVertexAttrib4fv);
    LGLP(glVertexAttrib4iv);
    LGLP(glVertexAttrib4s);
    LGLP(glVertexAttrib4sv);
    LGLP(glVertexAttrib4ubv);
    LGLP(glVertexAttrib4uiv);
    LGLP(glVertexAttrib4usv);
    LGLP(glVertexAttribPointer);
	if(*terminate)
		return GL_FALSE;
	TRACE_1ARG("OpenGL 2.0 %s.",
		flag ? "initialized" : "failed to initialize");
	return flag;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * OpenGL 2.0 support routine.
 */

static GLboolean
ag_load_gl(void)
{
	GLboolean pass = GL_TRUE, terminate = GL_FALSE;
	pass = (GLboolean)(pass && vgl_load_gl_1_2(&terminate));
	pass = (GLboolean)(pass && vgl_load_gl_1_3(&terminate));
	pass = (GLboolean)(pass && vgl_load_gl_1_4(&terminate));
	pass = (GLboolean)(pass && vgl_load_gl_1_5(&terminate));
	pass = (GLboolean)(pass && vgl_load_gl_2_0(&terminate));
	if(terminate)
	{
		TRACE("Unable to switch to another GL implementation after first init.");
		return GL_FALSE;
	}
	return pass;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Event queue management.
 */

void
uiAutoRepeat(GLboolean enable)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return;
	ctx->wnd->auto_repeat_disabled = enable ? GL_FALSE : GL_TRUE;
}

void
uiAutoWarp(GLboolean enable)
{
	gl_context_p ctx;
	arch_window_p aw;
	IF_NO_GLCONTEXT(ctx)
		return;
	aw = ctx->wnd;
	if(enable)
	{
		int cx, cy;
		if(aw->auto_warp_enabled)
			return;
		aw->auto_warp_enabled = GL_TRUE;
		aw->auto_warp_saved_x = aw->last_mx;
		aw->auto_warp_saved_y = aw->last_my;
		cx = aw->width >> 1;
		cy = aw->height >> 1;
		aw->last_vmx = 0;
		aw->last_vmy = 0;
		aw->last_mx = cx;
		aw->last_my = cy;
		archWinShowPointer(aw, GL_FALSE);
		if(archWinWarpPointer(aw, cx, cy))
		{
			archWinGrabMouse(aw, GL_TRUE);
			return;
		}
		aw->auto_warp_enabled = GL_FALSE;
		aw->last_mx = aw->last_vmx = aw->auto_warp_saved_x;
		aw->last_my = aw->last_vmy = aw->auto_warp_saved_y;
		archWinShowPointer(aw, GL_TRUE);
	}
	if(!aw->auto_warp_enabled)
		return;
	aw->auto_warp_enabled = GL_FALSE;
	aw->last_mx = aw->last_vmx = aw->auto_warp_saved_x;
	aw->last_my = aw->last_vmy = aw->auto_warp_saved_y;
	archWinGrabMouse(aw, GL_FALSE);
	archWinWarpPointer(aw, aw->auto_warp_saved_x, aw->auto_warp_saved_y);
	archWinShowPointer(aw, GL_TRUE);
}

GLboolean
uiPending()
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	archWinPoll(ctx->wnd,GL_FALSE);
	return (GLboolean)(ctx->wnd->msg_que ? GL_TRUE : GL_FALSE);
}

ui_event_p
uiPeekEvent(unsigned offset_from_first_push)
{
	ui_event_p ptr;
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return NULL;
	if(!eventDeqPeekBack(ctx->wnd->msg_que,offset_from_first_push,&ptr))
		return NULL;
	return ptr;
}

ui_event_p
uiPeekLastEvent(unsigned offset_from_last_push)
{
	ui_event_p ptr;
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return NULL;
	if(!eventDeqPeek(ctx->wnd->msg_que,offset_from_last_push,&ptr))
		return NULL;
	return ptr;
}

GLboolean
uiRemoveEvent(unsigned offset_from_first_push,ui_event_p ev_return)
{
	ui_event_p ptr;
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	if(!eventDeqPeekBack(ctx->wnd->msg_que,offset_from_first_push,&ptr))
		return GL_FALSE;
	if(ev_return)
		memcpy(ev_return,ptr,sizeof(ui_event_t));
	ctx->wnd->msg_que = eventDeqRemoveBack(ctx->wnd->msg_que,
		offset_from_first_push);
	return GL_TRUE;
}

GLboolean
uiRemoveLastEvent(unsigned offset_from_last_push,ui_event_p ev_return)
{
	ui_event_p ptr;
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	if(!eventDeqPeek(ctx->wnd->msg_que,offset_from_last_push,&ptr))
		return GL_FALSE;
	if(ev_return)
		memcpy(ev_return,ptr,sizeof(ui_event_t));
	ctx->wnd->msg_que =
		eventDeqRemove(ctx->wnd->msg_que,offset_from_last_push);
	return GL_TRUE;
}

GLboolean
uiNextEvent(ui_event_p ev_return)
{
	gl_context_p ctx;
	IF_NO_GLCONTEXT(ctx)
		return GL_FALSE;
	archWinPoll(ctx->wnd,GL_TRUE);
	ctx->wnd->msg_que = eventDeqPopBack(ctx->wnd->msg_que,ev_return);
	return GL_TRUE;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Initialization and shutting down.
 */

GLboolean
agWindowConfig(char *title,GLuint width,GLuint height,GLboolean fullscreen)
{
	if(global_context)
		return GL_FALSE;
	if(title)
	{
		strncpy(ag_window_config.title,title,512);
	}
	ag_window_config.title[511] = 0;
	ag_window_config.width = width;
	ag_window_config.height = height;
	ag_window_config.fullscreen = fullscreen;
	return GL_TRUE;
}

GLboolean
agInit(GLuint required_version,GLboolean need_depth,
	GLboolean need_stencil,GLboolean need_dst_alpha)
{
	if(global_context)
		agShutdown();
	if(required_version != AG_REQUIRED_1_1 &&
			required_version != AG_REQUIRED_2_0)
		return GL_FALSE;
	global_context = vglInit(
		ag_window_config.title,
		0,0,ag_window_config.width,ag_window_config.height,
		ag_window_config.fullscreen,
		GL_FALSE,need_depth,need_stencil,need_dst_alpha);
	if(!global_context)
		return GL_FALSE;
	if(required_version == AG_REQUIRED_2_0 && !ag_load_gl())
	{
		agShutdown();
		return GL_FALSE;
	}
	return GL_TRUE;
}
